/************************************************************************
 *                                                                      *
 *               GATT - Genetic Algorithm Time-Tabler                   *
 *                                                                      *
 *                By Dave Corne                                         *
 *                   Peter Ross (peter@aisb.ed.ac.uk),                  *
 *                   Emma Collingwood (emmac@aisb.ed.ac.uk)             *
 *                                                                      *
 ************************************************************************/

/* 
   Copyright (C) 1996, Dave Corne, Peter Ross and Emma Collingwood

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. 
*/

/**********************************************************************
 *  Filename: chromo.c
 *
 *  functions to create, copy, and free chromosomes.
 *
 **********************************************************************/ 

#include "gatt.h"

extern int nummods;
extern int numrooms;
extern int numslots;
extern int roomsmatter;
extern int agentsmatter;
extern int slotseatconstraints;
extern int sloteventconstraints;
extern int geno_size;
extern int chromo_id;


void 
create_chromo (newchromo)
     CHROMOSOME *newchromo;
{
  newchromo->gdp = (GENODATA *) malloc (sizeof (GENODATA));
  newchromo->gdp->genotype = (char *) malloc (geno_size + 1);
  if (newchromo->gdp->genotype == NULL)
    {
      printf ("not enough memory for genotype\n");
      exit (1);
    }
  newchromo->gdp->genotype[geno_size] = '\0';
  newchromo->gdp->id = chromo_id++;

  if (roomsmatter)
    {
      newchromo->eloads = (int *) calloc (numrooms * numslots, sizeof (int));
      newchromo->sloads = (int *) calloc (numrooms * numslots, sizeof (int));
      if (newchromo->eloads == (int *) NULL || newchromo->sloads == (int *) NULL)
	{
	  printf ("Not enough memory\n");
	  exit (1);
	}

    }

  if (agentsmatter)
    {
      newchromo->aeloads = (int *) calloc (numagents, sizeof (int));
      newchromo->ahloads = (int *) calloc (numagents, sizeof (int));
      if (newchromo->aeloads == (int *)NULL || newchromo->ahloads == (int *)NULL)
	{
	  printf ("Not enough memory\n");
	  exit (1);
	}
    }

  if (slotseatconstraints)
    {
      newchromo->ssloads = (int *) calloc (numslots, sizeof (int));
      if (newchromo->ssloads == NULL)
	{
	  printf ("Not enough memory\n");
	  exit (1);
	}
    }

  if (sloteventconstraints)
    {
      newchromo->seloads = (int *) calloc (numslots, sizeof (int));
      if (newchromo->seloads == NULL)
	{
	  printf ("Not enough memory\n");
	  exit (1);
	}
    }
}

void 
copy_genotype (from, to)
     CHROMOSOME *from, *to;
{
  int i;

  for (i = 0; i < geno_size; i++)
    to->gdp->genotype[i] = from->gdp->genotype[i];
}

void 
copy_loads (from, to)
     CHROMOSOME *from, *to;
{
  int i;

  if (roomsmatter)
    for (i = 0; i < numrooms * numslots; i++)
      {
	to->eloads[i] = from->eloads[i];
	to->sloads[i] = from->sloads[i];
      }
  if (agentsmatter)
    for (i = 0; i < numagents; i++)
      {
	to->aeloads[i] = from->aeloads[i];
	to->ahloads[i] = from->ahloads[i];
      }

  if (slotseatconstraints)
    for (i = 0; i < numslots; i++)
      to->ssloads[i] = from->ssloads[i];
  if (sloteventconstraints)
    for (i = 0; i < numslots; i++)
      to->seloads[i] = from->seloads[i];
}



void 
freechromo (chromo)
     CHROMOSOME *chromo;
{
  free (chromo->gdp->genotype);
  free (chromo->gdp);
  if (roomsmatter)
    {
      free (chromo->eloads);
      free (chromo->sloads);
    }
  if (agentsmatter)
    {
      free (chromo->aeloads);
      free (chromo->ahloads);
    }
  if (slotseatconstraints)
    free (chromo->ssloads);
  if (sloteventconstraints)
    free (chromo->seloads);
}


void 
copy_chromo (copy, chromo)
     CHROMOSOME *copy, *chromo;
{

  copy_genotype (chromo, copy);
  copy_loads (chromo, copy);
  copy->agentclash = chromo->agentclash;
  copy->agenttoomany = chromo->agenttoomany;
  copy->fitness = chromo->fitness;
  copy->clashes = chromo->clashes;
  copy->toocloses = chromo->toocloses;
  copy->modexcslots = chromo->modexcslots;
  copy->modavslots = chromo->modavslots;
  copy->juxtviols = chromo->juxtviols;
  copy->juxtscore = chromo->juxtscore;
  copy->nperday = chromo->nperday;
  copy->eventcapviols = chromo->eventcapviols;
  copy->studcapviols = chromo->studcapviols;
  copy->modexcrooms = chromo->modexcrooms;
  copy->modavrooms = chromo->modavrooms;
  copy->agentexcslots = chromo->agentexcslots;
  copy->agentavslots = chromo->agentavslots;
  copy->agentoverh = chromo->agentoverh;
  copy->agentunderh = chromo->agentunderh;
  copy->agentovere = chromo->agentovere;
  copy->agentexcmod = chromo->agentexcmod;
  copy->agentavmod = chromo->agentavmod;
  copy->agentprefs = chromo->agentprefs;
  copy->seload = chromo->seload;
  copy->ssload = chromo->ssload;
  copy->nclash = chromo->nclash;
  copy->ntooclose = chromo->ntooclose;
  copy->consecutives = chromo->consecutives;
  copy->student_over = chromo->student_over;
  copy->nclash_exams = chromo->nclash_exams;
  
}

CHROMOSOME 
copychromo (chromo)
     CHROMOSOME *chromo;
{
  CHROMOSOME *copy;

  /**************************************************
   * 22/5/96
   * Added code to allocate memory for copy
   **************************************************/

  copy = (CHROMOSOME *)malloc(sizeof(CHROMOSOME));
  create_chromo (copy);
  copy_genotype (chromo, copy);
  copy_loads (chromo, copy);

  /******************************************************
   * 22/5/96
   * added line to copy juxtscore from original to copy
   *******************************************************/
  copy->agentclash = chromo->agentclash;
  copy->agenttoomany = chromo->agenttoomany;
  copy->juxtscore = chromo->juxtscore;
  copy->fitness = chromo->fitness;
  copy->clashes = chromo->clashes;
  copy->toocloses = chromo->toocloses;
  copy->modexcslots = chromo->modexcslots;
  copy->modavslots = chromo->modavslots;
  copy->juxtviols = chromo->juxtviols;
  copy->nperday = chromo->nperday;
  copy->eventcapviols = chromo->eventcapviols;
  copy->studcapviols = chromo->studcapviols;
  copy->modexcrooms = chromo->modexcrooms;
  copy->modavrooms = chromo->modavrooms;
  copy->agentexcslots = chromo->agentexcslots;
  copy->agentavslots = chromo->agentavslots;
  copy->agentoverh = chromo->agentoverh;
  copy->agentunderh = chromo->agentunderh;
  copy->agentovere = chromo->agentovere;
  copy->agentexcmod = chromo->agentexcmod;
  copy->agentavmod = chromo->agentavmod;
  copy->agentprefs = chromo->agentprefs;
  copy->seload = chromo->seload;
  copy->ssload = chromo->ssload;
  copy->nclash = chromo->nclash;
  copy->ntooclose = chromo->ntooclose;
  copy->consecutives = chromo->consecutives;
  copy->student_over = chromo->student_over;
  copy->nclash_exams = chromo->nclash_exams;
  return (*copy);
}
