/************************************************************************
 *                                                                      *
 *               GATT - Genetic Algorithm Time-Tabler                   *
 *                                                                      *
 *                By Dave Corne                                         *
 *                   Peter Ross (peter@aisb.ed.ac.uk),                  *
 *                   Emma Collingwood (emmac@aisb.ed.ac.uk)             *
 *                                                                      *
 ************************************************************************/

/* 
   Copyright (C) 1996, Dave Corne, Peter Ross and Emma Collingwood

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. 
*/




/**********************************************************************
 *  Filename: init.c 
 *
 *  Initialisation functions to intialise the populations at the 
 *  beginning of the GA.
 *
 **********************************************************************/ 

#include "gatt.h"

#define NOGO_C(a,b) cantgo_c[nummods*b+a]
#define NOGO_TC(a,b) cantgo_tc[nummods*b+a]
#define NOGO_S(a,b) cantgo_s[nummods*b+a]
#define NOGO_E(a,b) cantgo_e[nummods*b+a]

void analyse_initial_pop ();
extern int numchromo;
void nogos_c ();
void nogos_tc ();
void nogos_e ();
void nogos_s ();
int in_newnogos();

int *possibles_c, *possibles_tc, *possibles_s, *possibles_e, *possibles, npossibles_c,
  npossibles_tc, npossibles_s, npossibles_e, npossibles;
char *cantgo_c, *cantgo_tc, *cantgo_s, *cantgo_e;


extern double (*init_evaluate) ();
CHROMOSOME *init_pops ();
void init_genotype ();
void init_genotype_greedy ();
extern void create_chromo();
extern double drandom ();
extern FILE *fopen ();
extern int fscanf ();
extern char *malloc ();
extern char *strdup ();
extern long random ();
extern int chromo_id;
extern int geno_size;
extern char *modslotexcs;
extern char *modroomexcs;
extern int roomsmatter;
extern int agentsmatter;
extern char *modagentexcs;
extern void check_edge_violation ();
extern void check_juxt_violation ();
extern void check_juxt_constraint();
extern int slotseatconstraints;
extern int nsimgroups;
extern int numpop;
extern int last_line;
extern double *total_abs_fitness;
extern int *bestinpop;
extern FILE *output;
typedef struct _nogo
{
  int event, slot;
} _NOGO;

typedef struct _newnogos
{
  int n;			/* Number of nogos here */
  _NOGO *nogos[MAXMODS * MAXSLOTS];
} _NEWNOGOS;

_NOGO *newnogoalloc ();
_NEWNOGOS *newnewnogoalloc ();
_NEWNOGOS *nng, *sim_nng;

int nngs_initialised = 0;
/************************************************************************/
/* This initialises a chromosome                                        */
/************************************************************************/

void 
init_genotype (gdp)
     GENODATA *gdp;
{
  int i, allele,j;
  int ok = 0;
  gdp->refcount = 1;
  gdp->id = chromo_id++;
  gdp->parent1 = gdp->parent2 = 0;
  gdp->genotype = (char *) malloc (geno_size + 1);
  if (gdp->genotype == NULL)
    {
      printf ("No memory for genotype\n");
      exit (0);
    }
  gdp->genotype[geno_size] = '\0';	/* for printing */

  gdp->slotperm = (char *) malloc (numslots + 1);
  if (gdp->slotperm == NULL)
    {
      printf ("No memory for slotperm\n");
      exit (0);
    }


  /* first check that the while loop can terminate - ie
   * there are no exams that are excluded from all slots */

  for (i = 0; i < nummods; i++){
    for (j=0;j<numslots;j++){
      if (modslotexcs[nummods * j + i] == 0){
	/* this module is ok so break */
	ok = 1;
	break;
      }
    }
    /* if we are here and ok is still 0, then module has all slots
     * excluded */
    if (ok == 0){
      printf("trouble...module %s is excluded from all slots\n", a_modules[i]->name);
      endwin();
      exit(0);
    }
    else
      ok = 0;
  }


  for (i = 0; i < nummods; i++)
    {
      if (a_modules[i]->slotspec > -1)
	gdp->genotype[i] = a_modules[i]->slotspec;
      else
	{
	  allele = random ()% numslots;
	  while (modslotexcs[nummods * allele + i] == 1)
	    allele = random ()% numslots;
	  gdp->genotype[i] = allele;
	}
    }

  if (roomsmatter)
    for (i = 0; i < nummods; i++)
      {
	if (a_modules[i]->roomspec > -1)
	  gdp->genotype[nummods + i] = a_modules[i]->roomspec;
	else
	  {
	    allele = random ()% numrooms;
	    while (modroomexcs[nummods * allele + i] == 1)
	      allele = random ()% numrooms;
	    gdp->genotype[nummods + i] = allele;
	  }
      }

  if (agentsmatter)
    for (i = 0; i < nummods; i++)
      {
	if (a_modules[i]->agentspec > -1)
	  gdp->genotype[nummods + (nummods * roomsmatter) + i]
	    = a_modules[i]->agentspec;
	else {
	  allele = random ()% numagents;
	  while (modagentexcs[nummods * allele + i] == 1)
	    allele = random ()% numagents;

	  gdp->genotype[nummods + (nummods * roomsmatter) + i] = allele;
	}
	/***************************************************************
         * 22/5/96
         * Removed the following line as it is unecessary and undoes
         * previous correct assignments
         ****************************************************************/
	/*	gdp->genotype[nummods + (nummods * roomsmatter) + i] =
	  (random ()% numagents);*/
      }

}

/***********************************************************************/
/* Try to make a clash-free genotype -- initialise cleverly            */
/***********************************************************************/
void 
init_genotype_greedy (gdp)
     GENODATA *gdp;
{
  int i, j, k, l, s, sg, z, zz, min, minslot;
  int *placements, *fills, nummin = 1;
  FILE *dbo;
  if (nngs_initialised == 0)
    {
      nng = newnewnogoalloc ();
      sim_nng = newnewnogoalloc ();
      for (i = 0; i < nummods * numslots; i++)
	{
	  nng->nogos[i] = newnogoalloc ();
	  sim_nng->nogos[i] = newnogoalloc ();
	}

      if ((dbo = fopen ("zwump", "a")) == NULL)
	{
	  printf ("Failed to open file\n");
	  exit (1);
	}

      nngs_initialised = 1;
      cantgo_c = (char *) calloc (nummods * numslots, sizeof (char));
      if (cantgo_c == NULL)
	{
	  printf ("No memory for cantgo_c\n");
	  exit (0);
	}

      cantgo_tc = (char *) calloc (nummods * numslots, sizeof (char));
      if (cantgo_tc == NULL)
	{
	  printf ("No memory for cantgo_tc\n");
	  exit (0);
	}

      cantgo_s = (char *) calloc (nummods * numslots, sizeof (char));
      if (cantgo_s == NULL)
	{
	  printf ("No memory for cantgo_s\n");
	  exit (0);
	}

      cantgo_e = (char *) calloc (nummods * numslots, sizeof (char));
      if (cantgo_e == NULL)
	{
	  printf ("No memory for cantgo_e\n");
	  exit (0);
	}

      possibles = (int *) malloc (numslots * sizeof (int));
      if (possibles == NULL)
	{
	  printf ("No memory for possibles\n");
	  exit (0);
	}

      possibles_c = (int *) malloc (numslots * sizeof (int));
      if (possibles_c == NULL)
	{
	  printf ("No memory for possibles_c\n");
	  exit (0);
	}

      possibles_tc = (int *) malloc (numslots * sizeof (int));
      if (possibles_tc == NULL)
	{
	  printf ("No memory for possibles_tc\n");
	  exit (0);
	}

      if (slotseatconstraints)
	{
	  possibles_s = (int *) malloc (numslots * sizeof (int));
	  if (possibles_s == NULL)
	    {
	      printf ("No memory for possibles_s\n");
	      exit (0);
	    }
	}

      possibles_e = (int *) malloc (numslots * sizeof (int));
      if (possibles_e == NULL)
	{
	  printf ("No memory for possibles_e\n");
	  exit (0);
	}

    }
  gdp->refcount = 1;
  gdp->id = chromo_id++;
  gdp->parent1 = gdp->parent2 = 0;
  gdp->genotype = (char *) malloc (geno_size + 1);
  if (gdp->genotype == NULL)
    {
      printf ("No memory for genotype\n");
      exit (0);
    }
  gdp->genotype[geno_size] = '\0';	/* for printing */

  for (i = 0; i < nummods * numslots; i++)
    cantgo_c[i] = cantgo_tc[i] = cantgo_s[i] = cantgo_e[i] = 0;


  /* For each module, put it somewhere it can go without clashing with */
  /* anything. One by one. If not possible, then put it in a cango2 place*/
  /* If this not possible, then put it in a not cango1 but cango2       */

  /* In particular, fill rooms up as much as possible. So prefer to put it */
  /* in the place with fewest rooms it can go */


  /* Now deal with exclusions */

  for (i = 0; i < nummods; i++)
    for (j = 0; j < numslots; j++)
      if (modslotexcs[nummods * j + i] == 1)
	NOGO_E (i, j) = 1;

  placements = (int *) calloc (nummods, sizeof (int));
  for (i = 0; i < nummods; i++)
    placements[i] = -1;
  fills = (int *) calloc (numslots, sizeof (int));
  if (fills == NULL)
    {
      printf ("No memory for fills\n");
      exit (1);
    }


  /* Deal with specified slots */

  for (i = 0; i < nummods; i++)
    if (a_modules[i]->slotspec > -1)
      {
	s = gdp->genotype[i] = a_modules[i]->slotspec;
	placements[i] = s;

	if (a_modules[i]->sim == 0)
	  {			/* If not in a a simgroup */
	    /* Then just fill nng with the new nogos */
	    nogos_c (i, placements, s);
	    for (j = 0; j < nng->n; j++)
	      NOGO_C (nng->nogos[j]->event, nng->nogos[j]->slot) = 1;
	    nogos_tc (i, placements, s);
	    for (j = 0; j < nng->n; j++)
	      NOGO_TC (nng->nogos[j]->event, nng->nogos[j]->slot) = 1;
	    if (slotseatconstraints)
	      {
		nogos_s (i, placements, s, fills);
		for (j = 0; j < nng->n; j++)
		  NOGO_S (nng->nogos[j]->event, nng->nogos[j]->slot) = 1;
	      }
	    /* Deal with fills */
	    if (slotseatconstraints)
	      fills[s] += a_modules[i]->size;
	  }
	else
	  {			/* i is a member of a simgroup */
	    for (j = 0; j < simgroups[a_modules[i]->simgroup]->num; j++)
	      placements[simgroups[a_modules[i]->simgroup]->sims[j]] = s;
	    for (j = 0; j < simgroups[a_modules[i]->simgroup]->num; j++)
	      {
		k = simgroups[a_modules[i]->simgroup]->sims[j];
		nogos_c (k, placements, s);
		for (l = 0; l < nng->n; l++)
		  NOGO_C (nng->nogos[l]->event, nng->nogos[l]->slot) = 1;
		nogos_tc (k, placements, s);
		for (l = 0; l < nng->n; l++)
		  NOGO_TC (nng->nogos[l]->event, nng->nogos[l]->slot) = 1;
		if (slotseatconstraints)
		  {
		    nogos_s (k, placements, s, fills);
		    for (l = 0; l < nng->n; l++)
		      NOGO_S (nng->nogos[l]->event, nng->nogos[l]->slot) = 1;
		  }
		/* Deal with fills */
		if (slotseatconstraints)
		  fills[s] += simgroups[a_modules[i]->simgroup]->size;
	      }
	  }
      }

  /* So, some of the timetable is sorted out; specifieds are in place */
  /* and the consequent nogos are given. Now, place the other events  */

  for (i = 0; i < nummods; i++)
    if (placements[i] < 0)
      {				/* This has not been placed yet */
	if (a_modules[i]->sim == 0)
	  {			/* and it's not a sim member */
	    /* Initialise all possibles */
	    npossibles = npossibles_e = npossibles_c = npossibles_tc = 0;
	    if (slotseatconstraints)
	      npossibles_s = 0;
	    for (j = 0; j < numslots; j++)
	      possibles[j] = possibles_e[j] = possibles_tc[j] = possibles_c[j] = 0;
	    if (slotseatconstraints)
	      possibles_s[j] = 0;

	    /* Find out where it can actually go */
	    npossibles_e = 0;
	    for (j = 0; j < numslots; j++)
	      if (NOGO_E (i, j) == 0)
		{
		  possibles_e[j] = 1;
		  npossibles_e++;
		}
	    npossibles_c = 0;
	    for (j = 0; j < numslots; j++)
	      if (NOGO_C (i, j) == 0)
		{
		  possibles_c[j] = 1;
		  npossibles_c++;
		}
	    npossibles_tc = 0;
	    for (j = 0; j < numslots; j++)
	      if (NOGO_TC (i, j) == 0)
		{
		  possibles_tc[j] = 1;
		  npossibles_tc++;
		}
	    if (slotseatconstraints)
	      {
		npossibles_s = 0;
		for (j = 0; j < numslots; j++)
		  if (NOGO_S (i, j) == 0)
		    {
		      possibles_s[j] = 1;
		      npossibles_s++;
		    }
	      }

	    /* So, possibles now contains the slots i can go in */

	    /* How to decide where to place this event? */

	    /* Well, there are various methods. Here, put it somewhere which */
	    /* is possible_e, possible_c, and possible_s, and which minimises */
	    /*  the number of new nogo_tc's which would arise */

	    /* First, find all the possibles like this */

	    npossibles = 0;
	    for (j = 0; j < numslots; j++)
	      if (possibles_e[j] && possibles_c[j] &&
		  ((slotseatconstraints == 0) || possibles_s[j]))
		{
		  possibles[j] = 1;
		  npossibles++;
		}

	    /* Now, find out what would happen in terms of tcs at */
	    /* each of the possible slots */
	    /* But first, make sure possibles is not 0. If it is */
	    /* then just look for possibles in terms of exclusions */

	    if (npossibles == 0)
	      for (j = 0; j < numslots; j++)
		if (possibles_e[j])
		  {
		    possibles[j] = 1;
		    npossibles++;
		  }

	    min = nummods * numslots;
	    nummin = 1;
	    minslot = 0;
	    for (j = 0; j < numslots; j++)
	      if (possibles[j])
		{
		  nogos_tc (i, placements, j);
		  if (nng->n < min)
		    {
		      nummin = 1;
		      min = nng->n;
		      minslot = j;
		    }
		  else if (nng->n == min)
		    {
		      nummin++;
		      if (drandom ()< (1.0 / (double) nummin))
			minslot = j;
		    }
		}
	    /* minslot is now the best choice. So, go for it */
	    placements[i] = minslot;
	    nogos_c (i, placements, minslot);
	    for (j = 0; j < nng->n; j++)
	      NOGO_C (nng->nogos[j]->event, nng->nogos[j]->slot) = 1;
	    nogos_tc (i, placements, minslot);
	    for (j = 0; j < nng->n; j++)
	      NOGO_TC (nng->nogos[j]->event, nng->nogos[j]->slot) = 1;
	    nogos_s (i, placements, minslot, fills);
	    for (j = 0; j < nng->n; j++)
	      NOGO_S (nng->nogos[j]->event, nng->nogos[j]->slot) = 1;
	    fills[minslot] += a_modules[i]->size;
	    gdp->genotype[i] = minslot;
	  }
	else
	  {			/* i is a member of a sim group */
	    /* Initialise all possibles */
	    npossibles = npossibles_e = npossibles_c = npossibles_tc = 0;
	    if (slotseatconstraints)
	      npossibles_s = 0;
	    for (j = 0; j < numslots; j++)
	      possibles[j] = possibles_e[j] = possibles_tc[j] = possibles_c[j] = 0;
	    if (slotseatconstraints)
	      possibles_s[j] = 0;

	    /* Which sim group is it in ? */
	    sg = a_modules[i]->simgroup;

	    /* Find out where it can actually go */
	    for (z = 0; z < simgroups[sg]->num, zz = simgroups[sg]->sims[z]; z++)
	      {
		for (j = 0; j < numslots; j++)
		  if (NOGO_E (zz, j) == 0)
		    possibles_e[j] = 1;
		for (j = 0; j < numslots; j++)
		  if (NOGO_C (zz, j) == 0)
		    possibles_c[j] = 1;
		for (j = 0; j < numslots; j++)
		  if (NOGO_TC (zz, j) == 0)
		    possibles_tc[j] = 1;
		if (slotseatconstraints)
		  for (j = 0; j < numslots; j++)
		    if (NOGO_S (zz, j) == 0)
		      possibles_s[j] = 1;

	      }
	    /* Find all the possibles like this */

	    npossibles = 0;
	    for (j = 0; j < numslots; j++)
	      if (possibles_e[j] && possibles_c[j] &&
		  ((slotseatconstraints == 0) || possibles_s[j]))
		{
		  possibles[j] = 1;
		  npossibles++;
		}

	    /* Now, find out what would happen in terms of tcs at */
	    /* each of the possible slots */
	    /* Since we are actually dealing with simgroup */
	    /* we have to do this for all of the members of it, collecting */
	    /* and merging the nogos as we go */

	    /* So, initialise the sim_nng;  */
	    sim_nng->n = 0;
	    min = nummods * numslots;
	    nummin = 1;
	    minslot = 0;
	    for (j = 0; j < numslots; j++)
	      if (possibles[j])
		{
		  sim_nng->n = 0;
		  for (z = 0; z < simgroups[sg]->num, zz = simgroups[sg]->sims[z]; z++)
		    {
		      nogos_tc (z, placements, j);
		      /* Merge this with the sim_nng */
		      for (k = 0; k < nng->n; k++)
			if (in_newnogos (nng->nogos[k]->event,
					 nng->nogos[k]->slot,
					 sim_nng) == 0)
			  {
			    sim_nng->nogos[sim_nng->n]->event = nng->nogos[k]->event;
			    sim_nng->nogos[sim_nng->n++]->slot = nng->nogos[k]->slot;
			  }
		    }
		  if (sim_nng->n < min)
		    {
		      nummin = 1;
		      min = sim_nng->n;
		      minslot = j;
		    }
		  else if (sim_nng->n == min)
		    {
		      nummin++;
		      if (drandom ()< (1.0 / (double) nummin))
			minslot = j;
		    }
		}

	    /* minslot is now the best choice. So, go for it */
	    placements[i] = minslot;
	    /* Update for new nogos */
	    for (z = 0; z < simgroups[sg]->num, zz = simgroups[sg]->sims[z]; z++)
	      {
		placements[zz] = minslot;
		nogos_c (zz, placements, minslot);
		for (j = 0; j < nng->n; j++)
		  NOGO_C (nng->nogos[j]->event, nng->nogos[j]->slot) = 1;
		nogos_tc (zz, placements, minslot);
		for (j = 0; j < nng->n; j++)
		  NOGO_TC (nng->nogos[j]->event, nng->nogos[j]->slot) = 1;
		nogos_s (zz, placements, minslot, fills);
		for (j = 0; j < nng->n; j++)
		  NOGO_S (nng->nogos[j]->event, nng->nogos[j]->slot) = 1;
		gdp->genotype[zz] = minslot;
	      }
	    fills[minslot] += simgroups[sg]->size;
	  }
      }

  free (fills);
  free (placements);

}


void 
fix_for_sim_groups (gdp)
     GENODATA *gdp;
{
  int i, j, slot;

  for (i = 0; i < nsimgroups; i++)
    {
      slot = gdp->genotype[simgroups[i]->sims[0]];
      for (j = 1; j < simgroups[i]->num; j++)
	gdp->genotype[simgroups[i]->sims[j]] = slot;
    }
}

/************************************************************************/
/* This creates and randomly initialises the chromosome populations     */
/************************************************************************/

CHROMOSOME *
init_pops (int numpops, int numchromo)
{
  register int chromo, pop, i;
  CHROMOSOME *pops;

  pops = (CHROMOSOME *) malloc (numpop * numchromo * sizeof (CHROMOSOME));
  if (pops == (CHROMOSOME *) NULL)
    {
      mvprintw (POPTABLE_ROW + 2, POPTABLE_COL,
		"Cannot get enough memory - goodbye");
      move (last_line - 1, 0);
      refresh ();
      exit (1);
    }

  for (pop = 0; pop < numpops; pop++)
    total_abs_fitness[pop] = 0.0;
  for (chromo = 0; chromo < numpops * numchromo; chromo += 1)
    {				/* Initialise the */
      create_chromo (&pops[chromo]);	/* pops */
      init_genotype (pops[chromo].gdp);
      fix_for_sim_groups (pops[chromo].gdp);
      pops[chromo].fitness =
	init_evaluate (pops[chromo].gdp->genotype, 0,
		       &pops[chromo], 0,
		       chromo / numchromo);
      /* total_abs_fitness[n] is total fitness of population n */
      total_abs_fitness[chromo / numchromo] += fabs (pops[chromo].fitness);
    }

  /*  if (!spatialchosen) {
      for(pop=0; pop<numpops; pop+=1) {
      qsort((char *)((int)pops+(pop*numchromo*sizeof(CHROMOSOME))),
      numchromo, sizeof(CHROMOSOME), compchromo);
      bestinpop[pop]=0;
      }
      }*/

  /*  else { */
  for (i = 0; i < numpop; i++)
    bestinpop[i] = 0;
  for (pop = 0; pop < numpops; pop += 1)
    {
      for (chromo = 0; chromo < numchromo; chromo++)
	{
	  if (pops[pop * numchromo + chromo].fitness >
	      pops[pop * numchromo + bestinpop[pop]].fitness)
	    bestinpop[pop] = chromo;
	}
    }

  if (output)
    analyse_initial_pop (pops);
  return (pops);
}

void 
analyse_initial_pop (pops)
     CHROMOSOME *pops;
{
  int chromo, clashes, toocloses;

  clashes = toocloses = 0;

  for (chromo = 0; chromo < numpop * numchromo; chromo++)
    {
      clashes += pops[chromo].clashes;
      toocloses += pops[chromo].toocloses;
    }
  fprintf (output, "Initial pops averaged %g clases, %g toocloses\n",
	   (double) ((double) clashes / (double) (numpop * numchromo)),
	   (double) (((double) toocloses / (double) (numpop * numchromo))));
}

/***********************************************************************/
/*  This is used by various forms of greedy initialisation. Given an   */
/* event e, the current timetable so far, the current nogo-record so   */
/* far, find what the _new_ set of nogos will be if e is put into slot */
/* s. That is, how restrictive on the placements of other events will  */
/* it be to put e in slot s?  In particular, how _extra_ restrictive   */
/* will it be? There are various ways of looking at this; eg: we can   */
/* consider only extra restrictions on events _yet to be timetabled_,  */
/* or we can consider extra restrictions on all other events. The      */
/* latter way, for example, may produce timetables which are more      */
/* easily changed by further search. Should allow for either of these  */
/* to be chosen.                                                       */
/***********************************************************************/

void 
nogos_c (int e, int *tt, int s)
{
  int i, n, cl, tc;
  _MODULE *m;
  LL_EDGES *lle;
  m = a_modules[e];

  /* initialise the new nogos structure */
  nng->n = 0;

  /* If e goes into slot s, then e's edge-neighbours will be affected */
  /* See just how affected they are */

  for (lle = m->edges; lle; lle = lle->next)
    {
      if (lle->e->mod1->id == e)
	n = lle->e->mod2->id;	/* Retrieve e's */
      else
	n = lle->e->mod1->id;	/* neighbour, n */
      /* Carry on, only if n has not already been timetabled */
      if (tt[n] == -1)
	{
	  for (i = 0; i < numslots; i++)
	    if ((NOGO_E (n, i) == 0) &&
		(NOGO_C (n, i) == 0))
	      {			/* If, so far, n can be in slot i */
		/* see if it now can't be */
		check_edge_violation (e, n, s, i, &cl, &tc);
		if (cl > 0)
		  {		/* If bad, then make a new nogo */
		    nng->nogos[nng->n]->event = n;
		    nng->nogos[nng->n++]->slot = i;
		  }
	      }
	}
    }
}
void 
nogos_tc (int e, int *tt, int s)
{
  int i, n, cl, tc, viol;
  _MODULE *m;
  LL_EDGES *lle;
  LL_JUXTS *llj;
  m = a_modules[e];

  /* initialise the new nogos structure */
  nng->n = 0;

  /* If e goes into slot s, then e's edge-neighbours will be affected */
  /* See just how affected they are */

  for (lle = m->edges; lle; lle = lle->next)
    {
      if (lle->e->mod1->id == e)
	n = lle->e->mod2->id;	/* Retrieve e's */
      else
	n = lle->e->mod1->id;	/* neighbour, n */

      /* Carry on, only if n has not already been timetabled */
      if (tt[n] == -1)
	{
	  for (i = 0; i < numslots; i++)
	    if ((NOGO_E (n, i) == 0) &&	/* If, so far, n can be in slot i */
		(NOGO_TC (n, i) == 0))
	      {			/* see if it now can't be */
		check_edge_violation (e, n, s, i, &cl, &tc);
		if (tc > 0)
		  {		/* If bad, then make a new nogo */
		    nng->nogos[nng->n]->event = n;
		    nng->nogos[nng->n++]->slot = i;
		  }
	      }
	}
    }
  /* Now do the same for e's juxt neighbours */

  for (llj = m->juxts; llj; llj = llj->next)
    {
      if (llj->j->a->id == e)
	n = llj->j->b->id;	/* Retrieve e's */
      else
	n = llj->j->a->id;	/* neighbour, n */

      /* Carry on, only if n has not already been timetabled */
      if (tt[n] == -1)
	for (i = 0; i < numslots; i++)
	  if ((NOGO_E (n, i) == 0) &&	/* If, so far, n can be in slot i */
	      (NOGO_TC (n, i) == 0) &&
	      (in_newnogos (n, i, nng) == 0))
	    {			/* see if it now can't be */
	      if (e < n)
		check_juxt_constraint (llj->j, s, i, &viol);
	      else
		check_juxt_constraint (llj->j, i, s, &viol);
	      if (viol > 0)
		{		/* If bad, then make a new nogo */
		  nng->nogos[nng->n]->event = n;
		  nng->nogos[nng->n++]->slot = i;
		}
	    }
    }
}

void
nogos_s (int e, int *tt, int s, int *fills)
{
  int i;
  _MODULE *m;
  m = a_modules[e];

  /* initialise the new nogos structure */
  nng->n = 0;

  if (slotseatconstraints)
    for (i = 0; i < nummods; i++)
      if ((i != e) && (tt[i] == -1))
	{			/* Not e itself, and untimetabled, */
	  if (a_modules[i]->sim == 0)
	    {			/* and not in a simgroup */

	      /* check if it can go in slot s without overload */
	      if ((NOGO_S (i, s) == 0) &&
		  (in_newnogos (i, s, nng) == 0))
		{
		  if ((fills[s] + a_modules[i]->size) > slots[s]->sscap)
		    {
		      /* Add a new nogo for this */
		      nng->nogos[nng->n]->event = i;
		      nng->nogos[nng->n++]->slot = s;
		    }
		}
	    }
	  else
	    /* We add to fills the whole size */ if ((NOGO_S (i, s) == 0) &&
						     (NOGO_E (i, s) == 0) &&
					     (in_newnogos (i, s, nng) == 0))
	    {
	      if ((fills[s] + simgroups[a_modules[i]->simgroup]->size)
		  > slots[s]->sscap)
		{
		  /* Add a new nogo for this */
		  nng->nogos[nng->n]->event = i;
		  nng->nogos[nng->n++]->slot = s;
		}
	    }
	}
}


int 
in_newnogos (int n, int s, _NEWNOGOS * nng)
{
  int i;
  for (i = 0; i < nng->n; i++)
    if ((nng->nogos[i]->event == n) &&
	(nng->nogos[i]->slot == s))
      return (1);
  return (0);
}

_NOGO *
newnogoalloc ()
{
  _NOGO *temp;

  if ((temp = (_NOGO *) malloc (sizeof (_NOGO))) != NULL)
    return temp;
  else
    {
      printf ("Trouble: Couldn't find enough space for storing nogos\n");
      exit (0);
    }
}

_NEWNOGOS *
newnewnogoalloc ()
{
  _NEWNOGOS *temp;

  if ((temp = (_NEWNOGOS *) malloc (sizeof (_NEWNOGOS))) != NULL)
    return temp;
  else
    {
      printf ("Trouble: Couldn't find enough space for storing newnogos\n");
      exit (0);
    }
}


