/************************************************************************
 *                                                                      *
 *               GATT - Genetic Algorithm Time-Tabler                   *
 *                                                                      *
 *                By Dave Corne                                         *
 *                   Peter Ross (peter@aisb.ed.ac.uk),                  *
 *                   Emma Collingwood (emmac@aisb.ed.ac.uk)             *
 *                                                                      *
 ************************************************************************/

/* 
   Copyright (C) 1996, Dave Corne, Peter Ross and Emma Collingwood

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. 
*/


/**********************************************************************
 *  Filename: input.c
 *
 *  Functions to deal with all data supplied in the input parameter
 *  files to set up GATT for running. 
 *
 **********************************************************************/ 



#include "gatt.h"		/* Include struct defn's and defines */
#include <stdio.h>

#define GRAPH(a,b) graph[a*nummods+b]
#define CCM_SCORE(a,b) ccm_score[a*nummods+b]
extern FILE *output;
int *bests, nbests;
int nsimgroups = 0;
int nsubprobs = 0;
void sort_mod_array ();
void initialise_ccm ();

extern int maxof();

struct ms
{
  int id, size;
};
struct ms msizes[MAXMODS];

extern div_t div ();

int n_cce = 0;

struct binary_constraint
{
  EDGE *e;
  JUXT *j;
};

static int modcompare ();

/* ADDED TO CHECK TALLY BETWEEN DAYS ENTERED AND
 * DAYS IN SLOTS */

extern int maxdaysofar;

/********************/

extern int allmodsexcludenum;
extern int numchromo;
extern int numpops;
extern double (*evaluate) ();
extern double *dchecks;
extern double *devals;
extern double efm_pressure;
extern int debug;
extern int chromoformat;
extern int geno_size;
extern int *bestinpop;
extern int numjuxts;
extern int numslotsgiven;
extern int numstudentsgiven;
extern int numagentsgiven;
extern int nummodsgiven;
extern int numroomsgiven;
extern int numrooms;
extern int numslots;
extern int numagents;
extern int nummods;
extern int numedges;
extern int default_agentmaxmods;
extern int default_agentmaxhours;
extern int default_agentminmods;
extern int default_agentminhours;
extern int default_roommodcap;
extern int default_roomstudcap;
extern int default_modlength;
extern int maxpref;
extern int roomsmatter;
extern int agentsmatter;
extern int *modprefs;
extern int spatialchosen;
extern int *evals;
extern int *checks;
extern int numpop;
extern int vdm;
extern int efm;
extern int slotseatconstraints;
extern int sloteventconstraints;
extern int slotseatmax;
extern int sloteventmax;
extern double dsm_rate;
extern int efms;
extern int sefms;
extern int crosses;
extern int scrosses;

int *graph, *ccm_score;
int *tnscs, *tnrcs, *tnacs;

char *agentslotexcs,		/* array indicating agents' slot exclusions */
 *modagentexcs,			/* array indicating agents' module exclusions */
 *modslotexcs,			/* array indicating modules' slot exclusions */
 *modroomexcs;			/* array indicating modules' room exclusions */

int in_studs ();
int nextday();
int nextyear();
int nextmonth();
int isleapyear();
void addnewcce ();
void addtosimgroups ();
void addtosubprobs ();
void generate_speclists ();
void buildmodarray ();
void buildedgearray ();
void buildjuxtarray ();
void fillindates ();
void fillinslotseatdetails ();
void fillinsloteventdetails ();
void spread_restrictions_on_sim_members ();
void find_simgroup_sizes ();
void initialise_tnchange_sizes ();
_MODULE **modarrayalloc ();
EDGE **edgearrayalloc ();
JUXT **juxtarrayalloc ();
extern LL_SPEC *newllspecalloc ();
extern EDGE *addnewedge ();
extern LL_EDGES *newlledgealloc ();
extern LL_JUXTS *newlljuxtalloc ();
extern void ss_one_reproduction ();
extern void mix_ops_ss_one_reproduction ();
extern void (*reproduction) ();

int compmodedges ();

void 
dealwithinput ()
{
  int i, j, k, slotid, dayid, rel, elm, dsm, gap, exc;
  int trouble = 0, _room_ = 0, _tutor_ = 0;
  _MODULE *m;
  _STUDENT *s;
  LL_MODS *t, *u, *llm;
  LL_SPEC *lls, *llr, *lla;
  JUXT *ju;
  EDGE *ed;
  
  
  

    /* First, check that various input numbers tally with what's expected */
  if (numslotsgiven != numslots)
    {
      printf ("Trouble: number of slots entered is %d, but %d were expected\n",
	      numslotsgiven, numslots);
      trouble = 1;
    }
  if (numagentsgiven != numagents)
    {
      printf ("Trouble: number of agents entered is %d, but %d were expected\n",
	      numagentsgiven, numagents);
      trouble = 1;
    }
  if (nummodsgiven != nummods)
    {
      printf ("Trouble: number of modules entered is %d, but %d were expected\n",
	      nummodsgiven, nummods);
      trouble = 1;
    }
  if (numroomsgiven != numrooms)
    {
      printf ("Trouble: number of rooms entered is %d, but %d were expected\n",
	      numroomsgiven, numrooms);
      trouble = 1;
    }

  /* and check numbers dont exceed the limits */
  if (numslots > MAXSLOTS)
    {
      printf ("Trouble: number of slots entered exceeds maximum of %d\n",
	      MAXSLOTS);
      trouble = 1;
    }

  if (nummods > MAXMODS)
    {
      printf ("Trouble: number of modules entered exceeds maximum of %d\n",
	      MAXMODS);
      trouble = 1;
    }

  if (numstudents > MAXSTUDENTS)
    {
      printf ("Trouble: number of students exceeds maximum of %d\n", MAXSTUDENTS);
      trouble = 1;
    }

  if (numdays > MAXDAYS)
    {
      printf ("Trouble: number of days exceeds maximum of %d\n", MAXDAYS);
      trouble = 1;
    }

  
  if ((numdays-1) != maxdaysofar){
    printf("Trouble: number of days entered does not tally \n");
    trouble = 1;
  }			


  if (numrooms > MAXROOMS)
    {
      printf ("Trouble: number of rooms exceeds maximum of %d\n", MAXROOMS);
      trouble = 1;
    }

  if (numagents > MAXAGENTS)
    {
      printf ("Trouble: number of agents exceeds maximum of %d\n", MAXAGENTS);
      trouble = 1;
    }


      


  if (trouble == 1)
    exit (0);


  fillindates ();		/* Fill in date structures */
bests = (int *) malloc (numpop * numchromo * sizeof (int));
free(bests);

/* Now, build the modules array */
  a_modules = modarrayalloc ();	/* Get memory for the pointers */
  buildmodarray (t_modules);	/* Now travel through the module */
  /* tree slotting them into the array */

  /* Put in the given default values for agents */
  for (i = 0; i < numagents; i++)
    {
      if (agents[i]->maxmods < 0)
	agents[i]->maxmods = default_agentmaxmods;
      if (agents[i]->maxhours < 0)
	agents[i]->maxhours = default_agentmaxhours;
      if (agents[i]->minmods < 0)
	agents[i]->minmods = default_agentminmods;
      if (agents[i]->minhours < 0)
	agents[i]->minhours = default_agentminhours;
    }
  /* Put in the given default values for room capacities */
  for (i = 0; i < numrooms; i++)
    {
      if (rooms[i]->modcap < 0)
	rooms[i]->modcap = default_roommodcap;
      if (rooms[i]->studcap < 0)
	rooms[i]->studcap = default_roomstudcap;
    }

  /* Put in the given default values for module lengths */
  for (m = modules; m; m = m->next)
    if (m->len < 0)
      m->len = default_modlength;

  /* Construct the room capacity data for use in evaluation */
  for (i = 0; i < numrooms; i++)/* First put in capacities */
    for (j = 0; j < numslots; j++)
      {
	roomcaps[i][j][0] = rooms[i]->modcap;
	roomcaps[i][j][1] = rooms[i]->studcap;
      }
  /* Now put in the exceptions */
  for (i = 0; i < numrooms; i++)
    {
      for (j = 0; j < rooms[i]->nummexceptions; j++)
	roomcaps[i][rooms[i]->mexceptions[j]->slot][0] =
	  rooms[i]->mexceptions[j]->cap;
      for (j = 0; j < rooms[i]->numsexceptions; j++)
	roomcaps[i][rooms[i]->sexceptions[j]->slot][1] =
	  rooms[i]->sexceptions[j]->cap;
    }
  /* Put agent specifications into module structures */
  for (i = 0; i < numagents; i++)
    {
      for (j = 0; j < agents[i]->nummustdo; j++)
	{
	  a_modules[agents[i]->mustdolist[j]]->agentspec = i;
	}
    }

  /* For each module without a specified agent, see who's excluded */
  if (numagents > 0)
    {
      for (m = modules; m; m = m->next)
	{
	  if (m->agentspec < 0)
	    for (j = 0; j < numagents; j++)
	      {
		exc = 1;
		for (k = 0; k < agents[j]->numcando; k++)
		  {
		    if (agents[j]->cando[k]->mod == m->id)
		      {
			exc = 0;
			break;
		      }
		  }
		if (exc)
		  {
		    if (m->numexcagent == 0)
		      {
			m->s_exagents = newllspecalloc ();
			m->s_exagents->forced = 1;
			m->s_exagents->priority = 10;
			m->s_exagents->id = j; /* CHANGED TO AGENT */
			m->s_exagents->next = NULL;
			m->exagents = m->s_exagents;
			m->numexcagent++;
		      }
		    else
		      {
			m->p_exagents = m->exagents;
			m->exagents = newllspecalloc ();
			m->exagents->forced = 1;
			m->exagents->priority = 10;
			m->exagents->id = j; /* CHANGED TO AGENT */
			m->exagents->next = m->p_exagents;
			m->numexcagent++;
		      }
		  }
	      }
	}


      /* Work out maximum possible preference for agents' module choices */
    }
  
bests = (int *) malloc (numpop * numchromo * sizeof (int));  
free(bests);


  /* Build exclusion array for mods excluding slots */

  modslotexcs = (char *) malloc (numslots * nummods + 1);
  if (modslotexcs == NULL)
    {
      printf ("Failed to allocate memory\n");
      exit (1);
    }
  for (i = 0; i < nummods; i++)
    for (j = 0; j < numslots; j++)
      modslotexcs[nummods * j + i] = 0;

  for (m = modules; m; m = m->next){
    for (lls = m->exslots; lls; lls = lls->next)
      if (lls->forced)
	modslotexcs[nummods * (lls->id) + (m->id)] = 1;	/* excluded */
      else
	modslotexcs[nummods * (lls->id) + (m->id)] = 2;	/* avoided */
  }

  for (m = modules; m; m = m->next)
    for (i = 0; i < allmodsexcludenum; i++)
      modslotexcs[nummods * allmodsexclude[i] + (m->id)] = 1;
bests = (int *) malloc (numpop * numchromo * sizeof (int));
free(bests);

/* Build exclusion array for mods excluding rooms */

  modroomexcs = (char *) malloc (numrooms * nummods + 1);
  if (modroomexcs == NULL)
    {
      printf ("Failed to allocate memory\n");
      exit (1);
    }
  for (m = modules; m; m = m->next)
    for (j = 0; j < numrooms; j++)
      modroomexcs[nummods * j + (m->id)] = 0;
  bests = (int *) malloc (numpop * numchromo * sizeof (int));
  free(bests);
  
  for (m = modules; m; m = m->next)
    for (llr = m->exrooms; llr; llr = llr->next)
      if (llr->forced)
	modroomexcs[nummods * (llr->id) + m->id] = 1;	/* excluded */
      else
	modroomexcs[nummods * (llr->id) + m->id] = 2;	/* avoided */

  /* Build exclusion array for mods excluding agents */

  modagentexcs = (char *) malloc (numagents * nummods);
  if (modagentexcs == NULL)
    {
      printf ("Failed to allocate memory\n");
      exit (1);
    }

  for (i = 0; i < nummods; i++)
    for (j = 0; j < numagents; j++)
      modagentexcs[nummods * j + i] = 0;

  for (m = modules; m; m = m->next)
    for (lla = m->exagents; lla; lla = lla->next)
      if (lla->forced)
	modagentexcs[nummods * (lla->id) + m->id] = 1;	/* excluded */
      else
	modagentexcs[nummods * (lla->id) + m->id] = 2;	/* avoided */

  /* Build exclusion array for agents excluding slots */
  agentslotexcs = (char *) malloc (numslots * numagents);
  if (agentslotexcs == NULL)
    {
      printf ("Failed to allocate memory\n");
      exit (1);
    }
 bests = (int *) malloc (numpop * numchromo * sizeof (int));
  free(bests);
 
  for (i = 0; i < numagents; i++)
    for (j = 0; j < numslots; j++)
      agentslotexcs[numagents * j + i] = 0;

  for (i = 0; i < numagents; i++)
    {
      for (j = 0; j < agents[i]->numexcslots; j++)
	agentslotexcs[numagents * (agents[i]->excslots[j]) + i] = 1;	/* excluded */
      for (j = 0; j < agents[i]->numavslots; j++)
	agentslotexcs[numagents * (agents[i]->avslots[j]->mod) + i] = 2;	/* avoided */
    }

  
  /* Add to module structures the list of students taking it */


  /* Add to module structures pointers to the edge constraints they're in */

  /* put real size of modules in */

  for (i = 0; i < nummods; i++)
    a_modules[i]->size += a_modules[i]->numtakers;

  bests = (int *) malloc (numpop * numchromo * sizeof (int));

  /* DEBUGGING  */

   if (debug)
    {

      printf ("Numslots is %d, numslotsgiven is %d\n", numslots, numslotsgiven);

      for (i = 0; i < numslotsgiven; i++)
	{
	  printf ("slot %d is on day %d (%d) at time %d\n",
		  i, slots[i]->day, slots[i]->aday, slots[i]->time);

	}

      printf ("Numdays is %d\n", numdays);

      if (numdays)
	printf ("Date of first day is %d/%d/%d \n", dates[0]->day,
		dates[0]->month,
		dates[0]->year);
      fflush (stdout);
      for (s = students; s; s = s->next)
	{
	  printf (" Student %s takes : ", s->name);
	  for (llm = s->mods; llm; llm = llm->next)
	    {
	      printf ("%s ", llm->mod->name);
	    }
	  printf ("\n");
	}
      fflush (stdout);
      for (i = 0; i < numagentsgiven; i++)
	{
	  printf (" Agent name %s teaches max %d mods, %d hours, min %d mods, %d hours\n", agents[i]->name, agents[i]->maxmods, agents[i]->maxhours,
		  agents[i]->minmods, agents[i]->minhours);
	  printf (" Agent name %s must do %d mods, which are: \n",
		  agents[i]->name, agents[i]->nummustdo);
	  for (j = 0; j < agents[i]->nummustdo; j++)
	    {
	      printf ("%s ", a_modules[agents[i]->mustdolist[j]]->name);
	    }
	  printf (" \n and can do %d modules, which are \n",
		  agents[i]->numcando);
	  for (j = 0; j < agents[i]->numcando; j++)
	    {
	      printf ("%s %d ", a_modules[agents[i]->cando[j]->mod]->name,
		      agents[i]->cando[j]->des);
	    }
          printf("\n");
          
	}

      for (i = 0; i < numagentsgiven; i++)
	{
	  printf (" Agent %s avoids slots : \n", agents[i]->name);
	  for (j = 0; j < agents[i]->numavslots; j++)
	    {
	      printf ("%d %d    ", agents[i]->avslots[j]->mod,
		      agents[i]->avslots[j]->des);
	    } printf ("\n");
	}
      for (i = 0; i < numagentsgiven; i++)
	{
	  printf (" Agent %s excludes slots : \n", agents[i]->name);
	  for (j = 0; j < agents[i]->numexcslots; j++)
	    {
	      printf ("%d  ", agents[i]->excslots[j]);

	    } printf ("\n");
	}
      for (m = modules; m; m = m->next)
	{
	  printf (" Module %s excludes slots : \n", m->name);
	  for (lls = m->exslots; lls; lls = lls->next)
	    {
	      if (lls->forced)
		printf ("%d  ", lls->id);
	    } printf ("\n");
	}
      for (m = modules; m; m = m->next)
	{
	  printf (" Module %s avoids slots : \n", m->name);
	  for (lls = m->exslots; lls; lls = lls->next)
	    if (!(lls->forced))
	      printf ("%d %d     ", lls->id,
		      lls->priority);
	} printf ("\n");

      for (m = modules; m; m = m->next)
	if (m->slotspec > -1)
	  printf (" Module %s specifies slot %d : \n",
		  m->name, m->slotspec);

      for (ju = juxts; ju; ju = ju->next)
	{
	  rel = ju->rel;
	  elm = ju->elm;
	  dsm = ju->dsm;
	  gap = ju->gap;

	  printf (" %s ", ju->a->name);
	  if ((rel != 3) && (rel != 4))
	    {
	      if (elm == 0)
		printf (" EXACTLY ");
	      if (elm == 1)
		printf (" AT LEAST ");
	      if (elm == 2)
		printf (" AT MOST ");
	      printf (" %d ", gap);
	      if (dsm == 0)
		printf (" DAYS");
	      if (dsm == 1)
		printf (" MINUTES ");
	    }
	  if (rel == 0)
	    printf (" BEFORE ");
	  if (rel == 1)
	    printf (" AFTER ");
	  if (rel == 2)
	    printf (" BETWEEN ");
	  if (rel == 3)
	    printf (" SAME TIME ");
	  if (rel == 4)
	    printf (" SAME DAY ");
	  printf (" %s\n ", ju->b->name);
	}

      for (i = 0; i < numroomsgiven; i++)
	{
	  printf ("Room %s has student capacity %d and module capacity %d\n",
		  rooms[i]->name, rooms[i]->studcap, rooms[i]->modcap);
	  printf (" it has %d student capacity exceptions, which are : \n",
		  rooms[i]->numsexceptions);
	  for (j = 0; j < rooms[i]->numsexceptions; j++)
	    {
	      printf ("Slot %d scap %d, ", rooms[i]->sexceptions[j]->slot,
		      rooms[i]->sexceptions[j]->cap);
	    }
	  printf ("\n");
	  printf (" it has %d module capacity exceptions, which are : \n",
		  rooms[i]->nummexceptions);
	  for (j = 0; j < rooms[i]->nummexceptions; j++)
	    {
	      printf ("Slot %d mcap %d, ", rooms[i]->mexceptions[j]->slot,
		      rooms[i]->mexceptions[j]->cap);
	    }
	  printf ("\n");
	  printf (" it has %d excluded slots, which are : \n",
		  rooms[i]->numexcslots);
	  for (j = 0; j < rooms[i]->numexcslots; j++)
	    {
	      printf ("Slot %d, ", rooms[i]->excslots[j]);
	    }
	  printf ("\n");
	}
      exit (0);
    }
bests = (int *) malloc (numpop * numchromo * sizeof (int));
  /* Now, decide what chromosome format to use. If there is more than one */
  /* day in the timetable, and days are not explicitly disallowed in the  */
  /* chromosome, then include days */

  /* If rooms matter, then include rooms (if there are more than one) */
  if ((roomsmatter == 1) && (numrooms > 0))
    _room_ = 1;

  /* If there are teachers, and more than one, include them */
  if ((agentsmatter== 1) && (numagents > 0))
    _tutor_ = 1;


  /* So now choose the chromosome format */
bests = (int *) malloc (numpop * numchromo * sizeof (int));
  switch (_room_ + 2 * _tutor_)
    {
    case 0:
      chromoformat = _SLOT_;
      geno_size = nummods;
      break;
    case 1:
      chromoformat = _SLOT_ROOM_;
      geno_size = 2 * nummods;
      break;
    case 2:
      chromoformat = _SLOT_TUTOR_;
      geno_size = 2 * nummods;
      break;
    case 3:
      chromoformat = _SLOT_ROOM_TUTOR_;
      geno_size = 3 * nummods;
      break;
    }

  /* Now initialise the stuff to do with keeping track of distinct bests */

  nbests = 0;
  bests = (int *) malloc (numpop * numchromo * sizeof (int));
  if (bests == (int *) NULL)
    {
      printf ("Failed to allocate memory\n");
      exit (1);
    }

  /* Next, sort out the arrays which give days and ids of slots etc ... */

  slotid = 0;
  dayid = 0;
  for (i = 0; i < numslots; i++)
    {
      if (!(slots[i]->day == dayid))
	{
	  slotid = 0;
	  dayid = slots[i]->day;
	  slots[i]->id = slotid++;
	}
      else
	{
	  slots[i]->id = slotid++;
	}
    }

  /* Now, sort out the student or `schedule apart' data into  */
  /* edge constraints */

  for (s = students; s; s = s->next)
    {				/* For each student */
      for (t = s->mods; t->next; t = t->next)
	{
	  for (u = t->next; u; u = u->next)
	    {
	      /* check no corrupt data, ie same student taking an exam more than once ! */
	      if (strcmp (t->mod->name, u->mod->name))
		t_edges = addnewedge (t_edges, t->mod, u->mod, 1);
	    }
	}
    }

  a_edges = edgearrayalloc ();	/* Now all edges known, can build array */
  buildedgearray (t_edges);

  a_juxts = juxtarrayalloc ();	/* Do the same for juxts here too */
  buildjuxtarray (t_juxts);

  if (slotseatconstraints)
    fillinslotseatdetails ();
  if (sloteventconstraints)
    fillinsloteventdetails ();

  /* Sort out selection */

  /*    if(spatialchosen==1) */ /* Spatial selection has been chosen */
  /*      reproduction = ss_one_reproduction; */

  if (reproduction == mix_ops_ss_one_reproduction)
    efms = sefms = crosses = scrosses = 0;

  evals = (int *) malloc (numpop * sizeof (int));
  checks = (int *) malloc (numpop * sizeof (int));

  schanges = (int *) malloc (nummods * sizeof (int));
  schanged = (char *) malloc (nummods);
  if (roomsmatter)
    {
      rchanges = (int *) malloc (nummods * sizeof (int));
      rchanged = (char *) malloc (nummods);
    }
  if (agentsmatter)
    {
      achanges = (int *) malloc (nummods * sizeof (int));
      achanged = (char *) malloc (nummods);
    }
  if (vdm || efm)
    {
      hotslots = (int *) malloc (nummods * sizeof (int));
      if (roomsmatter)
	hotrooms = (int *) malloc (nummods * sizeof (int));
      if (agentsmatter)
	hotagents = (int *) malloc (nummods * sizeof (int));
    }
  if (efm)
    {
      eslots = (int *) malloc (numslots * sizeof (int));
      if (roomsmatter)
	erooms = (int *) malloc (numrooms * sizeof (int));
      if (agentsmatter)
	eagents = (int *) malloc (numagents * sizeof (int));
    }

  if (roomsmatter)
    {
      roomeloads = (int *) malloc (numrooms * numslots * sizeof (int));
      roomsloads = (int *) malloc (numrooms * numslots * sizeof (int));
    }
  dchecks = (double *) malloc (numpop * sizeof (double));
  devals = (double *) malloc (numpop * sizeof (double));

  bestinpop = (int *) malloc (numpop * sizeof (int));


  /* Now sort out the sim groups */

  /* Find them */
  for (ju = juxts; ju; ju = ju->next)
    if (ju->rel == 3)
      addtosimgroups (ju);

  /* tell me about them */
  /*
    for(i=0;i<nsimgroups;i++)
      {printf("simgroup %d : ",i);
     for(j=0;j<simgroups[i]->num;j++)
        printf(" %s ", a_modules[simgroups[i]->sims[j]]->name);
     printf("\n");
   }
    exit(0);
  */
  /* Tell the modules about them */

  for (i = 0; i < nsimgroups; i++)
    {
      a_modules[simgroups[i]->sims[0]]->sim = 1;	/* base of group */
      a_modules[simgroups[i]->sims[0]]->simgroup = i;	/* this group */
      for (j = 1; j < simgroups[i]->num; j++)
	{
	  a_modules[simgroups[i]->sims[j]]->sim = 2;	/* in a group */
	  a_modules[simgroups[i]->sims[j]]->simgroup = i;	/* ... this one */
	}
    }
  /* Make sure base of a simgroup has all the restrictions on it that */
  /* the group's members have */

  spread_restrictions_on_sim_members ();
  find_simgroup_sizes ();

  /* Sort out all the allowed clashes. If some pair is set for */
  /* being concurrent, but students take both, then make ns 0  */

  for (ed = edges; ed; ed = ed->next)
    {
      if ((ed->mod1->sim > 0)	/* It's in a sim group */
	  && (ed->mod2->sim > 0)/* So is this */
	  && (ed->mod1->simgroup == ed->mod2->simgroup))	/* in the same group */

	{
	  ed->ns = 0;		/* Then zap consideration of this edge */
	  /*   printf("an edge : %s %s has been zapped\n",ed->mod1->name,
                      ed->mod2->name); */
	}
      if
	((ed->mod1->slotspec > -1) && (ed->mod2->slotspec > -1))
	ed->ns = 0;
    }

  generate_speclists ();

  /* Build the edge graph, and do some graphic things */

  graph = (int *) calloc (nummods * nummods, sizeof (int));
  ccm_score = (int *) calloc (nummods * nummods, sizeof (int));

  for (ed = edges; ed; ed = ed->next)
    GRAPH (ed->mod1->id, ed->mod2->id) =
      GRAPH (ed->mod2->id, ed->mod1->id) = 1;

  /* Find set relations between  non edges */

  /*  inx = (char *)calloc(nummods,1);
   iny = (char *)calloc(nummods,1);
 
   for(i=0;i<nummods-1;i++)
    for(j=i+1;j<nummods;j++)
      if(GRAPH(i,j)==0)
        { if (a_modules[i]->numedges>=a_modules[j]->numedges)
              {x = i; y = j; }
	      else {x=j; y=i;}
              for(k=0;k<nummods;k++)
                inx[k]=iny[k]=0;
           for(e=a_modules[y]->edges;e;e=e->next)
            if(e->e->mod1->id==y) iny[e->e->mod2->id]=1;
            else iny[e->e->mod1->id]=1;
 
             for(e=a_modules[x]->edges;e;e=e->next)
            if(e->e->mod1->id==x) inx[e->e->mod2->id]=1;
            else inx[e->e->mod1->id]=1;
 
          shares = ynox = xnoy=0;
          for(k=0;k<nummods;k++)
             if(iny[k]==1)
               {if (inx[k]==0) ynox++;
                else shares++;
	       }
             else { if (inx[k]==1) xnoy++;}
	 
         if(shares==ynox)
         printf(" %s %d  %s  %d \n",
                  a_modules[y]->name, a_modules[y]->numedges,
                  a_modules[x]->name, a_modules[x]->numedges);
	     } */
  /* exit(0); */

  /*  initialise_ccm(); */
  initialise_tnchange_sizes ();
}

void 
sort_mod_array ()
{
  int i;

  for (i = 0; i < nummods; i++)
    {
      msizes[i].id = i;
      msizes[i].size = a_modules[i]->size;
    }
  qsort ((char *) msizes, nummods, sizeof (struct ms), modcompare);
}



static int 
modcompare (i, j)
     struct ms *i, *j;
{
  return (j->size - i->size);
}


void 
spread_restrictions_on_sim_members ()
{
  int i, j, specfound, specs[MAXSLOTS], nspecsfound;


  for (i = 0; i < nsimgroups; i++)
    {
      specfound = nspecsfound = 0;
      for (j = 0; j < simgroups[i]->num; j++)
	if (a_modules[simgroups[i]->sims[j]]->slotspec > -1)
	  {
	    specfound = 1;
	    specs[nspecsfound++] = a_modules[simgroups[i]->sims[j]]->slotspec;
	  }
      /* we have found all specs in this group */
      /* but see if there is a problem */
      for (j = 0; j < nspecsfound; j++)
	if (specs[0] != specs[j])
	  {
	    printf ("specifications and simgroups don't match !!");
	    exit (0);
	  }
    }
}

void 
find_simgroup_sizes ()
{
  int i, j, n;
  short int studs[MAXSTUDENTS];
  _MODULE *m;
  LL_STUDS *lls;
  for (i = 0; i < nsimgroups; i++)
    {
      n = 0;
      simgroups[i]->size = 0;
      simgroups[i]->total = 0;
      for (j = 0; j < simgroups[i]->num; j++)
	{
	  m = a_modules[simgroups[i]->sims[j]];
	  simgroups[i]->total += m->size;
	  for (lls = m->studs; lls; lls = lls->next)
	    if (in_studs (studs, n, lls->st->id) == 0)
	      studs[n++] = lls->st->id;
	}
      simgroups[i]->size = n;
    }
}

int 
in_studs (int *studs, int n, int id)
{
  int i;
  for (i = 0; i < n; i++)
    if (studs[i] == id)
      return (1);
  return (0);
}

void 
addtosubprobs (bc, id)
     struct binary_constraint bc;
     int id;
{
  int i, k, a_group, b_group, a_found, b_found, a, b, tmp, goes_in;

  if (id == 0)			/* it's an edge */
    {
      a = bc.e->mod1->id;
      b = bc.e->mod2->id;
    }
  else				/* It's a juxt */
    {
      a = bc.j->a->id;
      b = bc.j->b->id;
    }
  /* See if this edge needs a new subprob */

  a_found = b_found = 0;

  for (i = 0; i < nsubprobs; i++)
    for (k = 0; k < subprobs[i]->num; k++)
      {
	if (subprobs[i]->mems[k] == a)
	  {
	    a_group = i;
	    a_found = 1;
	  }
	else if (subprobs[i]->mems[k] == b)
	  {
	    b_group = i;
	    b_found = 1;
	  }
	if (a_found && b_found)
	  break;
      }

  if (a_found && b_found && (a_group == b_group))
    return;

  if (a_found && b_found && (a_group != b_group))
    {				/* Two groups need to be merged */
      if (a_group > b_group)
	{
	  tmp = a_group;
	  a_group = b_group;
	  b_group = tmp;
	}

      /* Put everything from b into a */
      for (i = 0; i < subprobs[b_group]->num; i++)
	{
	  goes_in = 1;
	  for (k = 0; k < subprobs[a_group]->num; k++)
	    if (subprobs[b_group]->mems[i] ==
		subprobs[a_group]->mems[k])
	      {
		goes_in = 0;
		break;
	      }
	  if (goes_in == 1)
	    subprobs[a_group]->mems[subprobs[a_group]->num++] =
	      subprobs[b_group]->mems[i];
	}
      /* swap b with the last group in the list, and */
      /* decrement the number of subprobs */
      subprobs[b_group] = subprobs[--nsubprobs];
    }
  else if (a_found)
    {				/* a is in a_group, so we need to add b to this group */
      subprobs[a_group]->mems[subprobs[a_group]->num++] = b;
    }
  else if (b_found)
    {				/* b is in b_group, so we need to add a to this group */
      subprobs[b_group]->mems[subprobs[b_group]->num++] = a;
    }
  else
    {
      /* a and b start a whole new subprob */
      subprobs[nsubprobs] =
	(struct subprob *) malloc (sizeof (struct subprob));
      subprobs[nsubprobs]->num = 2;
      subprobs[nsubprobs]->mems[0] = a;
      subprobs[nsubprobs++]->mems[1] = b;
    }
}



void 
addtosimgroups (j)
     JUXT *j;
{
  int i, k, a_group, b_group, a_found, b_found, a, b, tmp, goes_in;

  a = j->a->id;
  b = j->b->id;

  /* See if this sim constraint needs a new sim group */

  a_found = b_found = 0;

  for (i = 0; i < nsimgroups; i++)
    for (k = 0; k < simgroups[i]->num; k++)
      {
	if (simgroups[i]->sims[k] == a)
	  {
	    a_group = i;
	    a_found = 1;
	  }
	else if (simgroups[i]->sims[k] == b)
	  {
	    b_group = i;
	    b_found = 1;
	  }
	if (a_found && b_found)
	  break;
      }

  if (a_found && b_found && (a_group == b_group))
    return;

  if (a_found && b_found && (a_group != b_group))
    {				/* Two groups need to be merged */
      if (a_group > b_group)
	{
	  tmp = a_group;
	  a_group = b_group;
	  b_group = tmp;
	}

      /* Put everything from b into a */
      for (i = 0; i < simgroups[b_group]->num; i++)
	{
	  goes_in = 1;
	  for (k = 0; k < simgroups[a_group]->num; k++)
	    if (simgroups[b_group]->sims[i] ==
		simgroups[a_group]->sims[k])
	      {
		goes_in = 0;
		break;
	      }
	  if (goes_in == 1)
	    simgroups[a_group]->sims[simgroups[a_group]->num++] =
	      simgroups[b_group]->sims[i];
	}
      /* swap b with the last group in the list, and */
      /* decrement the number of sim groups */

      simgroups[b_group] = simgroups[--nsimgroups];
    }
  else if (a_found)
    {				/* a is in a_group, so we need to add b to this group */
      simgroups[a_group]->sims[simgroups[a_group]->num++] = b;
    }
  else if (b_found)
    {				/* b is in b_group, so we need to add a to this group */
      simgroups[b_group]->sims[simgroups[b_group]->num++] = a;
    }
  else
    {
      /* a and b start a whole new simgroup */
      simgroups[nsimgroups] =
	(struct simgroup *) malloc (sizeof (struct simgroup));
      simgroups[nsimgroups]->num = 2;
      simgroups[nsimgroups]->sims[0] = a;
      simgroups[nsimgroups++]->sims[1] = b;
    }
}

/***************************************************************************/
/*  Generate the lists sunspecs, runspecs, aunspecs, which list the events */
/*  which do not have specified slots, rooms, or agents. This helps make   */
/*  sure that mutation and such things don't waste time; random choices of */
/*  events to mutate don't need to be checked, for example.                */
/**************************************************************************/
void 
generate_speclists ()
{
  int i, j;

  nsunspecs = nrunspecs = naunspecs = 0;

  sunspecs = (int *) malloc (nummods * sizeof (int));
  if (roomsmatter)
    runspecs = (int *) malloc (nummods * sizeof (int));
  if (agentsmatter)
    aunspecs = (int *) malloc (nummods * sizeof (int));

  for (i = 0; i < nummods; i++)
    {
      if (a_modules[i]->slotspec < 0)	/* If no specified slot */
	sunspecs[nsunspecs++] = i;
      if (roomsmatter)
	if (a_modules[i]->roomspec < 0)	/* If no specified room */
	  runspecs[nrunspecs++] = i;
      if (agentsmatter)
	if (a_modules[i]->agentspec < 0)
	  aunspecs[naunspecs++] = i;
    }

  /* Now find the number of possible alleles for each gene */

  slot_alleles = (int *) malloc (nummods * sizeof (int));
  if (roomsmatter)
    room_alleles = (int *) malloc (nummods * sizeof (int));
  if (agentsmatter)
    agent_alleles = (int *) malloc (nummods * sizeof (int));

 for (i = 0; i < nummods; i++)
    {
      if (a_modules[i]->slotspec >= 0)
	slot_alleles[i] = 1;
      else
	{
	  slot_alleles[i] = numslots;
	  for (j = 0; j < numslots; j++)
	    if (modslotexcs[nummods * j + i] == 1)
	      slot_alleles[i]--;
	}
      if (roomsmatter)
	if (a_modules[i]->roomspec >= 0)
	  room_alleles[i] = 1;
	else
	  {
	    room_alleles[i] = numrooms;
	    for (j = 0; j < numrooms; j++)
	      if (modroomexcs[nummods * j + i] == 1)
		room_alleles[i]--;
	  }
      if (agentsmatter)
	if (a_modules[i]->agentspec >= 0)
	  agent_alleles[i] = 1;
	else
	  {
	    agent_alleles[i] = numagents;
	    for (j = 0; j < numagents; j++)
	      if (modagentexcs[nummods * j + i] == 1)
		agent_alleles[i]--;
	  }
    }

}

int 
compmodedges (mod1, mod2)
     _MODULE *mod1, *mod2;
{
  if (mod1->numedges > mod2->numedges)
    return 1;
  else if (mod1->numedges == mod2->numedges)
    return 0;
  else
    return -1;
}

_MODULE **
modarrayalloc ()
{
  _MODULE **temp;

  if ((temp = (_MODULE **) calloc (nummods, sizeof (_MODULE *))) == NULL)
    {
      printf (" Trouble: Not enough memory for the module array -- bye \n");
      exit (0);
    }
  else
    return (temp);
}

EDGE **
edgearrayalloc ()
{
  EDGE **temp;

  if ((temp = (EDGE **) calloc (numedges, sizeof (EDGE *))) == NULL)
    {
      printf (" Trouble: Not enough memory for the edge array -- bye \n");
      exit (0);
    }
  else
    return (temp);
}

JUXT **
juxtarrayalloc ()
{
  JUXT **temp;

  if ((temp = (JUXT **) calloc (numjuxts, sizeof (JUXT *))) == NULL)
    {
      printf (" Trouble: Not enough memory for the juxt array -- bye \n");
      exit (0);
    }
  else
    return (temp);
}

/****************************************************************/
/* Another K&R Adaptation. Traverses the binary tree of modules */
/****************************************************************/

void 
buildmodarray (node)
     _MODULE *node;
{
  if (node != NULL)
    {
      buildmodarray (node->left);
      a_modules[node->id] = node;
      buildmodarray (node->right);
    }
}


void 
buildedgearray (node)
     EDGE *node;
{
  if (node != NULL)
    {
      buildedgearray (node->left);
      a_edges[node->id] = node;
      buildedgearray (node->right);
    }
}

void 
buildjuxtarray (node)
     JUXT *node;
{
  if (node != NULL)
    {
      buildjuxtarray (node->left);
      a_juxts[node->id] = node;
      buildjuxtarray (node->right);
    }
}


void 
fillindates ()
{
  int i = 1, j = 0;


  while (i < numdays)
    {
      while (slots[j]->day < i)
	j++;

      dates[i] = (struct date *) malloc (sizeof (struct date));
      dates[i]->day = nextday (dates[i - 1], slots[j]->aday - slots[j - 1]->aday);
      dates[i]->month = nextmonth (dates[i - 1], slots[j]->aday - slots[j - 1]->aday);
      dates[i]->year = nextyear (dates[i - 1], slots[j]->aday - slots[j - 1]->aday);
      i++;
    }
}

int 
nextday (d, jump)
     struct date *d;
     int jump;
{
  if (d->day + jump < 29)
    return (d->day + jump);

  if (d->day + jump == 29)
    if ((d->month != 2) || isleapyear (d->year + 1900))
      return (29);
    else
      return 1;

  if (d->day + jump == 30)
    if (d->month == 2)
      if (isleapyear (d->year + 1900))
	return 1;
      else
	{
	  fprintf (stderr, "Dates seem corrupted -- this is not a leap year\n");
	  fprintf (stderr, "treating it as if it was ...\n");
	  return 1;
	}
    else
      return 30;

  if (d->day + jump == 31)
    switch (d->month)
      {
      case 1:
	return 31;
      case 2:
	fprintf (stderr, "Dates seem corrupted -- too many days in february!");
	return 1;
      case 3:
	return 31;
      case 4:
	return 1;
      case 5:
	return 31;
      case 6:
	return 1;
      case 7:
	return 31;
      case 8:
	return 31;
      case 9:
	return 1;
      case 10:
	return 31;
      case 11:
	return 1;
      case 12:
	return 31;
      }
  if (d->day + jump > 31)
    switch (d->month)
      {
      case 4:
      case 6:
      case 9:
      case 11:
	return (d->day + jump - 30);
      default:
	return (d->day + jump - 31);
      }
}

int 
nextmonth (d, jump)
     struct date *d;
     int jump;
{

  if (nextday (d, jump) < d->day)
    if (d->month == 12)
      return 1;
    else
      return (d->month + 1);
  else
    return d->month;
}

int 
nextyear (d, jump)
     struct date *d;
     int jump;
{
  if ((d->day + jump > 31) && (d->month == 12))
    return (d->year + 1);
  else
    return (d->year);
}

/*******************************************************************/
/* Is input a leap year ? Taken from /usr/local/include/gnu/time.h */
/*******************************************************************/
int 
isleapyear (year)
     int year;
{
  if
    ((year) % 4 == 0 && ((year) % 100 != 0 || (year) % 1000 == 0))
    return 1;
  else
    return 0;
}


void 
fillinslotseatdetails ()
{
  int s;
  for (s = 0; s < numslots; s++)
    if (slots[s]->sscap == -1)
      slots[s]->sscap = slotseatmax;
}

void 
fillinsloteventdetails ()
{
  int s;
  for (s = 0; s < numslots; s++)
    if (slots[s]->secap == -1)
      slots[s]->secap = sloteventmax;
}

/*
void initialise_mixops()
{

   vdm_rate = 0.7;
   crossover_rate = 0.15;
   efm_rate = 0.1;
   mutation_rate = 0.05;

   vdms = crosses = efms = mutations =
   svdms = scrosses = sefms = smutations = 0;

 }
*/

void 
initialise_tnchange_sizes ()
{
  int i;
  double p;
 
  tnscs = (int *) malloc (nummods * sizeof (int));
  if (roomsmatter)
    tnrcs = (int *) malloc (nummods * sizeof (int));
  if (agentsmatter)
    tnacs = (int *) malloc (nummods * sizeof (int));

  for (i = 0; i < nummods; i++)
    {
      p = (double) slot_alleles[i];
      tnscs[i] = maxof (1, (int) floor (log (1.0 - efm_pressure) / log ((p - 1.0) / p)));
      /*    fprintf(tncsfile,"%d slots %d \n",i,tnscs[i]); */
      if (roomsmatter)
	{
	  p = (double) room_alleles[i];
	  tnrcs[i] = maxof (1, (int) floor (log (1.0 - efm_pressure) / log ((p - 1.0) / p)));

	}
      if (agentsmatter)
	{
	  p = (double) agent_alleles[i];
	  tnacs[i] = maxof (1, (int) floor (log (1.0 - efm_pressure) / log ((p - 1.0) / p)));
	}
    }
  /*  fclose(tncsfile); */
}

/********************************************************************/
/*  Set up various things which will set ccm mutation to work:      */
/*  For each event,look at the co-edges among its neighbours. For   */
/*  every such, increment its `ccm'-score by 1. Do this only for    */
/*  events of degree > something. Eg, > numslots-5, to leave room   */
/* for manoeuvre for them.                                          */
/********************************************************************/
void 
initialise_ccm ()
{
  int i, j, k, m1, m2,tot, upto, numleft;
  _MODULE *m;
  LL_EDGES *e, *f;
  unsigned short *ccm_freq;

  /* First off, initialise array of critical co-edges */

  /* Now, look at each module in turn */

  for (i = 0; i < nummods; i++)
    {
      m = a_modules[i];		/* m is the module pointer */
      if (m->numedges > (numslots - 5))
	{
	  printf ("CC-ing module %s, with %d edges\n", m->name, m->numedges);

	  for (e = m->edges; e->next; e = e->next)
	    for (f = e->next; f; f = f->next)
	      {
		/* get the modules out */
		if (e->e->mod1->id == i)
		  m1 = e->e->mod2->id;
		else
		  m1 = e->e->mod1->id;
		if (f->e->mod1->id == i)
		  m2 = f->e->mod2->id;
		else
		  m2 = f->e->mod1->id;
		if (GRAPH (m1, m2) == 0)
		  {
		    if (m1 < m2)
		      CCM_SCORE (m1, m2)++;
		    else
		      CCM_SCORE (m2, m1)++;
		  }
	      }
	}
    }

  /* for(i=0;i<nummods-1;i++)
     for(j=i+1;j<nummods;j++){
       ccescores[i*nummods+j].id = i*nummods+j;
       ccescores[i*nummods+j].s = CCM_SCORE(i,j);
     }
      qsort((char *)ccescores, nummods*(nummods-1)/2, sizeof(struct cce),
     ccecompare); */

  /* Now, Get CCM score frequencies  */

  /* Initialise all frequencies to 0 */

  ccm_freq = (unsigned short *) calloc (nummods + 1, sizeof (unsigned short));

  for (i = 0; i < nummods - 1; i++)
    for (j = i + 1; j < nummods; j++)
      ccm_freq[CCM_SCORE (i, j)]++;	/* Get the frequencies */

  /* print the frequencies */
  for (i = 0; i < nummods + 1; i++)
    printf ("ccm freq %d %d \n", i, ccm_freq[i]);

  /* Now, we want 2,000 of these. So go from nummods down until */
  /* we accumulate 2,000 */

  tot = 0;
  for (i = nummods; i >= 0; i--)
    {
      tot += ccm_freq[i];
      if (tot >= 2000)
	break;
    }
  upto = i + 1;			/* We want all cce's with scores upto or above */
  numleft = 2000 - (tot - ccm_freq[i]);
  /* and numleft of those with frequency i */

  /* So collect these together into a list */

  k = 0;
  for (i = 0; i < nummods - 1; i++)
    for (j = i + 1; j < nummods; j++)
      {
	if (CCM_SCORE (i, j) >= upto)
	  addnewcce (i, j);
	else if (k < numleft)
	  if (CCM_SCORE (i, j) == (upto - 1))
	    {
	      addnewcce (i, j);
	      k++;
	    }
      }
  /* Now print these lovely cce's  */
  for (i = 0; i < 2000; i++)
    printf ("cce %s %s %d \n", a_modules[cces[i]->a]->name,
	    a_modules[cces[i]->b]->name,
	    CCM_SCORE (cces[i]->a, cces[i]->b));

}

void 
addnewcce (i, j)
     int i, j;
{

  cces[n_cce] = (struct cce *) malloc (sizeof (struct cce));
  cces[n_cce]->a = i;
  cces[n_cce++]->b = j;
}
