/************************************************************************
 *                                                                      *
 *               GATT - Genetic Algorithm Time-Tabler                   *
 *                                                                      *
 *                By Dave Corne                                         *
 *                   Peter Ross (peter@aisb.ed.ac.uk),                  *
 *                   Emma Collingwood (emmac@aisb.ed.ac.uk)             *
 *                                                                      *
 ************************************************************************/

/* 
   Copyright (C) 1996, Dave Corne, Peter Ross and Emma Collingwood

   This program is free software you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation- either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY- without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program-if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. 
*/


/***************************************************************
 * Filename: main.l
 *
 * translations of input data for use with lex/yacc
 ****************************************************************/



%{

#define MAXINPUTNUMS 150
#define NEWLINE 0
#include "y.tab.h"
#include <stdlib.h>
#include "gatt.h"

int inputnumlist[MAXINPUTNUMS]; 
double inputdouble;

int numnames = 0;
int numnums = 0;

void resetinputlists();
void addtonumlist();
void addtonamelist();

int  i;                           /* Used for dealing with */
char *sn1, *sn2;                  /* number ranges */
%}
%a 20000
%p 20000
%e 3000
%n 2000
whitespace [ \t]+
nl         (\n)
dotnl      ((\.)([ \t]*)(\n))
name       (([-a-zA-Z_-]+)([-a-zA-Z0-9_]*))
name2      (([0-9]+)([A-Z]+)([0-9]+))
number     ([0-9]+)
numrange   (([0-9]+)("-")([0-9]+))
double     (([0-9]+)(\.)([0-9]+))
colon      (":")
slash      ("/")      
comment    ("#".*)    
continued  ((\\)([ \t]*)(\n))
word       (([-a-zA-Z_-]+)([-a-zA-Z0-9_]*))
%%
{comment}         ;
{whitespace}      ;        /* ignore whitespace */
{continued}       ;        /* skip past newline if backslash before it */
{nl}             { return NEWLINE;} 
{dotnl}          { return NEWLINE;} 
the              { return THE; }
there            { return THERE; }
time             { return TIME; }
timetable        { return TIMETABLE; }
assign           { return ASSIGN; }
date             { return DATE; }
apart            { return APART; }
long             { return LONG; }
last             { return LAST; }
lasts            { return LASTS; }
are              { return ARE ;  }
{number}         { addtonumlist(atoi(yytext));
                   return NUMBER; }
{numrange}       { sn1 = (char *)malloc((i=strcspn(yytext,"-"))+1);
                   strncpy(sn1,yytext,i);
                   sn2 = (char *)malloc(strlen(yytext)-i);
                   strcpy(sn2,&yytext[i+1]);
                   for(i=atoi(sn1);i<=atoi(sn2);i++)
                      addtonumlist(i);
                   free(sn1);
                   free(sn2);
                   return NUMRANGE; }
{double}         { inputdouble = (double)atof(yytext); return DOUBLE; }
day              { return DAY; }
days             { return DAYS; }
slot             { return SLOT; }
slots            { return SLOTS; }
student          { return STUDENT; }
students         { return STUDENTS; }
module           { return MODULE ;  }
modules          { return MODULES ; }
room             { return ROOM ;   }
rooms            { return ROOMS ; }
hour             { return HOUR ; }
hours            { return HOURS; }
is               { return IS ; }
be               { return BE; }
between          { return BETWEEN ; }
on               { return ON ; }
in               { return IN_ ; }
at               { return AT ; }
all              { return ALL ; }
and              { addtonamelist("___"); return AND ; }
min              { return MIN; }
mins             { return MINS; }
minute           { return MINUTE; }
minutes          { return MINUTES; }
take             { return TAKE; }
takes            { return TAKES; }
schedule         { return SCHEDULE; }
p                { return P; }
prefer           { return PREFER; }
preferably       { return PREFERABLY; }
preference       { return PREFERENCE; }
place            { return PLACE; }
occur            { return OCCUR; }
into             { return INTO; }
with             { return WITH; }
priority         { return PRIORITY; }
before           { addtonamelist("___"); return BEFORE; }
after            { addtonamelist("___"); return AFTER; }
exactly          { return EXACTLY; }
weight           { return WEIGHT; }
precisely        { return PRECISELY; }
least            { return LEAST; }
most             { return MOST ; }
pair             { return PAIR ; }
expected         { return EXPECTED; }
size             { return SIZE; }
onto             { return ONTO; }
have             { return HAVE; }
happen           { return HAPPEN; }
per              { return PER; }
as               { return AS; }
unavailable      { return UNAVAILABLE; }
require          { return REQUIRE; }
requires         { return REQUIRES; }
debug            { return DEBUG; }
has              { return HAS; }
avoid            { return AVOID; }
avoids           { return AVOIDS; }
holds            { return HOLDS; }
can              { return CAN; }
extra            { return EXTRA; }
additional       { return ADDITIONAL; }
minimise         { return MINIMISE; }
minimum          { return MINIMUM; }
maximum          { return MAXIMUM; }
number           { return NUMBER_W; }
of               { return OF; }
no               { return NO; }
no\.             { return NUMBER_W; }
taking           { return TAKING; }
same             { return SAME; }
sits             { return SITS; }
sitting          { return SITTING; }
seats            { return SEATS; }
consec           { return CONSEC; }
consecutive      { return CONSECUTIVE; }
concurrently     { return CONCURRENTLY; }
simultaneously   { return SIMULTANEOUSLY; }
simulated_annealing { return SIMULATED_ANNEALING; }
allow            { return ALLOW; }
penalty          { return PENALTY; }
clash            { return CLASH; }
event-spread     { return EVENT_SPREAD; }
constraint       { return CONSTRAINT; }
constraints      { return CONSTRAINTS; }
specifications   { return SPECIFICATIONS; }
exclusions       { return EXCLUSIONS; }
exclude          { return EXCLUDE; }
excludes         { return EXCLUDES; }
specifies        { return SPECIFIES; }
specify          { return SPECIFY; }
ordering         { return ORDERING; }
to               { return TO; }
too              { return TOO; }
from             { return FROM; }
must             { return MUST; }
many             { return MANY; }
exam             { return EXAM; }
exams            { return EXAMS; }
lecture          { return LECTURE; }
lecturer         { return LECTURER; }
lecturers        { return LECTURERS; }
lectures         { return LECTURES; }
tutor            { return TUTOR;}
tutors           { return TUTORS;}
tutorial         { return TUTORIAL; }
tutorials        { return TUTORIALS; }
event            { return EVENT; }
events           { return EVENTS; }
teach            { return TEACH; }
teaches          { return TEACHES; }
teacher          { return TEACHER; } 
teachers         { return TEACHERS; }
invigilator      { return INVIGILATOR; }
invigilators     { return INVIGILATORS;}
invigilates      { return INVIGILATES; }
invigilate       { return INVIGILATE; } 
default          { return DEFAULT;  }
defaults         { return DEFAULTS; }
a                { return A; }
less             { return LESS; }
more             { return MORE; }
than             { return THAN; }
fewer            { return FEWER; }
being            { return BEING; }
close            { return CLOSE; }
share            { return SHARE; }
capacity         { return CAPACITY; }
consecutive_pen         { return CONSECPEN;}
minimize_courses_pen    { return MINCOURSES;}
too_many_per_day_pen    { return TOOMANYPERDAY; }
juxtaposition_pen       { return JUXTAPOSITION; }
too_close_pen           { return  TOOCLOSEPEN ; }
n_too_close_pen         { return  NTOOCLOSEPEN; }
event_exclude_slot_pen  { return MODEXCSLOTPEN; }
agent_preference_pen    { return  AGENTPREFPEN; }
event_avoid_slot_pen    { return  MODAVSLOTPEN; }
event_capacity_pen      { return  EVENTCAPPEN; }
student_capacity_pen    { return  STUDCAPPEN; }
event_exclude_room_pen  { return  MODEXCROOMPEN; }
event_avoid_room_pen    { return  MODAVROOMPEN; }
agent_exclude_slot_pen  { return  AGENTEXCSLOTPEN; }
agent_avoid_slot_pen    { return  AGENTAVSLOTPEN; }
agent_overworked_h_pen  { return  AGENTOVERHPEN; }
agent_underworked_h_pen { return  AGENTUNDERHPEN; }
agent_overworked_e_pen  { return AGENTOVEREPEN; }
agent_underworked_e_pen { return AGENTUNDEREPEN; }
agent_exclude_event_pen { return AGENTEXCMODPEN; }
agent_avoid_event_pen   { return AGENTAVMODPEN; }
agent_clash_pen         { return AGENTCLASHPEN;}
agent_too_many_pen      { return AGENTTOOMANYPEN;}
event_clash_pen         { return CLASHPEN; }
n_event_clash_pen       { return NCLASHPEN; }
slot_seat_load_pen      { return SSLOADPEN; }
slot_event_load_pen     { return SELOADPEN; }
ga_numpops              { return NUMPOPSIS; }
ga_popsize              { return POPSIZEIS; }
ga_gens_stage           { return GENSSTAGEIS; }
ga_reporting_interval   { return REPINITIS; }
ga_migint               { return MIGINITIS; }
use_violation_directed_mutation {return USE_VDM;}
use_event_freeing_mutation { return USE_EFM ; }
ga_efm_rate             { return EFM_RATE ; }
ga_crossover_rate       { return CROSSRATEIS; }
ga_mutation_rate        { return MUTATERATEIS; }
ga_permutation_rate     { return PERMRATEIS; }
ga_pop_mutation_rate    { return MUTATIONRATEIS; }
ga_reinit_rate          { return REINITRATEIS; }
ga_set_directed_mutation { return SETDM; }
ga_set_reinit           { return REINITIS;  }
ga_set_crossover        { return CROSSOVERIS; }
ga_set_selection        { return SELECTIONIS; }
ga_set_reproduction     { return REPRODUCTIONIS; }
use_stochastic_reproduction { return STOCHASTIC_REPRODUCTION; }
use_adaptive_efm        { return ADAPTIVEEFM; }
use_heuristic_eval      { return HEURISTIC_EVAL; }
non_interactive         { return NON_INTERACTIVE; }
expt_output_file        { return EXP_OUTPUT; }
rng_seed                { return SEEDIS; }
initial_temperature     { return INITIAL_TEMPERATURE; }
final_temperature       { return FINAL_TEMPERATURE; }
cooling_rate            { return COOLING_RATE; }
iterations              { return ITERATIONS; }
crowding_factor         { return CROWDING_FACTOR;}
ga_set_restarts         { return RESTARTSIS; }

{colon}          { return COLON; }
{slash}          { return SLASH; }
{name}           { addtonamelist(yytext);
                   return NAME;}
{name2}          { addtonamelist(yytext);
                   return NAME;}
{word}           { return WORD;}
%%


void addtonumlist(num)
int num;
{
 inputnumlist[numnums++] = num;
}

void addtonamelist(name)
char *name;
{
 inputnamelist[numnames] = 
        (struct inputname *)malloc(sizeof(struct inputname));
 inputnamelist[numnames]->n = (char *)malloc(strlen(name)+1);
 strcpy(inputnamelist[numnames++]->n,name);
}

void resetinputlists()
 {
  int i;

  numnums = 0;
  for(i=0;i<numnames;i++) {
   free(inputnamelist[i]->n);
   free(inputnamelist[i]);
  }
  numnames = 0;
}



