/************************************************************************
 *                                                                      *
 *               GATT - Genetic Algorithm Time-Tabler                   *
 *                                                                      *
 *                By Dave Corne                                         *
 *                   Peter Ross (peter@aisb.ed.ac.uk),                  *
 *                   Emma Collingwood (emmac@aisb.ed.ac.uk)             *
 *                                                                      *
 ************************************************************************/

/* 
   Copyright (C) 1996, Dave Corne, Peter Ross and Emma Collingwood

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. 
*/


/**********************************************************************
 *  Filename: main.y 
 *   
 *  main program
 *
 **********************************************************************/ 




%{
#include "gatt.h"             /* Include struct defn's and defines */

void report();              /* Prototypes; defined in this file. */
void free_pops();
void freechromo();
int declarenewroom();
int declarenewagent();
int compchromo();
int equal_timetables();

extern CHROMOSOME *init_pops();
void insert();
void getbestchromo();
void dump_chromos();
extern void init_genotype();
double drandom();
double insert_prob;
int simulated_annealing=0;
double temperature;
double initial_temperature = 20.0;
int iterations = 10000;
int crowding_factor = 5;
double final_temperature = 0.1;
double cooling_rate = 0.90;
int restart, restarts;
double efm_pressure;
extern void dealwithinput();
extern void printstats();        
double sabestfit;
double avhddb();
int sabest[200];
int vbestfitness=1000000;
int dch[100][500];
int dchnum=0;
void dcheck();
int saved_seed;
int seed_given;
int use_seed_given = 0;
extern int *bests;
extern int nbests;
extern int nsimgroups;
double best_fitness;
int replace_equal = 1;
extern void makerunscreen();
extern void makehelpscreen();
extern void makeprintscreen();
extern char *modagentexcs;
extern char *modslotexcs;
extern char *modroomexcs;
int maxconsecutive = 0;  /* max consecutive exams */
int allmodsexcludenum = 0;
/* Defined in chromo.c */

extern void create_chromo();
extern void copy_genotype();
extern void freechromo();
extern CHROMOSOME copychromo();

/* Defined in select.c */

extern CHROMOSOME rank_select();
extern CHROMOSOME fitprop_select();
extern CHROMOSOME tm_select();
extern CHROMOSOME tn_select();

/* Defined in reprod.c */

extern void hillclimb();
extern void one_reproduction();
extern void gen_reproduction();
extern void ss_gen_reproduction();
extern void ss_one_reproduction();
extern void ss_one_boltzmann_reproduction();
extern void mix_ops_ss_one_reproduction();
extern void evolutionary_reproduction();
extern void save_best_reinit();
extern void crossexpts();
extern int bestpen;
extern int bestclash;
extern int besttc;

/* Defined in cross.c */

extern CHROMOSOME one_pt_cross();
extern CHROMOSOME two_pt_cross();
extern CHROMOSOME uniform_cross();
extern CHROMOSOME temporal_stamp_recombination();
extern CHROMOSOME directed_temporal_stamp_recombination();
extern CHROMOSOME violation_directed_recombination();

/* Defined in mutate.c */

extern double adaptive();
extern double nonadaptive();
extern double diversity();
extern double diversity_measure;

/* Defined in reinit.c */

extern void no_reinit();
extern void converge_reinit();
extern void permute_slots_reinit();
extern void complete_reinit();

/* Defined in tt0.c */

extern CHROMOSOME mutate_tt0();
extern CHROMOSOME mutate_vdm();
extern CHROMOSOME mutate_efm();
/* int tnchange_size;
   int tnrchange_size;
   int tnachange_size;
   int tnchoose_size;
*/
/* Defined in print.c */

extern void print_best_tt();
extern void print_student_tts();

/* defined in main.l */

extern void addtonamelist();

/* defined in screens.c */
extern int sort_main_help_action();
extern int sort_out_print_action(); 
extern int current_state_id;
extern int poss_actions_id;
/* Standard C lib stuff */

extern FILE *fopen();
extern int fscanf();
extern char *malloc();
extern char *strdup();
extern long random();
/* Globals used in various files, but declared here  */

double avhd;
int interactive = 1;
int eval_nmodsperday = 0;
int npdpen = 1;
int consecpen = 1;
int clashpen = 1;
int nclashpen = 1;
int juxtpen = 1;
int tooclosepen = 1;
int ntooclosepen = 1;
int modexcslotpen = 1;
int modavslotpen = 1;
int eventcappen = 1;
int studcappen = 1;
int modexcroompen = 1;
int modavroompen = 1;
int agentexcslotpen = 1;
int agentavslotpen = 1;
int agentoverhpen = 1;
int agentunderhpen = 1;
int agentoverepen = 1;
int agentunderepen = 1;
int agentexcmodpen = 1;
int agentavmodpen = 1;
int agentprefpen = 1;
int agentclashpen = 1;
int agenttoomanypen = 1;
int ssloadpen = 1;
int seloadpen = 1;
int studentoomanypen = 1;
int clashperexam_pen = 50;

int slotseatconstraints = 0;
int sloteventconstraints = 0;
int slotseatmax, sloteventmax;
int hctrials;

int numpop = 1,            /* Number of populations */
    numchromo = 200,        /* Number of chromos in a population */
    geno_size = 32,        /* Length of a chromosome */
    gens_stage = 2000,      /* How many times to cycle before asking user */
    repinterval = 100,      /* Interval between screen/file reports */
    isadaptive = FALSE,    /* Using adaptive mutation? */
    last_line = 24,        /* Size of screen; actually found at run time */
    migint = 10,           /* Interval between migrations */
    tournament_size = 3,  /* Default size in tournament selection */
    chromo_id = 1,         /* Chromosome IDs start from this value */
    bestchrom,            /* This keeps track of the best chromosome */
    days,
    slotsperday,
    *bestinpop,           /* Keeps track of best chromo in each pop */
    vdm = FALSE,
    efm = FALSE,
    dsm = FALSE,
    debug = FALSE;

int default_agentmaxmods,
    default_agentmaxhours,
    default_agentminmods,
    default_agentminhours,
    default_expectedmodsize,
    default_agentspermod,
    maxpref,
    maxdaysofar = 0;

  /*  Memory allocators */

JUXT           *newjuxtalloc();
EDGE           *newedgealloc();
LL_MODS        *newllmodalloc();
LL_SPEC        *newllspecalloc();
LL_STUDS       *newllstudalloc();
LL_JUXTS       *newlljuxtsalloc();
LL_EDGES       *newlledgealloc();
LL_JUXTS       *newlljuxtalloc();
_MODULE        *newmodalloc();
_STUDENT       *newstudentalloc(); 
char           *namealloc();

FILE *crosspoints;                 /* File for crossover experiments */
FILE *output = (FILE *)NULL;       /* Output file if any */
FILE *chromofile = (FILE *)NULL;   /* Chromo file, if any output file */
int chromoutput = FALSE;           /* Doing any output? */
int twins = FALSE;         /* Crossover produces pairs, or just one child? */
char outfilename[256], chromofilename[256];
double mute_rate = 0.02,   /* Default proportion of chromo to mutate */
       mutation_rate = 0.2,/* Default proportion of pop to mutate */
       reinit_rate = 0.3,  /* Default reinit mutation rate */
       cross_rate = 0.6,   /* Default crossover rate if applicable */
       bias = 1.5,         /* Default bias for rank selection */
       dsm_rate = 1.0,     /* Default rate for domain-specific mutation*/
       perm_rate = 0.1;
 
double *dchecks, *devals; 
char *eval_name,           /* String giving name of problem */
     *maxfitness,          /* String giving max fitness (if known) */
     *cross_name,          /* String naming crossover type */
     *select_name,         /* String naming selection type */
     *repro_name;          /* String naming reproduction type */
int     *modprefs;            /* Array with maximum agent preferences */
double *total_abs_fitness; /* Used in roulette selection: size of wheel */
int *evals;                /* How often eval has been called in each pop */
int *checks;               /* How many constraint checks per pop */
int ss_xmax,ss_ymax,ss_n;  /* Size of grid and length of walk in ssN */
int n_formae;              /* No. formae to transmit in temp_stamp_rec */  
int n_vdr;                 /* No of genes to swap in viol_direc_recomb */
int trials;                /* Parameter for evolutionary reproduction */
int stochastic_reproduction = 0; /* Is reproduction stochastic ? */
extern double eval_tt();
extern double init_eval_tt();

/* Default bindings for function pointers */

CHROMOSOME (*selection)() = tn_select;
CHROMOSOME (*mutate)() = mutate_tt0;
CHROMOSOME (*crossover)() = uniform_cross;
void (*reproduction)() = one_reproduction;
void (*reinitialise)() = converge_reinit;
double (*mrate)() = nonadaptive;
double (*evaluate)() = eval_tt;
double (*init_evaluate)() = init_eval_tt;
extern void analyse_initial_pop();
extern int randchoose();
extern int abestchoose();
/* extern int tnchoose();*/
extern int randchange();
extern int abestchange();   
extern int tnchange();

/* int (*dmchoose)() = tnchoose;*/
int (*dmchange)() = tnchange; 


int default_roomstudcap = -1;
int default_roommodcap = -1;
int default_modlength = -1;

int minconmods = 0;
int minimods = -1;
int minidays = -1;

int numslotsgiven = 0;
int numroomsgiven = 0;
int numstudentsgiven = 0;
int nummodsgiven = 0;

int numagentsgiven = 0;
int aldaysbetweenmods = -1;
int alslotsbetweenmods = -1;
int alminsbetweenmods = -1;
int agentmodsperday = -1;
int studmodsperday = -1;
int numedges = 0;
int numjuxts = 0;

int roomsmatter = FALSE;
int agentsmatter = FALSE; 
int reprochosen = FALSE;
int spatialchosen = FALSE;
int onegeneforslots = FALSE;
int chromoformat = _SLOT_;
int adaptive_efm = FALSE; 
int lazy_eval = FALSE;
int numtocheck;
extern int inputnumlist[];
extern int numnums;
extern int numnames;
extern int reinits;
extern void resetinputlists();
extern double inputdouble;

/* functions declared here */
int find_num_distinct_bests(CHROMOSOME *);

JUXT      *addnewjuxt();
EDGE      *addnewedge();
_MODULE   *addnewmodule();
_STUDENT  *addnewstudent();

%}

%token THERE ARE NAME NUMBER SLOT SLOTS DAY DAYS IS ON AT BETWEEN 
%token MODULE MODULES ALL MINUTES HAS EXTRA STUDENT STUDENTS TAKES
%token FORCE FORCIBLY F PREFER PREFERABLY P WITH STRENGTH PRIORITY
%token IN_ INTO EXCLUDE FROM BEFORE AFTER EXACTLY LEAST MOST AND PAIR
%token ROOM ROOMS  HAVE MINIMISE NUMBER_W OF EXPECTED WEIGHT
%token TAKING CONSECUTIVE ALLOW TIME DATE LONG LAST MINS MIN LASTS
%token PRECISELY NO LESS THAN FEWER A MINIMUM MORE MAXIMUM 
%token SAME AS TO BE THE CONCURRENTLY SIMULTANEOUSLY MUST TAKE PLACE
%token OCCUR HAPPEN SITS EXAM EXAMS LECTURE LECTURES TIMETABLE
%token COLON SLASH ONTO MINUTE ADDITIONAL SITTING CONSEC SCHEDULE
%token STA SDA CAN DESIRABILITY DES TEACHER TEACHERS LECTURER LECTURERS
%token TUTOR TUTORS AVOID PROHIBIT PENALTY CLASH EVENT_SPREAD CONSTRAINT
%token CONSTRAINTS SPECIFICATIONS EXCLUSIONS ORDERING TOO MANY BEING 
%token CLOSE DOUBLE POPULATION SIZE SPREAD INTERMODULE REINITIS
%token INTERVAL GENERATIONS PER STAGE MIGRATION RANDOM SEED OFFSPRING
%token BREEDING USE EVENT_FREEING MUTATION RATE VIOLATION_DIRECTED
%token CROSSOVERIS NUMRANGE EFM_RATE ADAPTIVEEFM MUTATIONRATEIS RESTARTSIS
%token SELECTIONIS REPRODUCTIONIS REINITRATEIS SHARE STOCHASTIC_REPRODUCTION
%token ONE_POINT  TWO_POINT OUTPUT FILE_W DEMES ONLY GENE FOR APART
%token INVIGILATOR INVIGILATES DEFAULT DEFAULTS HOURS HOUR PREFERENCE
%token INVIGILATE AVOIDS EXCLUDES SPECIFIES SPECIFY TUTORIAL TUTORIALS
%token EVENT EVENTS ASSIGN HOLDS UNAVAILABLE REQUIRE REQUIRES DEBUG
%token TOOMANYPERDAY JUXTAPOSITION TOOCLOSEPEN MODEXCSLOTPEN AGENTPREFPEN
%token MODAVSLOTPEN EVENTCAPPEN STUDCAPPEN MODEXCROOMPEN MODAVROOMPEN 
%token AGENTEXCSLOTPEN AGENTAVSLOTPEN AGENTOVERHPEN AGENTUNDERHPEN 
%token AGENTOVEREPEN AGENTUNDEREPEN AGENTEXCMODPEN AGENTAVMODPEN CLASHPEN
%token NUMPOPSIS POPSIZEIS GENSSTAGEIS REPINITIS MIGINITIS USE_VDM USE_EFM 
%token CROSSRATEIS MUTATERATEIS HEURISTIC_EVAL EXP_OUTPUT SEATS CAPACITY
%token SSLOADPEN SELOADPEN SETDM PERMRATEIS NON_INTERACTIVE SEEDIS
%token INITIAL_TEMPERATURE FINAL_TEMPERATURE COOLING_RATE ITERATIONS
%token NCLASHPEN NTOOCLOSEPEN SIMULATED_ANNEALING CROWDING_FACTOR 
%token  WORD AGENTCLASHPEN AGENTTOOMANYPEN TEACH TEACHES INVIGILATORS
%token MINCOURSES CONSECPEN
%%

 valid_input:  declaration
           |   description
           |   slotdefault
           |   alldefault
           |   modlength
           |   modextrastud
           |   modsize
           |   studenttakes
           |   modexcslotspec
           |   intermodcon
           |   roommatter
           |   roomcap
           |   eventspread
           |   modroomcon
           |   teachcon
           |   lecturecon
           |   gaparameter
           |   blankline
           |   debug
           |   setpenalty
           |   slotseatexception
           |   sloteventexception
          /* |   experiment */
           ;

 debug:         DEBUG
                { debug = TRUE; }
           ;
 blankline :     { ; }    
           ;

 setpenalty: 
     TOOMANYPERDAY COLON NUMBER { npdpen = inputnumlist[0]; numnums=0;}
   | CLASHPEN COLON NUMBER {clashpen = inputnumlist[0]; numnums=0;}
   | NCLASHPEN COLON NUMBER {nclashpen = inputnumlist[0]; numnums=0;}
   | JUXTAPOSITION COLON NUMBER {juxtpen = inputnumlist[0]; numnums=0;}
   | TOOCLOSEPEN COLON NUMBER { tooclosepen = inputnumlist[0]; numnums=0;}
   | NTOOCLOSEPEN COLON NUMBER { ntooclosepen = inputnumlist[0]; numnums=0;}
   | MODEXCSLOTPEN COLON NUMBER {modexcslotpen = inputnumlist[0]; numnums=0;}
   | MODAVSLOTPEN COLON NUMBER {modavslotpen = inputnumlist[0]; numnums=0;}
   | SELOADPEN COLON NUMBER {seloadpen = inputnumlist[0]; numnums=0;}
   | SSLOADPEN COLON NUMBER {ssloadpen = inputnumlist[0]; numnums=0;}
   | EVENTCAPPEN COLON NUMBER {eventcappen = inputnumlist[0]; numnums=0;}
   | STUDCAPPEN COLON NUMBER { studcappen = inputnumlist[0]; numnums=0;}
   | MODEXCROOMPEN COLON NUMBER {modexcroompen = inputnumlist[0]; numnums=0;}
   | MODAVROOMPEN COLON NUMBER {modavroompen = inputnumlist[0]; numnums=0;}
   | AGENTEXCSLOTPEN COLON NUMBER{agentexcslotpen=inputnumlist[0];numnums=0;}
   | AGENTAVSLOTPEN COLON NUMBER {agentavslotpen=inputnumlist[0];numnums=0;}
   | AGENTOVERHPEN COLON NUMBER { agentoverhpen =inputnumlist[0];numnums=0;}
   | AGENTUNDERHPEN COLON NUMBER {agentunderhpen =inputnumlist[0];numnums=0;}
   | AGENTOVEREPEN COLON NUMBER {agentoverepen = inputnumlist[0];numnums=0;}
   | AGENTUNDEREPEN COLON NUMBER {agentunderepen =inputnumlist[0];numnums=0;}
   | AGENTEXCMODPEN COLON NUMBER {agentexcmodpen =inputnumlist[0];numnums=0;}
   | AGENTAVMODPEN COLON NUMBER {agentavmodpen = inputnumlist[0];numnums=0;}
   | AGENTPREFPEN COLON NUMBER { agentprefpen = inputnumlist[0];numnums=0;}
   | AGENTTOOMANYPEN COLON NUMBER { agenttoomanypen = inputnumlist[0];numnums = 0;}
   | AGENTCLASHPEN COLON NUMBER {agentclashpen =  inputnumlist[0];numnums=0;}
   | MINCOURSES COLON NUMBER {clashperexam_pen = inputnumlist[0];numnums=0;}
| CONSECPEN COLON NUMBER {consecpen = inputnumlist[0];numnums=0;}
  
;

 declaration:   declarer NUMBER slotsyn { act_declareslots(); }
           ;
 declaration:   declarer NUMBER daysyn  { act_declaredays(); }
           ;
 declaration:   declarer NUMBER studsyn { act_declarestudents();}
           ;
 declaration:   declarer NUMBER roomsyn { act_declarerooms(); }
           ;
 declaration:   declarer NUMBER modsyn  { act_declaremods(); }
           ;
 declaration:   declarer NUMBER teachsyn { act_declareagents(); }
           ;

 description:  slotdescription          { act_slotdescription(); }
           ;
 description:  roomdescription          { act_roomdescription(); }
           ;
 description:  moddescription           { act_moddescription(); }
           ;
 description:  daydescription           { act_daydescription(); }
           ;
 description:  datedescription          { act_datedescription(); }
           ;
 description:  agentdescription         { act_agentdescription(); }
           ; 


 

 slotdescription: slotsyn NUMBER IS ON DAY NUMBER AT time
               |  slotsyn NUMBER IS ON DAY NUMBER AT TIME time
               ;

 daydescription:  NUMBER daysyn is_are BETWEEN daysyn NUMBER AND NUMBER
               |  NUMBER daysyn is_are BETWEEN daysyn NUMBER AND daysyn NUMBER
               ;

 datedescription: daysyn NUMBER IS ON date
               |  daysyn NUMBER IS ON DATE date
               ;

 slotdefault:     slotsyn DEFAULT COLON NUMBER SEATS
                  { slotseatconstraints = 1;
                    slotseatmax = inputnumlist[0];
                    resetinputlists();}
                  ;
 slotdefault:     slotsyn DEFAULT COLON NUMBER modsyn
                  { sloteventconstraints = 1;
                    sloteventmax = inputnumlist[0];
                    resetinputlists();}
                  ;


 slotseatexception: slotsyn numlist havesyn CAPACITY NUMBER SEATS
                  {
		    slotseatconstraints = 1;
                    act_slotseatexception(); resetinputlists();
                  }
                  ;

 sloteventexception: slotsyn numlist havesyn CAPACITY NUMBER modsyn
                  {
		    sloteventconstraints = 1;
                   act_sloteventexception(); resetinputlists();
                  }
                  ;
 roomdescription: roomsyn ARE COLON namelist
                ;

 moddescription:  modsyn ARE COLON namelist
               ;
 
 agentdescription: teachsyn ARE COLON namelist
               ;


 modlength:       modsyn namelist lastsyn NUMBER minsyn
                  { act_modlength(); }
         ;
 
 modextrastud:    modsyn NAME HAS NUMBER extrasyn studsyn 
                  { act_modextrastud(); }
            ;
 modsize:         EXPECTED SIZE modsyn namelist NUMBER
                  { act_expmodsize(); }
            ;
 
 studenttakes:    studsyn NAME sitsyn namelist
                  { act_studenttakes(0); } 
            |     NAME sitsyn namelist
                  { act_studenttakes(0); }
            ;

 modexcslotspec: ALL modsyn excludesyn slotsyn numlist
                  { act_allmodexcslot();
                    resetinputlists();
                  }
 modexcslotspec:  modsyn namelist excludesyn slotsyn numlist
                  { act_modexcslot(1);
                    resetinputlists(); }
              ;

 modexcslotspec:  modsyn namelist avoidsyn slotsyn numlist WEIGHT NUMBER
                  { act_modexcslot(0); 
                    resetinputlists(); }
              ;
 modexcslotspec:  modsyn namelist specsyn slotsyn NUMBER 
                  { act_modsspecslot(); }
               ;
 intermodcon:     modsyn namelist EXACTLY NUMBER minsyn BEFORE modsyn namelist WEIGHT NUMBER
                  { act_modprecmod(BEFORE,EXACTLY,MINUTES); }
           ; 
 intermodcon:    modsyn namelist EXACTLY NUMBER daysyn BEFORE modsyn namelist WEIGHT NUMBER
                  { act_modprecmod(BEFORE,EXACTLY,DAYS,1); }
           ; 
 intermodcon:     modsyn namelist AT LEAST NUMBER minsyn BEFORE modsyn namelist WEIGHT NUMBER
                  { act_modprecmod(BEFORE,LEAST,MINUTES,1); }
           ;
 intermodcon:     modsyn namelist AT LEAST NUMBER daysyn BEFORE modsyn namelist WEIGHT NUMBER
                 { act_modprecmod(BEFORE,LEAST,DAYS,1); }
           ;
 intermodcon:     modsyn namelist AT MOST NUMBER minsyn BEFORE modsyn namelist WEIGHT NUMBER
                 { act_modprecmod(BEFORE,MOST,MINUTES,1); }
           ;
 intermodcon:     modsyn namelist AT MOST NUMBER daysyn BEFORE modsyn namelist WEIGHT NUMBER
                 { act_modprecmod(BEFORE,MOST,DAYS,1); }
           ;
 intermodcon: schedsyn simulsyn modsyn namelist WEIGHT NUMBER
                { act_modprecmod(SAME,0,0,1); }
                ;
 intermodcon: schedsyn ON THE SAME DAY modsyn namelist WEIGHT NUMBER
                { act_modprecmod(DAY,0,0,1); }
                ;
 intermodcon: EXACTLY NUMBER minsyn BETWEEN modsyn namelist AND modsyn namelist WEIGHT NUMBER
                { act_modprecmod(BETWEEN,EXACTLY,MINUTES,1); }
                ;
 intermodcon: EXACTLY NUMBER daysyn BETWEEN modsyn namelist AND modsyn namelist WEIGHT NUMBER
                { act_modprecmod(BETWEEN,EXACTLY,DAYS,1); }
                ;
intermodcon: AT LEAST NUMBER minsyn BETWEEN modsyn namelist AND modsyn namelist WEIGHT NUMBER
               { act_modprecmod(BETWEEN,LEAST,MINUTES,1); }
                ;
intermodcon: AT LEAST NUMBER daysyn BETWEEN modsyn namelist AND modsyn namelist WEIGHT NUMBER
              { act_modprecmod(BETWEEN,LEAST,DAYS,1); }
                ;
 intermodcon: AT MOST NUMBER minsyn BETWEEN modsyn namelist AND modsyn namelist WEIGHT NUMBER
              { act_modprecmod(BETWEEN,MOST,MINUTES,1); }
                ;
 intermodcon: AT MOST NUMBER daysyn BETWEEN modsyn namelist AND modsyn namelist WEIGHT NUMBER
               { act_modprecmod(BETWEEN,MOST,DAYS,1); }
                ;

 eventspread:     minimiseconsec
           |      allowbetween
           ;

 minimiseconsec:  MINIMISE numberof studsyn sittingsyn consecsyn modsyn
                  { act_minimisestudconmods(); }
              ;
 
minimiseconsec: ALLOW studsyn AT MOST NUMBER CONSECUTIVE modsyn 
{maxconsecutive = inputnumlist[0]; numnums=0;}

lecturecon: schedsyn APART COLON namelist
             { act_studenttakes(1); }
          ;
 
  havesyn: HAVE
         | HAS
         ;
  
 avoidsyn: AVOID 
        |  PROHIBIT
        |  AVOIDS
        ;
    
 numberof:        NUMBER_W OF 
        |         NUMBER_W
        ;

 modroomcon:      modsyn namelist specsyn roomsyn NAME
                  { act_modspecroom(); }
          ;
 modroomcon: modnamesexcroom namelist
                  { act_modexcroom(); }
          ;
 modroomcon: modnameallroom namelist
                  { act_modallroom(); }
          ;
 modnamesexcroom: modsyn namelist excludesyn roomsyn
                  { addtonamelist("___"); }
          ;
 modnameallroom: modsyn namelist requiresyn roomsyn
                  { addtonamelist("___"); }
          ;


 sittingsyn:      SITTING
          |       TAKING
          |       WITH
          ;

 consecsyn:       CONSECUTIVE
         |        CONSEC
         ;


 allowbetween:    ALLOW studsyn atleastsyn NUMBER minsyn BETWEEN modsyn
                  { alminsbetweenmods = inputnumlist[0];
                    resetinputlists();}
            ;
 allowbetween:    ALLOW studsyn atleastsyn NUMBER daysyn BETWEEN modsyn
                  { aldaysbetweenmods = inputnumlist[0];
                    resetinputlists();}
            ;

 allowbetween:    ALLOW teachsyn AT MOST NUMBER modsyn PER DAY
                  { agentmodsperday = inputnumlist[0];
                    resetinputlists();}
              ;
 allowbetween:    ALLOW studsyn AT MOST NUMBER modsyn PER DAY
                  { studmodsperday = inputnumlist[0]; 
                    resetinputlists();}
            ;
 atleastsyn:      AT LEAST
          |       NO LESS THAN 
          |       NO FEWER THAN 
          ;

 schedsyn:  SCHEDULE
        |   TIMETABLE
        ;

 teachcon: NAME teachsyn modsyn namelist
           { act_teachmustdomods();}
        ;
 teachcon: NAME CAN teachsyn modsyn namelist 
           { act_teachcanmods(); }
           ; 
 teachcon: NAME CAN teachsyn modsyn namelist withdessyn NUMBER
           { act_teachprefdomods(); }
         ;
 teachcon: NAME teachsyn MAXIMUM NUMBER modsyn
           { act_teachgivesmaxmods(); }
         ;
 teachcon: NAME teachsyn MAXIMUM NUMBER HOURS
           { act_teachgivesmaxhours(); }
         ;
 teachcon: NAME teachsyn MINIMUM NUMBER modsyn
           { act_teachgivesminmods(); }
         ;
 teachcon: NAME teachsyn MINIMUM NUMBER HOURS
           { act_teachgivesminhours(); }
         ;
 teachcon: teachsyn DEFAULT COLON MAXIMUM NUMBER modsyn
           { default_agentmaxmods = inputnumlist[0];
             resetinputlists(); }
         ;
 teachcon: teachsyn DEFAULT COLON MAXIMUM NUMBER HOURS
          { default_agentmaxhours = inputnumlist[0];
             resetinputlists(); }
          ;
 teachcon: teachsyn DEFAULT COLON MINIMUM NUMBER modsyn
          { default_agentminmods = inputnumlist[0];
             resetinputlists(); }
          ;
 teachcon: teachsyn DEFAULT COLON MINIMUM NUMBER HOURS
          { default_agentminhours = inputnumlist[0];
             resetinputlists(); }
          ;
 teachcon: modsyn DEFAULT COLON NUMBER PER modsyn
          { default_agentspermod = inputnumlist[0];
            resetinputlists(); }
          ;
 teachcon: modsyn DEFAULT COLON EXPECTED SIZE NUMBER
          { default_expectedmodsize = inputnumlist[0];
            resetinputlists(); }
          ;

 teachcon: NUMBER studsyn SHARE modsyn NAME AND NAME
          { act_nstudssharemods();
            resetinputlists();}
           ;
        
 teachcon: modsyn namelist requiresyn NUMBER teachsyn
          { act_modreqagents(); 
            resetinputlists(); }
          ;
 teachcon: teachsyn NAME EXCLUDES slotsyn numlist
          { act_agentexcslots();
             resetinputlists(); }
           ;
 teachcon: teachsyn NAME avoidsyn slotsyn numlist withdessyn NUMBER 
          { act_agentavslots();
             resetinputlists(); }
          ;

 roommatter:      ASSIGN EVENTS TO roomsyn
                  { roomsmatter = TRUE; }
          ;
 roommatter:      ASSIGN teachsyn TO modsyn 
                  { agentsmatter = TRUE; }
          ;
 roomcap:         roomhascap
       |          roomhascapinslot
       ;

 lastsyn:         LAST
       |          LASTS
       ; 
 roomhascap:      roomsyn NAME HOLDS MAXIMUM NUMBER studsyn
                  { act_roomhascapstud(); }
          ;
 roomhascap:      roomsyn NAME HOLDS MAXIMUM NUMBER modsyn
                  { act_roomhascapmod();  }
          ;
 roomhascapinslot:  roomsyn NAME HOLDS NUMBER studsyn IN_ slotsyn numlist
                  { act_roomhascapstudinslot(); }
          ;
 roomhascapinslot:  roomsyn NAME HOLDS NUMBER modsyn IN_ slotsyn numlist
                  { act_roomhascapmodinslot(); }
          ;
 roomhascapinslot: roomsyn NAME UNAVAILABLE IN_ slotsyn numlist
                  { act_roomunavailableinslot(); }
          ;

 simulsyn:        SIMULTANEOUSLY
        |         CONCURRENTLY 
        |         IN_ THE SAME slotsyn
        ;


 excludesyn:      EXCLUDE
          |       EXCLUDES 
          ;

 withdessyn: WITH DESIRABILITY
          |  WITH DES
          | PREFERENCE
          ;

 sitsyn:          TAKES
      |           SITS
      ;

 namelist:        NAME 
        ;
   
 namelist:        namelist NAME
        ;

 numlist:         NUMBER
       |          NUMRANGE
       ;

 numlist:         numlist NUMBER 
       |          numlist NUMRANGE
       ;


 requiresyn:   REQUIRE
          |    REQUIRES
          ;

 alldefault:      modsyn DEFAULT COLON NUMBER minsyn
                  { default_modlength = inputnumlist[0];
                    resetinputlists(); }
          ;

 alldefault:      roomsyn DEFAULT COLON MAXIMUM NUMBER modsyn
                  { default_roommodcap = inputnumlist[0];
                    resetinputlists(); }
          ;
 
 alldefault:      roomsyn DEFAULT COLON MAXIMUM NUMBER studsyn
                 { default_roomstudcap = inputnumlist[0];
                    resetinputlists(); }
          ; 

 declarer:    THERE IS
        |     THERE ARE
        ;

 slotsyn:     SLOT
        |     SLOTS
        ;

 daysyn:      DAY
      |       DAYS
      ;
  
 studsyn:     STUDENT
       |      STUDENTS
       ;

 roomsyn:     ROOM
       |      ROOMS
       ;

 specsyn:     SPECIFIES
       |      SPECIFY 
        ;

 modsyn:     MODULE
      |      MODULES
      |      EXAM
      |      EXAMS
      |      LECTURE
      |      LECTURES
      |      TUTORIAL 
      |      TUTORIALS
      |      EVENT
      |      EVENTS
      ;

 teachsyn:   TEACHER
        |    TEACHERS
        |    LECTURER
        |    LECTURERS
        |    TUTOR
        |    TUTORS
        |    INVIGILATOR
        |    INVIGILATORS
        |    INVIGILATES
        |    INVIGILATE 
        |    TEACHES
        |    TEACH
        ;

 time: NUMBER COLON NUMBER
    ;

 date: NUMBER SLASH NUMBER SLASH NUMBER
    ;

 is_are:  IS
      |   ARE
      ;

 minsyn: MINUTES
      |  MINUTE
      |  MINS
      |  MIN
      ;

 extrasyn: EXTRA
        |  MORE
        |  ADDITIONAL
        ;

 gaparameter: NUMPOPSIS COLON NUMBER { numpop = inputnumlist[0]; numnums=0;}
           |  POPSIZEIS COLON NUMBER { numchromo =inputnumlist[0]; numnums=0;}
           |  GENSSTAGEIS COLON NUMBER {gens_stage=inputnumlist[0]; numnums=0;}
           |  REPINITIS COLON NUMBER {repinterval=inputnumlist[0]; numnums=0;}
           |  MIGINITIS COLON NUMBER {migint = inputnumlist[0]; numnums=0;}
           |  USE_VDM { dsm=1;efm=0;}
           |  USE_EFM { dsm=1;efm=1;}
           |  EFM_RATE COLON DOUBLE { dsm_rate = inputdouble; }
           |  ADAPTIVEEFM { adaptive_efm = 1; }
           |  CROSSRATEIS COLON DOUBLE {cross_rate = inputdouble;}
           |  SELECTIONIS COLON NAME { act_set_select();resetinputlists();}
           |  SELECTIONIS COLON NAME NUMBER 
                                     { act_set_select();resetinputlists();}
           |  SELECTIONIS COLON NAME NUMBER NAME DOUBLE
                                     { insert_prob = inputdouble;
                                   act_set_select();resetinputlists();}
           |  REPRODUCTIONIS COLON NAME { act_set_repro();resetinputlists();}
           |  REINITIS COLON NAME { act_set_reinit(); resetinputlists();}
           |  STOCHASTIC_REPRODUCTION   { stochastic_reproduction = 1; }
           |  REINITRATEIS COLON DOUBLE { reinit_rate = inputdouble;}
           |  CROSSOVERIS COLON NAME { act_set_cross(); resetinputlists();}
           |  MUTATERATEIS COLON DOUBLE { mute_rate = inputdouble; }
           |  MUTATIONRATEIS COLON DOUBLE { mutation_rate = inputdouble; }
           |  PERMRATEIS COLON DOUBLE {perm_rate = inputdouble; }
           |  SETDM COLON DOUBLE { act_set_directed_mutation(); }
           |  HEURISTIC_EVAL { evaluate = eval_tt; }
           |  NON_INTERACTIVE { interactive=0;}
           |  CROWDING_FACTOR COLON NUMBER 
                    {crowding_factor = inputnumlist[0]; numnums=0;}
           |  EXP_OUTPUT COLON NAME { if(!strcmp(inputnamelist[0]->n,"none"))
                                      output=(FILE *)NULL;
                                      else { 
                                       if ((output = fopen(inputnamelist[0]->n,"a")) == NULL) {printf("Couldn't open output file");exit(0);}
                                resetinputlists();}}
       |  SEEDIS COLON NUMBER {use_seed_given=1; seed_given = inputnumlist[0];
                      resetinputlists();}
       |  SIMULATED_ANNEALING {simulated_annealing=1; 
                               numchromo=1; 
                               sabestfit=-1000000.0;}
       |  RESTARTSIS COLON NUMBER {restarts = inputnumlist[0]; numnums=0;restart=0;}
       |  INITIAL_TEMPERATURE COLON DOUBLE {initial_temperature=inputdouble;}
       |  ITERATIONS COLON NUMBER {iterations=inputnumlist[0];numnums=0;}
       |  FINAL_TEMPERATURE COLON DOUBLE {final_temperature=inputdouble;}
       |  COOLING_RATE COLON DOUBLE {cooling_rate=inputdouble;}
           ;

%%

char **filelist;
unsigned nfiles;
extern FILE *yyin;

main(argc,argv)
int argc;
char **argv;
{
  int i, pop,gen,mignext,action,gen_total,fileoutput,ee,evs;
  CHROMOSOME *pops,migchromo;
  filelist = argv+1;
  nfiles = argc-1;

  if(nfiles == 0) {
    fprintf(stderr, "Usage: gatt file1 [file2 ...]\n");
    exit(1);
  }

  printf(" Assimilating the data ... ");
  for(i=0;i<nfiles;i++) {
    yyin = fopen(filelist[i],"r");
    if (!(yyin)) {
      fprintf(stderr, "GATT could not open file %s\n",filelist[i]);
      exit(1);
    }
    fflush(stdout);
    while(!feof(yyin))
      yyparse();
  }
  printf(" Done.\n Initialising ... ");
  dealwithinput();
  
  printf(" Done.");
  
  if (use_seed_given==1)
    srandom(seed_given);
  else /* Initialise random number gen */
    srandom((saved_seed=(unsigned)time(0))); 

  mignext=0;
  
  maxfitness = "1";
  eval_name = "modules/slots";
  cross_name = "uniform";
  select_name = "tn_select";
  repro_name = "one";
  
  if((reproduction == ss_gen_reproduction)
     || (reproduction == ss_one_reproduction)
     || (reproduction == mix_ops_ss_one_reproduction) 
     || (reproduction == ss_one_boltzmann_reproduction)) {
    ss_xmax = (int) sqrt((double)numchromo);
    ss_ymax = numchromo/ss_xmax;
    numchromo = ss_xmax*ss_ymax;
    migint = 0;
  }

 
  if(interactive==0){ 
    
    /* check if we have an output file */
    if (output == (FILE *)NULL) {
      /* open the default one */
      if ((output = fopen("OUTPUT", "r")) == NULL){
	printf("Couldn't open default output file\n");
	exit(0);
      }
    }
    

    total_abs_fitness = (double *)malloc(numpop*sizeof(double));
    pops = init_pops(numpop,numchromo);
    if(simulated_annealing==1)
      cooling_rate =
        exp( (log(final_temperature/initial_temperature)/(gens_stage/iterations)));
    gen_total = 1;
    for(i=0;i<numpop;i++) evals[i]=0;
    for(i=0;i<numpop;i++) checks[i]=0;

    for(gen=1; gen <= gens_stage; gen++, gen_total++) {
      for(pop=0; pop<numpop; pop+=1) {            /* For each generation */
	/* if(pops[0].fitness<0.9){ */
	getbestchromo(pops);
	reproduction(pops,pop*numchromo,numchromo);        /*Reproduce. */
	getbestchromo(pops);
	reinitialise(pops,pop*numchromo,numchromo,reinit_rate);
	getbestchromo(pops);
      }                                           /* Reinitialise */
      if (numpop>1 && migint > 0 && gen%migint==0) {
	migchromo=selection(pops,mignext*numchromo,numchromo);/*Migrate if */
	for(pop=0; pop<numpop; pop+=1) {                   /*necessary. */
	  if(pop != mignext)               /* Don't insert to itssource */
	    insert(pops,migchromo,pop*numchromo,numchromo);
	}
	mignext=(mignext+1)%numpop;
      }

      if((numchromo==1) && (restart<restarts) && (gen==(gens_stage-1)))  {
	if (output){ 
	  if(simulated_annealing==1)
	    bestpen = (int)(1.0/sabestfit - 1);
	  fprintf(output,"Bestpen = %d, cl %d, tc %d, rns %d \n",
		  bestpen, bestclash, besttc, saved_seed);
	}
	restart++; dcheck(pops);
	if(restarts==1) gen=gens_stage; else {
	  free_pops(pops,numpop*numchromo);
	  gen_total = 1;gen=1;
	  pops = init_pops(numpop,numchromo);
	  for(i=0;i<numpop;i++) evals[i] = 0;
	  for(i=0;i<numpop;i++) checks[i] = 0;
	  report(0,fileoutput,pops,0);   
	  sabestfit=-100000.0;
	  bestpen=1000000;
	}
      }
      
    }


    if(numchromo==1)
      fprintf(output,"Bestpen %d, cl %d, tc %d, rns %d  ahd %g  \n",
	      bestpen, bestclash, besttc, saved_seed, avhd);
    else
      fprintf(output," Bestpen %d Num %d HD %g \n", bestpen,
	      find_num_distinct_bests(pops),avhddb(pops));
    exit(0);
  }
  
  stdscr = initscr(); 
  last_line = tgetnum("li"); 
  if(last_line < 0) last_line = 24; 
  /* crosspoints = fopen("cross","a");    */
  
  total_abs_fitness = (double *)malloc(numpop*sizeof(double));
  pops = init_pops(numpop,numchromo);
  getbestchromo(pops);
  gen_total = 1;
  for(i=0;i<numpop;i++) evals[i]=0;
  for(i=0;i<numpop;i++) checks[i]=0;
  action = ASK_AGAIN;
  if(output) {
    fprintf(output,   "\n\n===================================\n");
    fprintf(output,   "               eval = %s\n",eval_name);
    fprintf(output,   "        max-fitness = %s\n",maxfitness);
    fprintf(output,   "       reproduction = %s\n",repro_name);
    fprintf(output,   "             select = %s\n",select_name);
    if(!strcmp(select_name,"fitprop"))
      fprintf(output, "     selection-bias = n/a\n");
    else
      fprintf(output, "     selection-bias = %3.2f\n",bias);
    fprintf(output,   "  chromosome-length = %d\n",geno_size);
    fprintf(output,   "               pops = %d\n",numpop);
    fprintf(output,   "chromosomes-per-pop = %d\n",numchromo);
    fprintf(output,   " migration-interval = %d\n",migint);
    fprintf(output,   "     crossover-type = %s\n",cross_name);
    if(!strcmp(repro_name,"one"))
      fprintf(output, "     crossover-rate = n/a\n");
    else
      fprintf(output, "     crossover-rate = %3.2f\n",cross_rate);
    if(isadaptive)
      fprintf(output, "      mutation-rate = adaptive\n");
    else
      fprintf(output, "      mutation-rate = %3.2f\n", mute_rate);
    fprintf(output,   " reporting-interval = %d\n", repinterval);
    if(twins)
      fprintf(output, "              twins = yes\n");
    else
      fprintf(output, "              twins = no\n");
    fprintf(output,   "\n");
  }

  current_state_id=2;
  poss_actions_id=0;
  report(0,fileoutput,pops,0); 
  current_state_id=0; 
  poss_actions_id=1;
  while(action != QUIT) {
    do {
      switch(getch()) {
      case 'r': case 'R':
	action = RESTART;
	free_pops(pops,numpop*numchromo);
	gen_total = 1;
	pops = init_pops(numpop,numchromo);
	for(i=0;i<numpop;i++) evals[i] = 0;
	for(i=0;i<numpop;i++) checks[i] = 0;
	poss_actions_id = 0;
	report(0,fileoutput,pops,0);
	
	if(output)
	  fprintf(output,
		  "\n\n------ re-starting ------\n\n");
	break;
      case 'c': case 'C':
	action = CONTINUE;
	break;
      case 'q': case 'Q':
	action = QUIT;
	break;
      case 'h' : case 'H':
	action = ASK_AGAIN;
	clear();
	makehelpscreen(pops);
	refresh();
	evs=0;
	for(ee=0;ee<numpop;ee++) evs+=evals[ee];
	action=sort_main_help_action(pops,evs,gen);
	break;
      case 'p': case 'P':
	action = ASK_AGAIN;
	clear();
	makeprintscreen(pops);
	refresh();
	evs=0;
	for(ee=0;ee<numpop;ee++) evs+=evals[ee];
	action=sort_out_print_action(pops,evs,gen);
	break;
      default:
	action = ASK_AGAIN;
	move(0,30);
	clrtoeol();
	refresh(); 
	break;
      }
    } while(action == ASK_AGAIN);
    if(action == CONTINUE) {
      current_state_id=0;
      poss_actions_id=1;
      for(gen=1; gen <= gens_stage; gen++, gen_total++) {
	for(pop=0; pop<numpop; pop+=1) {            /* For each generation */
	  getbestchromo(pops);
	  reproduction(pops,pop*numchromo,numchromo);        /* Reproduce. */
	  getbestchromo(pops);
	  reinitialise(pops,pop*numchromo,numchromo,reinit_rate);
	  getbestchromo(pops);
	}
	
	if((numchromo==1) && (restart<restarts) && (gen==gens_stage)){
	  restart++; 
	  free_pops(pops,numpop*numchromo);
	  gen_total = 1; gen=1;
	  pops = init_pops(numpop,numchromo);
	  for(i=0;i<numpop;i++) evals[i] = 0;
	  for(i=0;i<numpop;i++) checks[i] = 0;
	  report(0,fileoutput,pops,0);
	  
	}
	
	if (numpop>1 && migint > 0 && gen%migint==0) {
	  migchromo=selection(pops,mignext*numchromo,numchromo);
	  /* Migrate if necessary */
	  for(pop=0; pop<numpop; pop+=1) {     
	    if(pop != mignext)               /* Don't insert to its source */
	      insert(pops,migchromo,pop*numchromo,numchromo);
	  }
	  mignext=(mignext+1)%numpop;
	}
	if (gen_total%repinterval==0) {                      /* Report if  */
	  report(gen_total,fileoutput,pops,gen);             /* necessary  */
	}
	if (gen+1==gens_stage)
	  {current_state_id=1; poss_actions_id=0;}
      } 
    }
  }
  if(output)
    fclose(output);
  if(chromoutput)
    fclose(chromofile);
  if (output) printstats();
  if(output) 
    fclose(output);
  free_pops(pops,numpop*numchromo);
  move(last_line-1,0);
  refresh();
  endwin(); 
  exit(0);
}

void act_set_cross()
{
  char *name;
  name = inputnamelist[0]->n;
  
  if      (!strcmp(name,"uniform")) 
    crossover = uniform_cross;
  else
    if (!strcmp(name,"two-point"))
      crossover = two_pt_cross;
    else
      if (!strcmp(name,"one-point"))
	crossover = one_pt_cross;
      else 
	if (!strncmp(name,"temporal_stamp",14)) {
	  crossover = temporal_stamp_recombination;
	  n_formae = atoi(name+14);
	}
	else
	  if (!strncmp(name,"directed_temporal_stamp",23)) { 
	    crossover = directed_temporal_stamp_recombination;
	    n_formae = atoi(name+23);
	  }
	  else
	    if (!strncmp(name,"violation_directed",18)) {
	      crossover = violation_directed_recombination;
	      n_vdr = atoi(name+18);
	    }
}

void act_set_directed_mutation()
{
  efm_pressure = inputdouble;
  resetinputlists();
}

void act_set_repro()
{
  char *name;
  name = inputnamelist[0]->n;
 
  if      (!strcmp(name,"one")) reproduction = one_reproduction;
  else if (!strcmp(name,"gen")) reproduction = gen_reproduction;
  else if (!strncmp(name,"evolutionary",12)) {
    reproduction = evolutionary_reproduction;
    trials = atoi(name+12);
  }
  else if (!strcmp(name,"mis_ops")) reproduction = mix_ops_ss_one_reproduction;
  else if (!strncmp(name,"hillclimb",9)) {
    hctrials = atoi(name+9);
  }
  else if (!strcmp(name,"spatial_one")) {
    reproduction = ss_one_reproduction;
  }
  else if (!strncmp(name, "spatial_one_boltzmann",21)) {
    reproduction = ss_one_boltzmann_reproduction;
  }
  
}

void act_set_reinit()
{
 char *name;
 name = inputnamelist[0]->n;

 if      (!strcmp(name,"none")) reinitialise = no_reinit;
 else if (!strcmp(name,"converge")) reinitialise = converge_reinit;
 else if (!strcmp(name,"permute")) reinitialise = permute_slots_reinit;
 else if (!strcmp(name,"complete")) reinitialise = complete_reinit;
}

void act_set_select()
{ 
 char *name, *name2;
 int num;
 name = inputnamelist[0]->n;
 if(numnames==2)
 name2 = inputnamelist[1]->n;

 num = inputnumlist[0];

 if  (!strcmp(name,"tn")) {
   selection = tn_select;
   tournament_size = num;
 }
 else if  (!strcmp(name,"rank")) {selection = rank_select;}
 else if  (!strcmp(name,"fitprop")) {selection = fitprop_select;}
 else if  (!strcmp(name,"spatial")) {
   spatialchosen = 1; ss_n = num;
   if(!strcmp(name2,"replace_better")) replace_equal=0;
 }
}

void act_declareslots()
{
  numslots = inputnumlist[0];
  resetinputlists();
}
void act_declaredays()
{
  numdays = inputnumlist[0];
  resetinputlists();
}
void act_declarestudents()
{
  numstudents = inputnumlist[0];
  resetinputlists();
}
void act_declarerooms()
{
  numrooms = inputnumlist[0];
  resetinputlists();
}
void act_declaremods()
{
  nummods = inputnumlist[0];
  resetinputlists();
}
void act_declareagents()
{
  numagents = inputnumlist[0];
  resetinputlists();
}

void act_slotdescription()
{
   int ident, day, hrs, mins;

   ident = inputnumlist[0];
   day = inputnumlist[1];
   hrs = inputnumlist[2];
   mins = inputnumlist[3];

   if (!(ident==numslotsgiven)) {
    printf("Error: slots given in wrong order\n");
    exit(0);
  }
  
   if (ident>=numslots) {
     printf("Error: more slots given than previously declared\n");
     exit(0);
   }

   if (day > maxdaysofar)
     maxdaysofar = day; 

   slots[ident] = (struct slot *)malloc(sizeof(struct slot));
   slots[ident]->id = ident;
   slots[ident]->day = day;
   slots[ident]->aday = day;
   slots[ident]->time = (60*hrs)+mins;
   slots[ident]->sscap = -1;
   slots[ident]->secap = -1;
   numslotsgiven++;

   resetinputlists();
 }

void act_roomdescription()
{
  char *roomname;
  int i;
 
 
  for(i=0;i<numnames;i++) {
    roomname = inputnamelist[i]->n;
    declarenewroom(roomname);    
  }
  resetinputlists();
}

void act_moddescription()
{ 
  char     *modname;
  int      i;
  _MODULE  *modp;
  for(i=0;i<numnames;i++) {
    modname = inputnamelist[i]->n;
    t_modules = addnewmodule(t_modules,modname);
    modp = f_module;
  }
  resetinputlists();
}
void act_agentdescription()
{
  char *agentname;
  int i;

  for(i=0;i<numnames;i++) {
    agentname = inputnamelist[i]->n;
    declarenewagent(agentname);
  }
  resetinputlists();
}


void act_daydescription()
{
  int gap, daya, dayb, s = -1, i;
  
  gap = inputnumlist[0];
  daya = inputnumlist[1];
  dayb = inputnumlist[2];

   /* Find the first slot after daya */
  while (!(slots[++s]->day > daya));
  
  /* Add gap to absolute days after slot s */
  for(i=s;i<numslots;i++){
    slots[i]->aday += gap;
  }
  
  resetinputlists();
}

void act_datedescription()
{
  int day, dd, mm, yy;
  
  day = inputnumlist[0];
  dd  = inputnumlist[1];
  mm  = inputnumlist[2];
  yy  = inputnumlist[3];
  
  if (day) {
    printf("Must give date only for day 0, not day %d \n", day);
    resetinputlists();
    return;
  } 
  
  dates[0] = (struct date *)malloc(sizeof(struct date));
  dates[0]->day = dd;
  dates[0]->month = mm;
  dates[0]->year  = yy;
  
  resetinputlists();
}

void act_modlength()
{
 _MODULE *modp;
 int modleng, i;
 
  modleng = inputnumlist[0];
 for(i=0;i<numnames;i++) {
   t_modules = addnewmodule(t_modules,inputnamelist[i]->n);
   modp = f_module;
   modp->len = modleng;
 }
 resetinputlists();
}

void act_expmodsize()
{
 _MODULE *modp;
 int modsize, i;

 modsize = inputnumlist[0];
 for(i=0;i<numnames;i++) {
   t_modules = addnewmodule(t_modules,inputnamelist[i]->n);
   modp = f_module;
   modp->size = modsize;
 }
 resetinputlists();
}

void act_modextrastud()
{
 char *modname;
 int modext;
 _MODULE *modp;

 modname = inputnamelist[0]->n;
 modext = inputnumlist[0];

 t_modules = addnewmodule(t_modules,modname);
 modp = f_module;
 modp->size+= modext;
 resetinputlists();
}

void act_studenttakes(virt)
int virt;
{
  int      i, nmods, start;
  char     *studname;
  _MODULE  *modp;
  _STUDENT *studp;
  
  if (virt) { studname = "virtual";
	      start = 0; }
  else      { studname = inputnamelist[0]->n;
	      start = 1;
	    }
  nmods = numnames;
  
  /* Initialise this student. Note: we assume this student is */
  /* entirely new to the system, and create a new structure   */
  
  studp = addnewstudent(studname);
  
  for(i=start;i<nmods;i++)  {
    /* Look for or place this module in the module tree  */
    t_modules = addnewmodule(t_modules,inputnamelist[i]->n);
    modp = f_module;

    /* Now add this to the list of modules that this student takes */
    /* -- this is a linked list, so the first one is a special case */
    
    if (studp->numtaken==0)   {   /* If this is the first */
      studp->s_mods = newllmodalloc();  /* Get space for it */
      studp->s_mods->mod = modp;        /* Point it to the module */
      studp->s_mods->next = NULL;
      studp->mods = studp->s_mods;
      studp->numtaken++;
    }
    else  {
      studp->p_mods = studp->mods;      /* p_mods points to previous */
      studp->mods = newllmodalloc();    /* Get space for new one */
      studp->mods->mod = modp;          /* Point it to the module */
      studp->mods->next = studp->p_mods;   /* Make the link */
      studp->numtaken++;
    }
    
    /* Similarly, now add this student to the list of takers of the module */
    
    if ((modp->numtakers)==0) {     /* If this is the first */
      modp->s_studs = newllstudalloc();  /* Get space for it */
      modp->s_studs->st = studp;        /* Point it to the student */
      modp->s_studs->next = NULL;
      modp->studs = modp->s_studs;
      modp->numtakers++;
    }
    else {
      modp->p_studs = modp->studs;      /* p_studs points to previous */
      modp->studs = newllstudalloc();   /* Get space for new one */
      modp->studs->st = studp;          /* Point it to the student */
      modp->studs->next = modp->p_studs;   /* Make the link */
      modp->numtakers++;
    }
  }
  
  
  if (debug) printf("Numstudgiven is %d \n", numstudentsgiven);
  resetinputlists();
  /* Individual edge constraints and their sizes are built up afterwards */
}



int declarenewroom(roomname)
char *roomname;
{
  int j;
  /* Find out if room is already declared */
  j=0;
  while ((j<numroomsgiven)&&(strcmp(rooms[j]->name,roomname))) j++;
  
  if (j<numroomsgiven) { /* already declared; it's number is j*/
    return(j);
  }
  else { /* This room is new. Declare it */
    rooms[numroomsgiven] = (struct room *)malloc(sizeof(struct room));
    rooms[numroomsgiven]->name = (char *)malloc(strlen(roomname)+1);
    strcpy(rooms[numroomsgiven]->name,roomname);
    rooms[numroomsgiven]->nummexceptions = 0;
    rooms[numroomsgiven]->numsexceptions = 0;
    rooms[numroomsgiven]->studcap = -1;
    rooms[numroomsgiven]->numexcslots = 0;
    rooms[numroomsgiven++]->modcap = -1;
    return(j);
  }
}

int declarenewagent(agentname)
     char *agentname;
{
  int j;
  
  /* Find out if agent is already declared */
  j=0;
  while ((j<numagentsgiven)&&(strcmp(agents[j]->name,agentname))) j++;
  
  if (j<numagentsgiven) { /* already declared; it's number is j*/
    return(j);
  }
  else { /* This agent is new. Declare it */
    agents[numagentsgiven] = (struct agent *)malloc(sizeof(struct agent));
    agents[numagentsgiven]->name = (char *)malloc(strlen(agentname)+1);
    strcpy(agents[numagentsgiven]->name,agentname);
    agents[numagentsgiven]->nummustdo = 0;
    agents[numagentsgiven]->maxmods = -1;
    agents[numagentsgiven]->minmods = -1;
    agents[numagentsgiven]->maxhours = -1;
    agents[numagentsgiven]->minhours = -1;
    agents[numagentsgiven]->numexcslots = 0;
    agents[numagentsgiven]->numavslots = 0; 
    agents[numagentsgiven]->maxpref = 0;
    agents[numagentsgiven++]->numcando = 0;
    return(j);
  }
}

void act_nstudssharemods()
{
  int   n;
  _MODULE  *mod1, *mod2;
  n = inputnumlist[0];
  
  t_modules  = addnewmodule(t_modules,inputnamelist[0]->n);
  mod1 = f_module;
  t_modules  = addnewmodule(t_modules,inputnamelist[2]->n);
  mod2 = f_module; 
  t_edges = addnewedge(t_edges,mod1,mod2,n);
}




  
void act_roomhascapstud()
{ 
  int roomindex, studcap;
  char *roomname;
  
  roomname = inputnamelist[0]->n;
  studcap = inputnumlist[0];
  
  roomindex = declarenewroom(roomname); 
  
  rooms[roomindex]->studcap = studcap;
  resetinputlists();
}

void act_roomhascapstudinslot()
{
  int roomindex, studcap, s;
  char *roomname;
  
  roomname = inputnamelist[0]->n;
  studcap = inputnumlist[0];
  roomindex = declarenewroom(roomname);
  
  for(s=1;s<numnums;s++) {
  rooms[roomindex]->sexceptions[rooms[roomindex]->numsexceptions] = 
    (struct cap * )malloc(sizeof(struct cap));
  rooms[roomindex]->sexceptions[rooms[roomindex]->numsexceptions]->slot = 
    inputnumlist[s];
  rooms[roomindex]->sexceptions[rooms[roomindex]->numsexceptions++]->cap =
    studcap;
}
  resetinputlists();
}

void act_roomhascapmodinslot()
{
  int roomindex, modcap, s;
  char *roomname;
  
  roomname = inputnamelist[0]->n;
  modcap = inputnumlist[0];
  roomindex = declarenewroom(roomname);

  for(s=1;s<numnums;s++) {
   rooms[roomindex]->mexceptions[rooms[roomindex]->nummexceptions] =
    (struct cap * )malloc(sizeof(struct cap));
   rooms[roomindex]->mexceptions[rooms[roomindex]->nummexceptions]->slot = 
     inputnumlist[s];
   rooms[roomindex]->mexceptions[rooms[roomindex]->nummexceptions++]->cap =
     modcap;
 }
  resetinputlists();
}

void act_roomhascapmod()
{
  int roomindex, modcap;
  char *roomname;
  
  roomname = inputnamelist[0]->n;
  modcap = inputnumlist[0];

  roomindex = declarenewroom(roomname);

  rooms[roomindex]->modcap = modcap;
  resetinputlists();
}

void act_roomunavailableinslot()
{
  char roomindex;
  int s;

  roomindex = declarenewroom(inputnamelist[0]->n);

  for(s=0;s<numnums;s++) {
    rooms[roomindex]->excslots[rooms[roomindex]->numexcslots++]
      = inputnumlist[s];
  }
  resetinputlists();
}
/* If module <modname> is new, declare it, if not, don't. In either case */
/* return its index */

_MODULE *addnewmodule(node,modname)
     _MODULE *node;
     char *modname;
{
  int direction;     /* Where next to go in the tree */

  if(node==NULL) {       /* This is new, so put it in the tree ... */
    
    node = newmodalloc();  /* Try to find memory for it */
    node->name = namealloc(modname);
    strcpy(node->name,modname);
    node->slotspec = node->dayspec = node->roomspec =
      node->agentspec = node->len = node->simgroup =-1;
    
    node->numexcslot = node->numexcagent = node->numexcroom =
      node->numexcday = node->numtakers = node->numedges =
	node->numjuxts = node->numoedges = node->numojuxts = 
	  node->size = node->sim = 0;

    node->id = nummodsgiven++;
    node->left = node->right = NULL;
    node->exslots = 
      node->exagents = 
      node->exrooms = NULL;

    
    node->juxts = NULL;    /* init juxts to NULL, 24/10/97 */
    
    f_module = node;                      /* Send up pointer to it */
    
    /* Now consider the linked list of modules */
    /* See if this is the very first one  */
    
    if(nummodsgiven==1) {   /* If so, special case ... */
      s_modules = node;
      s_modules->next = NULL; 
      modules = s_modules;
    }
    else  {                /* If not, just make the right links */
      p_modules = modules;
      modules = node;
      modules->next = p_modules;
    }     
  }
  else                    /* If this module is already in the tree */
    if ((direction = strcmp(modname,node->name)) == 0) 
      f_module = node;   /* Send up a pointer to it */
  /* Otherwise, continue looking */
   else  if (direction < 0)  
     node->left = addnewmodule(node->left,modname);
   else  if (direction > 0)
     node->right = addnewmodule(node->right,modname);
  return(node);
}


_STUDENT  *addnewstudent(name)
char *name;
{

    if (!numstudentsgiven) {   /* If this is the first student, it's a  */
                              /* special case for the linked list      */
 
   s_students = newstudentalloc();
   s_students->name = namealloc(name);
   strcpy(s_students->name,name);
   s_students->id = numstudentsgiven++;
   s_students->numtaken=0;
   s_students->next = NULL;
   students = s_students;
 }
   else { 
   
   p_students = students;
   students = newstudentalloc();
   students->name = namealloc(name);
   strcpy(students->name,name);
   students->id = numstudentsgiven++;
   students->numtaken=0;
   students->next = p_students;
 }
   return(students);
 }



void act_modexcslot(force)
int force;
{
  int      i, j, end, priority, *slotlist;
  _MODULE  *modp;

  slotlist = inputnumlist;

  if (force) { end=0; priority=10; }
  else { end = 1; priority = slotlist[numnums-1]; }

  for(i=0;i<numnames;i++) { /* For each module */
    t_modules = addnewmodule(t_modules,inputnamelist[i]->n);
    modp = f_module;
    for(j=0;j<numnums-end;j++)  { /* For each excluded slot */
     
      if (modp->numexcslot==0)
                            /* If it's the first, then do this :  */
        {
          modp->s_exslots = newllspecalloc();
          modp->s_exslots->id = slotlist[j];
          modp->s_exslots->forced = force;
          modp->s_exslots->priority = priority;
          modp->s_exslots->next = NULL;
          modp->exslots = modp->s_exslots;
          modp->numexcslot++;
        }
      else                  /* If not, then this : */
        {
          modp->p_exslots = modp->exslots;
          modp->exslots = newllspecalloc();
          modp->exslots->id = slotlist[j];
          modp->exslots->forced = force;
          modp->exslots->priority = priority;
          modp->exslots->next = modp->p_exslots;
          modp->numexcslot++;
	}
    }
  }
 resetinputlists();
}

void act_modreqagents()
{
  int      i, attendees;
  _MODULE  *modp;

   attendees = inputnumlist[0]; 
   for(i=0;i<numnames;i++) {
   t_modules =  addnewmodule(t_modules,inputnamelist[i]->n);
   modp = f_module;
   modp->attendees = attendees;
  }
}
      
void act_modsspecslot()
{
  int      i,s;
  _MODULE  *modp;

  s = inputnumlist[0];

  for(i=0;i<numnames;i++) { /* For each module */
   t_modules = addnewmodule(t_modules,inputnamelist[i]->n);
    modp = f_module;
    modp->slotspec = s;
  }
 resetinputlists();
}


/* rel: before, or between, or same day, or same time */
/* ex_m_l: exactly, at most, or at least */
/* days_mins: days or minutes */

void act_modprecmod(rel, ex_m_l, days_mins) 
int rel, ex_m_l, days_mins;
{
  int      index = 0, endlist1, startlist2, i, j;
  _MODULE  *modp1, *modp2;

  if ((rel==SAME)||(rel==DAY)) {
    /* if these modules are to be at the same time or same day */
    /* then there is just one input list of names, with no divider */
    for(i=0;i<numnames-1; i++) {
     t_modules =  addnewmodule(t_modules,inputnamelist[i]->n);
       modp1 = f_module;
       for(j=i+1;j<numnames; j++){
     t_modules =  addnewmodule(t_modules,inputnamelist[j]->n);
       modp2 = f_module;     
     t_juxts = addnewjuxt(t_juxts,modp1,modp2,rel,ex_m_l,days_mins);
      }
     }
  }
  else {
   /* there are two lists, divided by a "___" */
  while (strcmp(inputnamelist[index]->n,"___")) index++;
   endlist1=index;
   startlist2=index+1;
  
     for(i=0;i<endlist1;i++){
     t_modules =  addnewmodule(t_modules,inputnamelist[i]->n); 
     modp1 = f_module;
       for(j=startlist2;j<numnames;j++) {
       t_modules =  addnewmodule(t_modules,inputnamelist[j]->n);
       modp2 = f_module;
     t_juxts = addnewjuxt(t_juxts,modp1,modp2,rel,ex_m_l,days_mins);
       }
     }
  }
  resetinputlists();
}

JUXT *addnewjuxt(node,modp1,modp2,rel,ex_m_l,days_mins)
int      rel, ex_m_l, days_mins;
JUXT     *node;
_MODULE  *modp1, *modp2;
{
  int      direction, weight,gap=0;
  long     index;
  _MODULE  *temp;

  /* First, capture the gap and/or priority if either of these exist */
  /* If rel is between or before, then there is a gap in force */
  
  if((rel==BEFORE)||(rel==AFTER)||(rel==BETWEEN)) 
      {gap = inputnumlist[0]; weight = inputnumlist[1];}

  /* In all other cases, there is only the weight to be concerned with */
   else weight = inputnumlist[0];
  
  /* Now, put the modules in the right order */
  
   if(modp1->id>modp2->id) { temp=modp1;modp1=modp2;modp2=temp;
                   if (rel==BEFORE) rel=AFTER; 
                   else if (rel==AFTER) rel=BEFORE;}

   index = (long)10000*(modp1->id)+(modp2->id);      /* A hack */

  if(node==NULL) {       /* This is new, so put it in the tree ... */

     node = newjuxtalloc();  /* Try to find memory for it */
     node->id = numjuxts++;
     node->a = modp1;
     node->b = modp2;
     node->index = index; 
     switch (rel) {
          case BEFORE:  node->rel = 0; break;
          case AFTER:   node->rel = 1; break;
          case BETWEEN: node->rel = 2; break;
          case SAME:    node->rel = 3; break;
          case DAY:     node->rel = 4; break;
	  }
     switch (ex_m_l) {
          case EXACTLY: node->elm = 0; break;
          case LEAST:   node->elm = 1; break;
          case MOST:    node->elm = 2; break;
	  }
     switch (days_mins) {
          case DAYS:    node->dsm = 0; break;
          case MINUTES: node->dsm = 1; break;
	  }
      node->gap = gap;
      node->priority = weight;
      node->left = node->right = NULL;


    /* Now update the linked list */

    if(numjuxts==1)  /* It's the first, so a special case : */
      { s_juxts = node; 
        s_juxts->next = NULL;
        juxts = s_juxts;
      }
    else            /* Otherwise, just make the right links */
      { p_juxts = juxts;
        juxts = node;
        juxts->next = p_juxts;
      }

    /* Now add this juxt to the lists in the module structures involved */

     if (modp1->numjuxts==0)     /*  If the first one, then : */
       { 
         modp1->s_juxts = newlljuxtalloc();  /* Get space for it */
         modp1->s_juxts->j = node;
         modp1->s_juxts->next = NULL;
         modp1->juxts = modp1->s_juxts;
         modp1->numjuxts++;
       }
     else {
         modp1->p_juxts = modp1->juxts;
         modp1->juxts = newlljuxtalloc();
         modp1->juxts->j = node;
         modp1->juxts->next = modp1->p_juxts;
         modp1->numjuxts++;
       }

     if (modp2->numjuxts==0)     /*  If the first one, then : */
       {
         modp2->s_juxts = newlljuxtalloc();  /* Get space for it */
         modp2->s_juxts->j = node;
         modp2->s_juxts->next = NULL;
         modp2->juxts = modp2->s_juxts;
         modp2->numjuxts++;
       }
     else {
         modp2->p_juxts = modp2->juxts;
         modp2->juxts = newlljuxtalloc();
         modp2->juxts->j = node;
         modp2->juxts->next = modp2->p_juxts;
         modp2->numjuxts++;
       }
    }    
   else                    /* If this module is already in the tree */
         if ((direction = (index-node->index)) == 0)
           {  /* It's already there, so replace or augment it with the */
              /* new details */

           switch (rel) {
	   case BEFORE:  node->rel = 0; break;
	   case AFTER:   node->rel = 1; break;
	   case BETWEEN: node->rel = 2; break;
	   case SAME:    node->rel = 3; break;
	   case DAY:     node->rel = 4; break;
	   }
           switch (ex_m_l) {
           case EXACTLY: node->elm = 0; break;
           case LEAST:   node->elm = 1; break;
           case MOST:    node->elm = 2; break;
          }
           switch (days_mins) {
           case DAYS:    node->dsm = 0; break;
           case MINUTES: node->dsm = 1; break;
           }
           node->gap = gap;
           node->priority=weight;
	 }
                           /* Otherwise, continue looking */
   else  if (direction < 0)
        node->left = addnewjuxt(node->left,modp1,modp2,rel,ex_m_l,days_mins);
   else  if (direction > 0)
        node->right = addnewjuxt(node->right,modp1,modp2,rel,ex_m_l,days_mins);

   return(node);
}
     


void act_minimisestudconmods()
 { 
   minconmods=1;
 }


void act_modspecroom()
{
  char      *roomname;
  int       i, roomindex;
  _MODULE   *modp;

  roomname = inputnamelist[numnames-1]->n;
  roomindex = declarenewroom(roomname);

  for(i=0;i<numnames-1;i++) {

   t_modules = addnewmodule(t_modules,inputnamelist[i]->n);
    modp = f_module;
    modp->roomspec = roomindex;
  }
  resetinputlists();
}

void act_modexcroom()
{
  int       i, j, roomindex, priority, index, endlist1, startlist2;
  _MODULE   *modp;

  index=0;
  priority = inputnumlist[0];

 while (strcmp(inputnamelist[index]->n,"___")) index++;
   endlist1=index;
   startlist2=index+1;

  for(i=0;i<endlist1;i++) {  
   t_modules = addnewmodule(t_modules,inputnamelist[i]->n);
   modp = f_module;
     for(j=startlist2;j<numnames;j++) {
     roomindex = declarenewroom(inputnamelist[j]->n);

    if (modp->numexcroom==0)
                            /* If it's the first, then do this :  */
      {
          modp->s_exrooms = newllspecalloc();
          modp->s_exrooms->id = roomindex;
          modp->s_exrooms->forced = 1;
          modp->s_exrooms->priority = 10;
          modp->s_exrooms->next = NULL;
          modp->exrooms = modp->s_exrooms;
          modp->numexcroom++;
        }
      else                  /* If not, then this : */
        {
          modp->p_exrooms = modp->exrooms;
          modp->exrooms = newllspecalloc();
          modp->exrooms->id = roomindex;
          modp->exrooms->forced = 1;
          modp->exrooms->priority = 10;
          modp->exrooms->next = modp->p_exrooms;
          modp->numexcroom++;
        }
   }
  }
  resetinputlists();
}

void act_modallroom()
{

  int       i, j, r[MAXROOMS], priority, index, endlist1, startlist2;
  _MODULE   *modp;

  index=0;
  priority = inputnumlist[0];

 while (strcmp(inputnamelist[index]->n,"___")) index++;
   endlist1=index;
   startlist2=index+1;


  /* ASSUMING ALL ROOMS HAVE BEEN ENTERED, we will exclude all the rooms */
  /* NOT mentioned in endlist. So find these rooms */

  for(i=0;i<numroomsgiven;i++) r[i]=0;  /* Mark all rooms as excluded */

  for(j=startlist2;j<numnames;j++)
      r[declarenewroom(inputnamelist[j]->n)]=1; /* Unmark allowed rooms */

    for(i=0;i<endlist1;i++) {
   t_modules = addnewmodule(t_modules,inputnamelist[i]->n);
   modp = f_module;
   for(j=0;j<numroomsgiven;j++){
       if(r[j]==0) {
    if (modp->numexcroom==0)
	/* If it's the first, then do this :  */
	{
          modp->s_exrooms = newllspecalloc();
          modp->s_exrooms->id = j;
          modp->s_exrooms->forced = 1;
          modp->s_exrooms->priority = 10;
          modp->s_exrooms->next = NULL;
          modp->exrooms = modp->s_exrooms;
          modp->numexcroom++;
        }
    else                  /* If not, then this : */
        {
          modp->p_exrooms = modp->exrooms;
          modp->exrooms = newllspecalloc();
          modp->exrooms->id = j;
          modp->exrooms->forced = 1;
          modp->exrooms->priority = 10;
          modp->exrooms->next = modp->p_exrooms;
          modp->numexcroom++;
        }
       }
   }
  }
  resetinputlists();
}

void act_teachgivesmaxmods()
{
  char *agentname;
  int maxm, agentindex;

  agentname = inputnamelist[0]->n;
  maxm = inputnumlist[0];
  agentindex = declarenewagent(agentname);
  agents[agentindex]->maxmods = maxm;
  resetinputlists();
}
void act_teachgivesmaxhours()
{
  char *agentname;
  int maxh, agentindex;

  agentname = inputnamelist[0]->n;
  maxh = inputnumlist[0];
  agentindex = declarenewagent(agentname);
  agents[agentindex]->maxhours = maxh;
  resetinputlists();
}

void act_teachgivesminmods()
{
  char *agentname;
  int minm, agentindex;

  agentname = inputnamelist[0]->n;
  minm = inputnumlist[0];
  agentindex = declarenewagent(agentname);
  agents[agentindex]->minmods = minm;
  resetinputlists();
}

void act_teachgivesminhours()
{
  char *agentname;
  int minh, agentindex;

  agentname = inputnamelist[0]->n;
  minh = inputnumlist[0];
  agentindex = declarenewagent(agentname);
  agents[agentindex]->minhours = minh;
  resetinputlists();
}


void act_teachmustdomods()
{
  char     *teachname;
  int      i, teachindex;
  _MODULE  *modp;

   teachname = inputnamelist[0]->n;

   teachindex = declarenewagent(teachname);

   for(i=1;i<numnames;i++){
   t_modules = addnewmodule(t_modules,inputnamelist[i]->n);
   modp = f_module; 
     agents[teachindex]->mustdolist[agents[teachindex]->nummustdo++] = 
       modp->id;
   }
   resetinputlists();
}
void act_agentexcslots()
{
  char *agentname;
  int i, agentindex;
  
   agentname = inputnamelist[0]->n;
   agentindex = declarenewagent(agentname);
  
   for(i=0;i<numnums; i++) {
    agents[agentindex]->excslots[agents[agentindex]->numexcslots++]
    = inputnumlist[i];
  }
}



void act_agentavslots()
{
  char *agentname;
  int i, agentindex;

   agentname = inputnamelist[0]->n;
   agentindex = declarenewagent(agentname);

   for(i=0;i<numnums-1; i++) {
    agents[agentindex]->avslots[agents[agentindex]->numavslots] = 
       (struct option *)malloc(sizeof(struct option));
    agents[agentindex]->avslots[agents[agentindex]->numavslots]->mod =
    inputnumlist[i];
    agents[agentindex]->avslots[agents[agentindex]->numavslots++]->des =
    inputnumlist[numnums-1];
  }
}

void act_teachprefdomods()
{
   char     *teachname;
   int      i, teachindex, desire;
   _MODULE  *modp;
   int max_desire_sofar = 0;
   
   teachname = inputnamelist[0]->n;
   desire = inputnumlist[0];

   teachindex = declarenewagent(teachname);

   for(i=1;i<numnames;i++){
    t_modules = addnewmodule(t_modules,inputnamelist[i]->n);
    modp = f_module;
     agents[teachindex]->cando[agents[teachindex]->numcando] =
       (struct option *)malloc(sizeof(struct option));
     agents[teachindex]->cando[agents[teachindex]->numcando]->mod =
       modp->id;
     
     /**************************************************************/
     /* V1.6 - KEEP A RECORD OF THE MAXIMUM PREFERENCE OF AN AGENT */
     if (desire > agents[teachindex]->maxpref)
       agents[teachindex]->maxpref = desire;
     /**************************************************************/
     
     agents[teachindex]->cando[agents[teachindex]->numcando++]->des =
       desire;
   }
   resetinputlists();
 }
void act_teachcanmods()
{
   char     *agentname;
   int      i, agentindex, desire;
   _MODULE  *modp;

   agentname = inputnamelist[0]->n;
   desire = 0; 

   agentindex = declarenewagent(agentname);

   for(i=1;i<numnames;i++){
    t_modules =  addnewmodule(t_modules,inputnamelist[i]->n);
    modp = f_module;
     agents[agentindex]->cando[agents[agentindex]->numcando] =
       (struct option *)malloc(sizeof(struct option));
     agents[agentindex]->cando[agents[agentindex]->numcando]->mod =
       modp->id;  
     agents[agentindex]->cando[agents[agentindex]->numcando++]->des =
       desire;
   }
   resetinputlists();
 }


/************************************************************************/
/* This updates the screen and, if necessary, the output file           */
/************************************************************************/

void report(gen_total,fileoutput,pops,gen)
int gen_total, fileoutput,gen;
CHROMOSOME *pops;
{
  int pop, chromo, evs=0;
  double average, best, tmp;
  for(pop=0;pop<numpop;pop++) evs+=evals[pop];

  if(output) {
       fprintf(output, "\nGenerations=%d  Evaluations-so-far=%d\n",
                       gen_total,evs);

	 fprintf(output, "exlcuded: %d\n", pops[bestchrom].agentexcslots);
     }
   for(pop=0; pop<numpop; pop+=1) {
      average=0.0;
       for(chromo=0; chromo<numchromo; chromo+=1) {
          tmp = pops[pop*numchromo+chromo].fitness;
          average+=tmp;
        }
      average/=(double)numchromo;
      best = pops[bestchrom].fitness;

      if(output) {
          fprintf(output, "  %-4d    %12.7f     %12.7f\n",
                             pop,    average,   best);
        }
    }
    if(interactive==1) {
      makerunscreen(pops,evs,gen); 
      clrtoeol();
      refresh();}
}


/************************************************************************/
/* This frees whole population                                          */
/************************************************************************/

void free_pops(whole_pop,num)
CHROMOSOME *whole_pop;
int num;
{
  int i;
  for(i=0; i<num; i++) {
    freechromo(&whole_pop[i]);
  }
  free(whole_pop);
}



/************************************************************************/
/* Dumps chromosomes to a file. Filename is outfilename.gen, eg foo.500 */
/************************************************************************/

void dump_chromos(pops,g)
CHROMOSOME *pops;
int g;
{
  int i,j;

  if(!chromoutput) {
     sprintf(chromofilename,"%s.chr",outfilename);
     chromofile = fopen(chromofilename,"a");
     chromoutput = (chromofile != (FILE *)NULL);
   }

  if(chromoutput) {
    for(i=0;i<numpop;i++) {
      for(j=0;j<numchromo;j++) {
        fprintf(chromofile,
                "%-2d %-4d %-5d %12.7f  ",
                  i,   j,   g,    pops[i*numchromo+j].fitness);
        fprintf(chromofile,
                "   %5d %5d %5d\n",
                                 pops[i*numchromo+j].gdp->id,
                                 pops[i*numchromo+j].gdp->parent1,
                                 pops[i*numchromo+j].gdp->parent2);
      }
      fprintf(chromofile,"\n\n");
    }
  }
}


/************************************************************************/
/* This function inserts 'chromo' into the correct place (sorted by     */
/* fitness) within the subpopulation defined by 'base' and 'range' of   */
/* the populations array 'pops'.                                        */
/************************************************************************/

void insert(pops,chromo,base,range)
CHROMOSOME *pops,chromo;
int base,range;
{
  int current, pop;

  pop = base/numchromo;         /* which pop we're hitting */

                                /* Can it be inserted at all? */
  if (chromo.fitness>pops[base+range-1].fitness) {
    freechromo(&pops[base+range-1]);
    current=base+range-2;       /* Start at the second last place. */
    do {                        /* If 'chromo' is fitter ... */
      if (chromo.fitness>pops[current].fitness) {
        pops[current+1]=pops[current]; /* ... then push it down one. */
      }
      else {
        total_abs_fitness[pop] += fabs(chromo.fitness)
                                   - fabs(pops[current+1].fitness);
        pops[current+1]=chromo;        /* ... else insert 'chromo' below */
        current=base;                  /* and finish.                    */
      }
      current-=1;               /* Work up the array until ... */
    } while (current>=base);    /* we get past the 'base'.     */
    if (chromo.fitness>pops[base].fitness) {
      total_abs_fitness[pop] += fabs(chromo.fitness)
                                 - fabs(pops[base].fitness);
      pops[base]=chromo;        /* If chromo is the best put it in. */
    }
  }
   else {     /* It cannot be inserted, it is terrible */
        freechromo(&chromo);
    }
}

/************************************************************************/
/* Returns >0 if chromo2 is fitter than chromo1, <0 if chromo 1 is      */
/* fitter than chromo2, and 0 if they are equally fit. This is passed   */
/* to qsort as its comparison function.                                */
/************************************************************************/

int compchromo(chromo1,chromo2)
CHROMOSOME *chromo1,*chromo2;
{
  if((chromo2->fitness) > (chromo1->fitness))
     return(1);
  else if((chromo2->fitness) == (chromo1->fitness))
     return(0);
  else
     return(-1);
}

/***********************************************************************/
/* Returns a random double in range 0.0 to 1.0 (never 1.0 exactly)     */
/***********************************************************************/
double drandom()
{
  return((double)random()/2147483647.0);
}

EDGE     *addnewedge(node,mod1,mod2,ns)
EDGE     *node;
_MODULE  *mod1, *mod2;
int ns;
  {
    int direction ;
    long index;

    _MODULE *temp;

                            /* Put the modules in the rigt order */
    if(mod1->id>mod2->id) {temp = mod1; mod1 = mod2; mod2 = temp;}
                            /* work out the index. A hashing hack */
     index = (mod1->id)*10000+(mod2->id);

   if(node==NULL) {       /* This is new, so put it in the tree ... */

     node = newedgealloc();  /* Try to find memory for it */
     node->mod1 = mod1;
     node->mod2 = mod2;
     node->ns = ns; 
     node->id = numedges++;
     node->left = node->right = NULL;

     f_edges = node;                   /* Send up the new entry */
      /* Now consider the linked list of edges */
      /* See if this is the very first one  */

    if(numedges==1)  {  /* If so, special case ... */
       s_edges = node;
       s_edges->next = NULL;
       edges = s_edges;
     }
     else {                  /* If not, just make the right links */
       p_edges = edges;
       edges = node;
       edges->next = p_edges;
     }

     /* Now, make this edge known to the modules involved */
     
     if(mod1->numedges==0) {     /* Mod1's first so :    */
       mod1->s_edges = newlledgealloc();
        mod1->s_edges->e = node;
        mod1->s_edges->next = NULL;
        mod1->edges = mod1->s_edges;
        mod1->numedges++;
     }
     else {                  /* If not first, just make the right links  */
       mod1->p_edges = mod1->edges;
        mod1->edges = newlledgealloc();
        mod1->edges->e = node;
        mod1->edges->next = mod1->p_edges;
        mod1->numedges++;
      }
                                      /* Do the same for mod2 as well */

     if(mod2->numedges==0)      /* Mod1's first so :    */
      {
        mod2->s_edges = newlledgealloc();
        mod2->s_edges->e = node;
        mod2->s_edges->next = NULL;
        mod2->edges = mod2->s_edges;
        mod2->numedges++;
      }
     else                  /* If not first, just make the right links  */
       {
        mod2->p_edges = mod2->edges;
        mod2->edges = newlledgealloc();
        mod2->edges->e = node;
        mod2->edges->next = mod2->p_edges;
        mod2->numedges++;
      }
   }
   else 
     /* Not found this node yet */ 
     if ((direction =(index-(10000*(node->mod1->id)+node->mod2->id))) == 0)
       /* This is already in the tree, so just increment the number */
       /* of sharers of this edge, and return it */
       {node->ns+=ns; f_edges = node; }
    /* Otherwise, continue looking */
     else  if (direction < 0)
       node->left = addnewedge(node->left,mod1,mod2,ns);
     else  if (direction > 0)
       node->right = addnewedge(node->right,mod1,mod2,ns);
    
    return(node);
  }



_MODULE *newmodalloc()
{
 _MODULE *temp;

 if ((temp=(_MODULE *)malloc(sizeof(_MODULE)))!=NULL)
   return temp;
 else {
   printf("Trouble: Couldn't find enough space for storing modules\n");
   exit(0);
 }
}

JUXT *newjuxtalloc()
{
 JUXT *temp;

 if ((temp=(JUXT *)malloc(sizeof(JUXT)))!=NULL)
    return temp;
 else {
   printf("Trouble: Couldn't find space for a juxtaposition constrant\n");
   exit(0);
 }
}


_STUDENT *newstudentalloc()
{
 _STUDENT *temp;
 
 if ((temp=(_STUDENT *)malloc(sizeof(_STUDENT)))!=NULL)
    return temp;
 else {
   printf("Trouble: Couldn't find enough space for storing students\n");
   exit(0);
 }
}

EDGE *newedgealloc()
{
 EDGE *temp;
  
 if ((temp=(EDGE *)malloc(sizeof(EDGE)))!=NULL)
    return temp;
 else {
   printf("Trouble: Couldn't find enough space for storing edges\n");
   exit(0);
 }
}

LL_MODS *newllmodalloc()
{
 LL_MODS *temp;

 if ((temp=(LL_MODS *)malloc(sizeof(LL_MODS)))!=NULL)
    return temp;
 else {
   printf("Trouble: Couldn't find enough space for a new pointer!\n");
   exit(0);
 }
}

LL_SPEC *newllspecalloc()
{
 LL_SPEC *temp;

 if ((temp=(LL_SPEC *)malloc(sizeof(LL_SPEC)))!=NULL)
    return temp;
 else {
   printf("Trouble: Couldn't find enough space for a new pointer!\n");
   exit(0);
 }
}

LL_STUDS *newllstudalloc()
{
 LL_STUDS *temp;

 if ((temp=(LL_STUDS *)malloc(sizeof(LL_STUDS)))!=NULL)
    return temp;
 else {
   printf("Trouble: Couldn't find enough space for a new pointer!\n");
   exit(0);
 }
}

LL_JUXTS *newlljuxtalloc()
{
 LL_JUXTS *temp;

 if ((temp=(LL_JUXTS *)malloc(sizeof(LL_JUXTS)))!=NULL)
    return temp;
 else {
   printf("Trouble: Couldn't find enough space for a new pointer!\n");
   exit(0);
 }
}

LL_EDGES *newlledgealloc()
{
 LL_EDGES *temp;

 if ((temp=(LL_EDGES *)malloc(sizeof(LL_EDGES)))!=NULL)
    return temp;
 else {
   printf("Trouble: Couldn't find enough space for a new pointer!\n");
   exit(0);
 }
}


char *namealloc(name)
char *name;
{
 char *temp;

 if ((temp=(char *)malloc(strlen(name)+1))!=NULL)
    return temp;
 else {
   printf("Trouble: Couldn't find enough space for a name\n");
   exit(0);
 }
}

void getbestchromo(pops)
CHROMOSOME *pops;
{
   int pop;
   double tmp;
      bestchrom = 0;
       for(pop=1; pop<numpop*numchromo;pop++) {
	 tmp = pops[pop].fitness;
	 if(pops[bestchrom].fitness < tmp)
           bestchrom = pop;
       }
 }

/* void update_bests(CHROMOSOME *pops,
                  int newbest,
                  int place,
                  int oldwasbest)
{
    int index;

    if(newbest){
      nbests=1;
      bests[0]=place;
    }
    else 
      if((oldwasbest==1) && (pops[place].fitness<best_fitness)) {
	index=0;
	while( equal_timetables(pops[bests[index]].gdp->genotype,
				pops[place].gdp->genotype)==0)
	  index++;
	bests[index] = bests[--nbests];
      }
      else 
	if((oldwasbest==1) && (pops[place].fitness==best_fitness)) {
          distinct=0;
	  for(index=0;index<numpop*numchromo;index++)
            
	    
int find_num_distinct_bests(pops)
CHROMOSOME *pops;
{
     int pop, b, same;

     nbests=0;

     getbestchromo(pops);
     best_fitness = pops[bestchrom].fitness;

     for(pop=0; pop<numpop*numchromo;pop++)
       if (pops[pop].fitness == best_fitness)
	   {
           same=0;
           for(b=0;b<nbests;b++)
	       {
              if(equal_timetables(pops[bests[b]].gdp->genotype,
                                  pops[pop].gdp->genotype)==1)
		  {same=1; break;}
	       }
           if(same==0)
*/

int find_num_distinct_bests(pops)
CHROMOSOME *pops;
{
     int pop, b, same;

     nbests=0;

     getbestchromo(pops);
     best_fitness = pops[bestchrom].fitness;

     for(pop=0; pop<numpop*numchromo;pop++) 
       if (pops[pop].fitness == best_fitness)
         {
           same=0;
           for(b=0;b<nbests;b++)
             {
              if(equal_timetables(pops[bests[b]].gdp->genotype,
                                  pops[pop].gdp->genotype)==1)
               {same=1; break;}
             }
           if(same==0)
           bests[nbests++]=pop;
         }
    return(nbests);
}

double avhddb(pops)
CHROMOSOME *pops;
{
     int pop, b, same,i,j,k,n;
     double d,h;
     nbests=0;

     getbestchromo(pops);
     best_fitness = pops[bestchrom].fitness;

     for(pop=0; pop<numpop*numchromo;pop++) 
       if (pops[pop].fitness == best_fitness)
         {
           same=0;
           for(b=0;b<nbests;b++)
             {
              if(equal_timetables(pops[bests[b]].gdp->genotype,
                                  pops[pop].gdp->genotype)==1)
               {same=1; break;}
             }
           if(same==0)
           bests[nbests++]=pop;
         }
      d=0.0;n=0;
      for(i=0;i<(nbests-1);i++)
	  for(j=(i+1);j<nbests;j++)
	      {h=0;
               for(k=0;k<nummods;k++)
                  if(pops[bests[i]].gdp->genotype[k]!=
                     pops[bests[j]].gdp->genotype[k]) h++;
               d+=h;         
               n++;
              }
 return(d/n);
}

int equal_timetables(char *tt1, char *tt2)
{
  int l,d;
  d=0;
  for(l=0;l<geno_size;l++)
    if (tt1[l]!=tt2[l]) {d=1; break;}

  if(d==1) return(0); else return(1);
}

void act_slotseatexception()
{
  int i,cap;

  cap = inputnumlist[numnums-1];

  for(i=0;i<numnums-1;i++)
  slots[inputnumlist[i]]->sscap = cap;
}

void act_sloteventexception()
{
  int i,cap;

  cap = inputnumlist[numnums-1];

  for(i=0;i<numnums-1;i++)
    slots[inputnumlist[i]]->secap = cap;
}
  
void act_allmodexcslot()
{
 int i;

  for(i=0;i<numnums;i++)
    allmodsexclude[allmodsexcludenum++]=inputnumlist[i];
}

void adjust_temperature()
{
  static int iterations_so_far=0, changes_so_far=0;
  iterations_so_far++;

  if (changes_so_far ==0)
     temperature = initial_temperature;
  if (iterations_so_far>=iterations)
      {iterations_so_far=0;
       changes_so_far++;
       temperature*=cooling_rate;
     }
  if(temperature<=final_temperature)
     temperature=final_temperature;

}


void dcheck(pops)
CHROMOSOME *pops;
{
  int i,j,k,d,n,h;

  if(simulated_annealing==1) bestpen = (int) (1.0/sabestfit - 1);

  if(bestpen<vbestfitness) {
    vbestfitness=bestpen ;
    dchnum=0;
    if(simulated_annealing==1)
      for(i=0;i<nummods;i++)
	dch[dchnum][i]=sabest[i];
    else
      for(i=0;i<nummods;i++)
	dch[dchnum][i]=pops[0].gdp->genotype[i];
    dchnum++;
  }
  else if(bestpen==vbestfitness){
    if(simulated_annealing==1)
      for(i=0;i<nummods;i++)
	dch[dchnum][i]=sabest[i];
    else 
      for(i=0;i<nummods;i++) 
        dch[dchnum][i]=pops[0].gdp->genotype[i];
    dchnum++;
  }
  
 d=0;
 if (restart==restarts){
   if(dchnum==1) avhd=0.0; else {
     n=0;
     for(i=0;i<(dchnum-1);i++)
       for(j=i+1;j<dchnum;j++)
	 {h=0;n++;
	  for(k=0;k<nummods;k++)
	    if (dch[i][k]!=dch[j][k]) h++;
	  d+=h;
	}
     avhd=d/n;
   } 
 }
  
}

































