/************************************************************************
 *                                                                      *
 *               GATT - Genetic Algorithm Time-Tabler                   *
 *                                                                      *
 *                By Dave Corne                                         *
 *                   Peter Ross (peter@aisb.ed.ac.uk),                  *
 *                   Emma Collingwood (emmac@aisb.ed.ac.uk)             *
 *                                                                      *
 ************************************************************************/

/* 
   Copyright (C) 1996, Dave Corne, Peter Ross and Emma Collingwood

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. 
*/


/**************************************************************************
 *  Filename:  print.c :
 *
 *  routines to print final timetables
 *************************************************************************/

#include "gatt.h"		/* Include everything */
extern int maxconsecutive;
extern int intcompare();

extern int bestchrom;		/* Index to a best chromo in pops    */
extern int last_line;		/* Where messages go                 */
extern int *evals;		/* Part of information printed       */
extern char *modslotexcs;	/* See which exclusions violated     */
extern int slotseatconstraints;
extern int roomsmatter;
extern int agentsmatter;
extern int alminsbetweenmods;
extern int *bests;

/*  prototypes defined here */
void print_different_tts();
void print_best_tt ();
void print_student_tts ();
void print_tt ();
void print_clashes ();
void report_juxt_violation ();
void print_agent_tts();
/*  prototypes defined in tt0.c */

extern void check_edge_violation ();
extern void check_juxt_constraint ();

char *print_time ();
char *print_student ();
char *print_date ();
char *print_d_or_m ();

char *words[11] =
{"student", "students", "person", "people", "suffer",
 "suffers", "day", "days", "minute", "minutes"};
char dstring[9], tstring[6];

char *
print_people (num)
     int num;
{
  if (num == 1)
    return (words[2]);
  else
    return (words[3]);
}

char *
print_suffer (num)
     int num;
{
  if (num == 1)
    return (words[5]);
  else
    return (words[4]);
}

char *
print_student (num)
     int num;
{
  if (num == 1)
    return (words[0]);
  else
    return (words[1]);
}

/**************************************************************************/
/*  print_best_tt:   This chooses a maximally fit timetable, and asks     */
/*                   print_tt() to print it                               */
/**************************************************************************/

void 
print_best_tt (pops)
     CHROMOSOME *pops;
{
    print_tt (pops[bestchrom]);
}


/***************************************************************************/
/*  print_tt(chromo):  prints the timetable represented by chromo, along   */
/*                     with detailed information about which constraints   */
/*                     it violates and how many                            */
/***************************************************************************/

void 
print_tt (chromo)
     CHROMOSOME chromo;
{
  int e1, e2,t1, t2, charsprinted, *cl, *tc, tmp;
  int i, j, k, ns, h1,mn1, thisroomevents, thisroomstuds;
  int viol;
  FILE *tt;
  EDGE *e;
  LL_STUDS *s, *t;
  _MODULE *m, *m1, *m2;
  JUXT *ju;
  _STUDENT *thisStudent;
  LL_MODS *thisMod;
  int student_id,index;
  int pos;
  int **usesSlots;
  int *studTotal;
  char *sname[30];
  
   
  if ((tt = fopen ("best_tt", "a")) == NULL)
    {
      printf (" Sorry, cannot open or create the file \"best_tt\" ");
      return;
    }

  /* First print an intro to this timetable */

  fprintf (tt,
	   "\n---------------------------------------------------------------------\n");
  fprintf (tt, " A Best timetable found after %d evaluations\n\n\n", evals[0]);

  /* Now print out details of the various unsatisfied constraints, if any */

  if (chromo.clashes == 0)	/* If no clashes */
    fprintf (tt, " There are no clashes in this timetable.\n ");
  else
    print_clashes (chromo, tt);

  if (chromo.toocloses == 0)
    fprintf (tt, " \n The minimum distance between events is satisfied for everyone.\n ");
  else
    {
      fprintf (tt, " \n There are %d \"too close\" problems in this timetable. The events and\n", chromo.toocloses);
      fprintf (tt, " people involved, are as follow:\n\n");

      for (e = edges; e; e = e->next)
	{
	  tc = 0;
	  e1 = e->mod1->id;
	  e2 = e->mod2->id;
	  t1 = chromo.gdp->genotype[e1];
	  t2 = chromo.gdp->genotype[e2];
	  ns = e->ns;
	  check_edge_violation (e1, e2, t1, t2, &cl, &tc);
	  if (tc != 0)
	    {
	      if (t1 >= t2)
		{
		  tmp = e1;
		  e1 = e2;
		  e2 = tmp;
		  tmp = t1;
		  t1 = t2;
		  t2 = tmp;
		}
	      m1 = a_modules[e1];
	      m2 = a_modules[e2];
	      fprintf (tt,
		   "  Event %s finishes at %s on %s, but event %s starts\n",
	      m1->name, print_time (t1, m1->len), print_date (t1), m2->name);
	      fprintf (tt,
		       "  at %s on %s , leaving only %d %s between. In all, \n %d %s %s: ",
		       print_time (t2, 0), print_date (t2),
		       ((alminsbetweenmods > -1) ?
		   (tmp = (slots[t2]->time - (slots[t1]->time + m1->len))) :
			(tmp = (slots[t2]->aday - slots[t1]->aday))),
		       print_d_or_m (alminsbetweenmods, tmp), ns,
		       print_people (ns), print_suffer (ns));

	      charsprinted = 3;
	      for (s = e->mod1->studs; s; s = s->next)
		{
		  for (t = e->mod2->studs; t; t = t->next)
		    {
		      if (s->st->id == t->st->id)
			{
			  if ((charsprinted + strlen (s->st->name) > 75) ||
			      (charsprinted == 3))
			    {
			      fprintf (tt, "\n   %s  ", s->st->name);
			      charsprinted = 3 + strlen (s->st->name) + 2;
			    }
			  else
			    {
			      fprintf (tt, "%s  ", s->st->name);
			      charsprinted += (strlen (s->st->name) + 2);
			    }
			}
		    }
		}
	      fprintf (tt, "\n\n");
	    }
	}
    }

  /* juxtapositions */

  for (ju = juxts; ju; ju = ju->next)
    {
      e1 = ju->a->id;
      e2 = ju->b->id;
      check_juxt_constraint (ju, chromo.gdp->genotype[e1],
			     chromo.gdp->genotype[e2], &viol);
      if (viol)
	report_juxt_violation (tt, ju, e1, e2, chromo.gdp->genotype[e1],
			       chromo.gdp->genotype[e2]);
    }

  /* room event capacities */

  if (roomsmatter)
    if (chromo.eventcapviols == 0)
      fprintf (tt,
	" \n All of the room event capacity constraints are satisfied.\n ");
    else
      {
	fprintf (tt,
		 " \nThere are %d room event capacity violations \n",
		 chromo.eventcapviols);

      }

  if (roomsmatter)
    if (chromo.studcapviols == 0)
      fprintf (tt,
      " \n All of the room student capacitiy constraints are satisfied.\n ");
    else
      {
	fprintf (tt,
		 " \nThere are %d student event capacity violations \n",
		 chromo.studcapviols);
      }

  /* agent stuff */
  if (agentsmatter) {
    if (chromo.agentclash == 0)
      fprintf(tt, 
	      "\n All the constraints between the agents are satisfied\n");
    else
      fprintf(tt, "\n There are %d clashes involving agents only:\n",
	      chromo.agentclash);
      
    if (chromo.agenttoomany > 0)
      fprintf(tt, "\n There are %d agents who exceed the maximum daily load\n",
	      chromo.agenttoomany);

    if (chromo.agentexcslots > 0)
      fprintf(tt,"\n There are %d agents who take events in slots from which they are excluded\n", chromo.agentexcslots);
    
    if (chromo.agentovere > 0 || chromo.agentoverh > 0)
      fprintf(tt, "\n There are overloaded agents(s)\n");

    if (chromo.agentundere > 0 || chromo.agentunderh > 0)
      fprintf(tt, "\n There are underloaded agents(s)\n");
  }

  /* Now look at the exclusions */
  
  if (chromo.modexcslots == 0)
    fprintf (tt,
	     " \n All of the exclusion constraints are satisfied in this timetable.\n ");
  else
    {
      fprintf (tt,
	       " \nThere are %d events in slots from which they should be excluded if\n",
	       chromo.modexcslots);
      fprintf (tt,
	       " possible. These are as follow:\n\n");
      charsprinted = 3;
      for (m = modules; m; m = m->next)
	if (modslotexcs[nummods * chromo.gdp->genotype[m->id] + m->id] != 0)
	  {
	    if ((charsprinted + strlen (m->name) > 75) ||
		(charsprinted == 3))
	      {
		fprintf (tt, "\n   %s  ", m->name);
		charsprinted = 3 + strlen (m->name) + 2;
	      }
	    else
	      {
		fprintf (tt, "%s  ", m->name);
		charsprinted += (strlen (m->name) + 2);
	      }
	  }
      fprintf (tt, "\n\n");
    }


  if (slotseatconstraints)
    {
      for (i = 0; i < numslots; i++)
	{
	  if (chromo.ssloads[i] > slots[i]->sscap)
	    {
	      fprintf (tt, " slot %d is overcrowded. It needs %d seats but has capacity %d\n", i, chromo.ssloads[i], slots[i]->sscap);
	      fprintf (tt, " The offending events are: \n");
	    }
	  else
	    {
	      fprintf (tt, " Slot %d is within capacity. It requires %d seats from %d\n", i, chromo.ssloads[i], slots[i]->sscap);
	      fprintf (tt, " The events involved are: \n");
	    }
	  for (j = 0; j < nummods; j++)
	    if (chromo.gdp->genotype[j] == i)
	      fprintf (tt, " %s (%d)   ",
		       a_modules[j]->name, a_modules[j]->numtakers);
	  fprintf (tt, "\n");
	}
    }

  fprintf (tt, "\n\n");

  /* print out stuff about consecutive exams */
  if (maxconsecutive > 0 && chromo.consecutives > 0){
    fprintf(tt, "%d students have more than %d consecutive exams\n",
            chromo.consecutives, maxconsecutive);
    fprintf(tt,"The students involved are: \n");
    /* find them !! */
    studTotal = (int *)calloc(numstudents, sizeof(int));
    usesSlots = calloc(numstudents,sizeof(int *));
    
    for (i=0;i<numstudents;i++){
      usesSlots[i] = (int *)calloc(numslots, sizeof(int));
    }
    
    
    /* for each students, go through exam lists and work out how many exams
     * per day - if number is greater than modsperday than increment count */
    thisStudent = students;
    for(thisStudent;thisStudent!=NULL;thisStudent=thisStudent->next){
      thisMod = thisStudent->mods;
      student_id = thisStudent->id;
      /*strcpy(sname[student_id],thisStudent->name);*/
      studTotal[student_id] = 0;
      for (thisMod;thisMod != NULL;thisMod = thisMod->next){
        usesSlots[student_id][studTotal[student_id]++] =
          chromo.gdp->genotype[thisMod->mod->id];
      }
    }

    /* now sort the arrays into ascending order of slots */
    for (i=0;i<numstudents;i++)
      qsort((char *)usesSlots[i],studTotal[i], sizeof(int),intcompare);

    /* now calculate biggest consecutive group for each student */
    for (i=0;i<numstudents;i++){
      index = 0;
      pos = 0;
      while (pos < studTotal[i]){
        while (usesSlots[i][pos+1] == usesSlots[i][pos]+1){
          index++;
          pos++;
        }
        if (index >= maxconsecutive){
          for (j=0;j<studTotal[i];j++)
            fprintf(tt,"student %d has exam in slot %d\n",i,usesSlots[i][j]);
          fprintf(tt,"\n");
        }
        index = 0;
        pos++;
      }
    }
  
    free(studTotal);
    for (i=0;i<numstudents;i++){
      free(usesSlots[i]);
    }
    
    free(usesSlots);
  }
  
  
  /* Now actually print the timetable  */

  fprintf (tt, "\n Here is the timetable: \n\n");

  for (i = 0; i < numslots; i++)
    {
      if ((i == 0) || ((i != 0) && (slots[i]->day != slots[i - 1]->day)))
	fprintf (tt,
		 " On %c%d%/%c%d%/%c%d at %c%d%:%c%d : ",
	dates[slots[i]->day]->day > 9 ? 32 : '0', dates[slots[i]->day]->day,
		 dates[slots[i]->day]->month > 9 ? 32 : '0', dates[slots[i]->day]->month,
		 dates[slots[i]->day]->year > 9 ? 32 : '0', dates[slots[i]->day]->year,
		 (h1 = (slots[i]->time/60)) > 9 ? 32 : '0', h1,
		 (mn1 = (slots[i]->time - 60 * h1)) > 9 ? 32 : '0', mn1);

      else
	fprintf (tt,
		 "             at %c%d:%c%d : ",
		 (h1 = (slots[i]->time/60)) > 9 ? 32 : '0', h1,
		 (mn1 = (slots[i]->time - 60 * h1)) > 9 ? 32 : '0', mn1);

      charsprinted = 24;
      for (m = modules; m; m = m->next)
	if (chromo.gdp->genotype[m->id] == i)
	  if (charsprinted + strlen (m->name) > 75)
	    {
	      fprintf (tt, "\n                        %s  ",
		       m->name);
	      charsprinted = 24 + strlen (m->name) + 2;
	    }
	  else
	    {
	      fprintf (tt, "%s  ", m->name);
	      charsprinted += strlen (m->name) + 2;
	    }
      fprintf (tt, "\n");

      /* Now print room information about this slot */


      if (roomsmatter)
	fprintf (tt, " In this slot, rooms are utilised as follows:\n ");

      for (j = 0; j < numrooms; j++)
	{
	  thisroomevents = 0;
	  thisroomstuds = 0;
	  for (k = 0; k < nummods; k++)
	    if ((chromo.gdp->genotype[nummods + k] == j) &&
		(chromo.gdp->genotype[k] == i))
	      {
		thisroomevents++;
		thisroomstuds += a_modules[k]->size;
	      }
	  fprintf (tt, " room %s ", rooms[j]->name);
	  if (roomcaps[j][i][0] > 0) 
	    fprintf(tt, "(event cap %d) ",  roomcaps[j][i][0]);
	  if (roomcaps[j][i][1] > 0) 
	    fprintf(tt, "(student cap  %d) ",  roomcaps[j][i][1]);
          fprintf(tt, "has %d events and %d students.\n",
		  thisroomevents, thisroomstuds);
	}
    }
  fprintf (tt,
	   "\n\n---------------------------------------------------------------------\n");

  for (i = 0; i < nummods; i++)
    fprintf (tt, " %s %d \n", a_modules[i]->name, chromo.gdp->genotype[i]);
  fprintf (tt, " \n");

  if (roomsmatter)
    for (i = 0; i < nummods; i++)
      fprintf (tt, " %s (%d) %s \n", a_modules[i]->name, a_modules[i]->size,
	       rooms[chromo.gdp->genotype[nummods + i]]->name);
}

void 
print_clashes (chromo, tt)
     CHROMOSOME chromo;
     FILE *tt;
{
  int e1, e2,t1, t2, charsprinted, *cl, *tc;
  int ns;
  EDGE *e;
  LL_STUDS *s, *t;


  fprintf (tt,
  " There are %d individual clashes in this timetable. The events which \n",
	   chromo.clashes);
  fprintf (tt,
	   " clash, and the people involved, are as follow:\n\n");

  for (e = edges; e; e = e->next)
    {
      e1 = e->mod1->id;
      e2 = e->mod2->id;
      ns = e->ns;
      t1 = chromo.gdp->genotype[e1];
      t2 = chromo.gdp->genotype[e2];
      check_edge_violation (e1, e2, t1, t2, &cl, &tc);
      if (cl != 0)
	{
	  fprintf (tt,
		   "  Events %s and %s clash at %s on %s; this involves\n",
		   e->mod1->name, e->mod2->name,
		   print_time (t1, 0), print_date (t1));

	  fprintf (tt,
		   "  a total of %d %s who should attend both, namely: ",
		   ns, print_student (ns));
	  charsprinted = 3;
	  for (s = e->mod1->studs; s; s = s->next)
	    {
	      for (t = e->mod2->studs; t; t = t->next)
		{
		  if (s->st->id == t->st->id)
		    {
		      if ((charsprinted + strlen (s->st->name) > 75) ||
			  (charsprinted == 3))
			{
			  fprintf (tt, "\n   %s  ", s->st->name);
			  charsprinted = 3 + strlen (s->st->name) + 2;
			}
		      else
			{
			  fprintf (tt, "%s  ", s->st->name);
			  charsprinted += (strlen (s->st->name) + 2);
			}
		    }
		}
	    }
	  fprintf (tt, "\n\n");
	}
    }
}



void print_different_tts(pops,num)
     CHROMOSOME *pops;
     int num;
{
  int i;

  for (i=0;i<num;i++)
    print_tt (pops[bests[i]]);
}

void  print_agent_tts(CHROMOSOME *pops)
{
  int i,start,finish,module,h1,mn1,x,a;
  FILE *tt;


  if ((tt = fopen ("agent_tt", "a")) == NULL)
    {
      printf (" Sorry, cannot open or create the file \"best_tt\" ");
      return;
    }
  
  /* go through the best chromosome array numagents times
   * looking for modules takesn by the agent
   */

  start = nummods + (nummods*roomsmatter);
  finish = nummods + (nummods*roomsmatter) + nummods;
 
  
  for (a=0;a<numagents;a++){
    fprintf(tt,"agent %s : \n", agents[a]->name);
    for (x=start;x<finish;x++){
      if (pops[bestchrom].gdp->genotype[x] == a){
	module = x-(nummods+(nummods*roomsmatter));
	i =
	  pops[bestchrom].gdp->genotype[module];
	fprintf (tt,
		 " On %c%d%/%c%d%/%c%d at %c%d%:%c%d :     %s",
		 dates[slots[i]->day]->day > 9 ? 32 : '0', 
		 dates[slots[i]->day]->day,
		 dates[slots[i]->day]->month > 9 ? 32 : '0', 
		 dates[slots[i]->day]->month,
		 dates[slots[i]->day]->year > 9 ? 32 : '0',
		 dates[slots[i]->day]->year,
		 (h1 = (slots[i]->time / 60)) > 9 ? 32 : '0', h1,
		 (mn1 = (slots[i]->time - 60 * h1)) > 9 ? 32 : '0', mn1,
		 a_modules[module]->name );
	fprintf(tt, "\n");
      }
    }
    fprintf(tt, "\n\n");
  }
}

void 
print_student_tts (CHROMOSOME *pops)
{

  int i,h1,mn1;
  _STUDENT *thisStudent;
  LL_MODS *thisMod;
  FILE *tt;

  if ((tt = fopen ("student_tt", "a")) == NULL)
    {
      printf (" Sorry, cannot open or create the file \"best_tt\" ");
      return;
    }

     /* go through the student array, and for each student, we want
	to print out his exam, time and place
     */

     thisStudent = students;
     for(thisStudent;thisStudent!=NULL;thisStudent=thisStudent->next){
       thisMod = thisStudent->mods;
       fprintf(tt, "Student %s\n", thisStudent->name);
       for (thisMod;thisMod != NULL;thisMod = thisMod->next){
	 i = pops[bestchrom].gdp->genotype[thisMod->mod->id];
	 fprintf (tt,
		  " On %c%d%/%c%d%/%c%d at %c%d%:%c%d :     %s",
		  dates[slots[i]->day]->day > 9 ? 32 : '0', dates[slots[i]->day]->day,
		  dates[slots[i]->day]->month > 9 ? 32 : '0', dates[slots[i]->day]->month,
		  dates[slots[i]->day]->year > 9 ? 32 : '0', dates[slots[i]->day]->year,
		  (h1 = (slots[i]->time / 60)) > 9 ? 32 : '0', h1,
		  (mn1 = (slots[i]->time - 60 * h1)) > 9 ? 32 : '0', mn1,
		  thisMod->mod->name );
	 fprintf(tt, "\n");
       }
       fprintf(tt, "\n\n");
     }
	 
}

char *
print_time (slot, add)
     int slot, add;
{
  int h, m, t;

  t = slots[slot]->time + add;
  h = t / 60;
  m = t - (60 * h);

  if ((h < 10) && (m < 10))
    sprintf (tstring, "0%d:0%d", h, m);
  else if (h < 10)
    sprintf (tstring, "0%d:%d", h, m);
  else if (m < 10)
    sprintf (tstring, "%d:0%d", h, m);
  else
    sprintf (tstring, "%d:%d", h, m);
  return (tstring);
}

char *
print_date (slot)
     int slot;
{
  int d, m, y;

  d = dates[slots[slot]->day]->day;
  m = dates[slots[slot]->day]->month;
  y = dates[slots[slot]->day]->year;

  if ((d < 10) && (m < 10) && (y < 10))
    sprintf (dstring, "0%d/0%d/0%d", d, m, y);
  else if ((d < 10) && (m < 10))
    sprintf (dstring, "0%d/0%d/%d", d, m, y);
  else if ((d < 10) && (y < 10))
    sprintf (dstring, "0%d/%d/0%d", d, m, y);
  else if ((m < 10) && (y < 10))
    sprintf (dstring, "%d/0%d/0%d", d, m, y);
  else if (d < 10)
    sprintf (dstring, "0%d/%d/%d", d, m, y);
  else if (m < 10)
    sprintf (dstring, "%d/0%d/%d", d, m, y);
  else if (y < 10)
    sprintf (dstring, "%d/%d/0%d", d, m, y);
  else
    sprintf (dstring, "%d/%d/%d", d, m, y);
  return (dstring);

}

char *
print_d_or_m (minutes, num)
     int minutes, num;
{
  if (minutes > -1)
    {
      if (num == 1)
	return words[8];
      else
	return words[9];
    }
  else
    {
      if (num == 1)
	return words[6];
      else
	return words[7];
    }
}



void 
report_juxt_violation (tt, j, e1, e2, s1, s2)
     FILE *tt;
     JUXT *j;
     int e1, e2, s1, s2;
{

  int rel, dsm, elm, gap, priority;
  char t1[6], t2[6], d1[9], d2[9];

  strcpy (t1, print_time (s1, 0));
  strcpy (t2, print_time (s2, 0));
  strcpy (d1, print_date (s1));
  strcpy (d2, print_date (s2));

  rel = j->rel;
  dsm = j->dsm;
  elm = j->elm;
  gap = j->gap;
  priority = j->priority;


  fprintf (tt, " The following juxtaposition constraint is violated:\n");


  switch (rel)
    {
    case 0:			/* a before b */
      if (elm == 0)
	{			/*    exactly  */
	  fprintf (tt, "%s should start exactly %d %s before %s, but\n",
		   j->a->name, gap, print_d_or_m ((dsm - 1), gap),
		   j->b->name);
	  fprintf (tt, "%s starts at %s on %s, and %s starts at %s on %s\n\n",
		   j->a->name, t1, d1, j->b->name, t2, d2);
	}
      else if (elm == 1)
	{			/*   at least  */
	  fprintf (tt, "%s should start at least %d %s before %s, but\n",
		   j->a->name, gap, print_d_or_m ((dsm - 1), gap),
		   j->b->name);
	  fprintf (tt, "%s starts at %s on %s, and %s starts at %s on %s\n\n",
		   j->a->name, t1, d1, j->b->name, t2, d2);
	}
      else if (elm == 2)
	{			/*  at most */
	  fprintf (tt, "%s should start at most %d %s before %s, but\n",
		   j->a->name, gap, print_d_or_m ((dsm - 1), gap),
		   j->b->name);
	  fprintf (tt, "%s starts at %s on %s, and %s starts at %s on %s\n\n",
		   j->a->name, t1, d1, j->b->name, t2, d2);
	}
      break;

    case 1:			/* mod1 after mod2 */
      if (elm == 0)
	{			/*    exactly  */
	  fprintf (tt, "%s should start exactly %d %s after %s, but\n",
		   j->a->name, gap, print_d_or_m ((dsm - 1), gap),
		   j->b->name);
	  fprintf (tt, "%s starts at %s on %s, and %s starts at %s on %s\n\n",
		   j->a->name, t1, d1, j->b->name, t2, d2);
	}
      else if (elm == 1)
	{			/*   at least  */
	  fprintf (tt, "%s should start at least %d %s after %s, but\n",
		   j->a->name, gap, print_d_or_m ((dsm - 1), gap),
		   j->b->name);
	  fprintf (tt, "%s starts at %s on %s, and %s starts at %s on %s\n\n",
		   j->a->name, t1, d1, j->b->name, t2, d2);
	}
      else if (elm == 2)
	{			/*  at most */
	  fprintf (tt, "%s should start at most %d %s after %s, but\n",
		   j->a->name, gap, print_d_or_m ((dsm - 1), gap),
		   j->b->name);
	  fprintf (tt, "%s starts at %s on %s, and %s starts at %s on %s\n\n",
		   j->a->name, t1, d1, j->b->name, t2, d2);
	}
      break;
    case 2:			/* between */
      if (elm == 0)
	{
	  fprintf (tt, "There should be exactly %d %s between %s and %s, but\n",
		gap, print_d_or_m ((dsm - 1), gap), j->a->name, j->b->name);

	  fprintf (tt, "%s starts at %s on %s, and %s starts at %s on %s\n\n",
		   j->a->name, t1, d1, j->b->name, t2, d2);
	}
      else if (elm == 1)
	{
	  fprintf (tt, "There should be at least %d %s between %s and %s, but\n",
		gap, print_d_or_m ((dsm - 1), gap), j->a->name, j->b->name);

	  fprintf (tt, "%s starts at %s on %s, and %s starts at %s on %s\n\n",
		   j->a->name, t1, d1, j->b->name, t2, d2);
	}
      else if (elm == 2)
	{
	  fprintf (tt, "There should be at most %d %s between %s and %s, but\n",
		gap, print_d_or_m ((dsm - 1), gap), j->a->name, j->b->name);

	  fprintf (tt, "%s starts at %s on %s, and %s starts at %s on %s\n\n",
		   j->a->name, t1, d1, j->b->name, t2, d2);
	}
      break;
    case 3:			/* same time */
      fprintf (tt, "%s and %s should start at the same time, but %s starts\n",
	       j->a->name, j->b->name, j->a->name);
      fprintf (tt, "at %s on %s and %s starts at %s on %s\n\n",
	       t1, d1, j->b->name, t2, d2);
      break;
    case 4:			/* same day */
      fprintf (tt, "%s and %s should start on the same day, but %s starts\n",
	       j->a->name, j->b->name, j->a->name);
      fprintf (tt, "at %s on %s and %s starts at %s on %s\n\n",
	       t1, d1, j->b->name, t2, d2);
      break;
    }

}
