/************************************************************************
 *                                                                      *
 *               GATT - Genetic Algorithm Time-Tabler                   *
 *                                                                      *
 *                By Dave Corne                                         *
 *                   Peter Ross (peter@aisb.ed.ac.uk),                  *
 *                   Emma Collingwood (emmac@aisb.ed.ac.uk)             *
 *                                                                      *
 ************************************************************************/

/* 
   Copyright (C) 1996, Dave Corne, Peter Ross and Emma Collingwood

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. 
*/


/**********************************************************************
 *  Filename: screens.c
 *
 *  Routines to format output screens for running GATT interactively 
 *  using curses functions
 **********************************************************************/ 
#include "gatt.h"


int ask_how_many_to_print();
void advise_printed_n_tts ();
int sort_help_action();
void advise_printed_student_tts();
void advise_printed_agent_tts();
void help_student_tt();
void help_agent_tt();
void help_number_tt();

extern int numstudents;

char *current_state[3] =
{" Trying to find an optimal timetable",
 " Stopped for now, awaiting instructions",
 " Just started, waiting for input"};
int current_state_id = 0;

char *possible_actions[2] =
{" (c)ontinue, (r)estart, (p)rint, (h)elp, (q)uit : ",
 " None; please wait "};

int poss_actions_id = 0;
extern int studentoomanypen;
extern int studmodsperday;
extern int consecpen;

extern int agentsmatter;
extern int roomsmatter;
extern int slotseatconstraints;
extern int sloteventconstraints;
extern int find_num_distinct_bests ();
extern char **filelist;
extern unsigned nfiles;
extern int bestchrom;
extern int clashpen;
extern int juxtpen;
extern int tooclosepen;
extern int modexcslotpen;
extern int modavslotpen;
extern int ssloadpen;
extern int seloadpen;
extern int eventcappen;
extern int studcappen;
extern int modexcroompen;
extern int modavroompen;
extern int agentexcslotpen;
extern int agentavslotpen;
extern int agentoverhpen;
extern int agentunderhpen;
extern int agentoverepen;
extern int agentunderepen;
extern int agentexcmodpen;
extern int agentavmodpen;
extern int agentprefpen; 
extern int agentclashpen;
extern int agenttoomanypen;
extern int *bests;
extern int numedges;
extern int numjuxts;
extern int gens_stage;
extern int  clashperexam_pen;

void advise_printed_best_tt ();
void cant_print_agent();

extern void print_best_tt();
extern void print_different_tts();
extern void print_student_tts();
extern void print_agent_tts();


void 
makerunscreen (pops, evs, gen)
     CHROMOSOME *pops;
     int evs, gen;
{
  CHROMOSOME c;
  int i, sofar;

  c = pops[bestchrom];
  /* mvprintw(16,1,"best chrom is %d\n", bestchrom); */
  
  mvprintw (1, 1,
	    "------------------------------------------------------------------------");
  mvprintw (2, 1,
	    " GATT: Genetic Algorithm TimeTabler, (c) Dave Corne and Peter Ross, 1995");
  mvprintw (3, 1,
	    "------------------------------------------------------------------------");
  mvprintw (4, 1,
	    " Problem Described in files: ");

  for (i = 0, sofar = 29; i < nfiles; i++)
    {
      mvprintw (4, sofar, " %s ", filelist[i]);
      sofar += (1 + strlen (filelist[i]));
    }
  mvprintw (6, 1, " Problem Details:");
  mvprintw (7, 3, " %d events", nummods);
  mvprintw (7, 18, "%d timeslots", numslots);
  mvprintw (7, 33, "%d rooms", numrooms);
  mvprintw (7, 48, "%d agents", numagents);

  mvprintw (8, 3, "%d clash constraints", numedges);
  mvprintw (8, 33, "%d other binary event constraints", numjuxts);

  mvprintw (10, 1, "Best timetable found so far:");
  clrtoeol ();
  mvprintw (11, 3, "%d (%d) clashes", c.clashes, c.clashes * clashpen);
  clrtoeol ();
  mvprintw (11, 38, "%d (%d) juxtapositions", c.juxtviols, c.juxtscore * juxtpen);
  clrtoeol ();
  mvprintw (12, 3, "%d (%d) too-closes", c.toocloses, c.toocloses * tooclosepen);
  clrtoeol ();
  mvprintw (13, 3, "%d (%d) events in avoided slots", c.modavslots,
	    c.modavslots * modavslotpen);
  
  /* if (studmodsperday  > 0)
     mvprintw (14, 3, "%d (%d) students with too many per day", c.student_over,  c.student_over * studentoomanypen); */
    
    mvprintw (14, 3, "%d (%d) mods in clashes", c.nclash_exams,
            c.nclash_exams *   clashperexam_pen);
  mvprintw(14,38,  "%d (%d) too many consecutive exams", c.consecutives, c.consecutives*consecpen);
  if (agentsmatter)
    {
      mvprintw (12, 38, "%d (%d) agent daily load ", c.agenttoomany,
		c.agenttoomany* agenttoomanypen);
      mvprintw (13, 38, "%d (%d) agent clashes    ", c.agentclash,
		c.agentclash * agentclashpen);
      mvprintw (15, 3, "%d (%d) agent event load ", c.agentundere + c.agentovere, (c.agentundere * agentunderepen)+(c.agentovere * agentoverepen));
      mvprintw (15, 38, "%d (%d) agent hourly load ", c.agentunderh + c.agentoverh, (c.agentunderh * agentunderhpen) + (c.agentoverh * agentoverhpen));
   
      mvprintw (16,3, "%d (%d) agents in excluded slots \n", c.agentexcslots,
		c.agentexcslots * agentexcslotpen);
      mvprintw (16,38, "%d (%d) agents doing excluded mods \n", c.agentexcmod,c.agentexcmod * agentexcmodpen);
      clrtoeol ();
    }
  if (roomsmatter)
    {
      mvprintw (16, 38, "%d (%d) events in avoided rooms", c.modavrooms,
		c.modavrooms * modavroompen);
      clrtoeol ();
      mvprintw (17, 3, "%d (%d) events overcrowding rooms", c.eventcapviols,
		c.eventcapviols * eventcappen);
      clrtoeol ();
      mvprintw (17, 38, "%d (%d) students overcrowding rooms", c.studcapviols,
		c.studcapviols * studcappen);
      clrtoeol ();

    }
  if (sloteventconstraints)
    mvprintw (18, 3, "%d (%d) events overcrowding slots", c.seload,
	      c.seload * seloadpen);
  clrtoeol ();
  if (slotseatconstraints)
    mvprintw (18, 38, "%d (%d) students overcrowding slots", c.ssload,
	      c.ssload * ssloadpen);
    
  clrtoeol ();
  mvprintw (20, 1, " Number of different timetables as good as this : %d ",
	    find_num_distinct_bests (pops));
  clrtoeol ();
  mvprintw (21, 1, "Current State of GATT: %s ", current_state[current_state_id]);
  clrtoeol ();

  mvprintw (22, 3, "Total evaluations so far: %d   Evaluations to go: %d ",
	    evs, gens_stage - gen);
  clrtoeol ();
  mvprintw (23, 1, "Possible Actions: %s ", possible_actions[poss_actions_id]);
  clrtoeol ();

}

void 
makeprintscreen (pops)
     CHROMOSOME *pops;
{
  clear();
  mvprintw (1, 1,
	    "------------------------------------------------------------------------");

  mvprintw (2, 1,
  " GATT: Genetic Algorithm TimeTabler, (c) Dave Corne and Peter Ross,1994");
  mvprintw (3, 1,
	    "------------------------------------------------------------------------");

  mvprintw (5, 3, "Main Printing Options :");
  mvprintw (7, 15, "p : print the best timetable");
  mvprintw (8, 15, "n : print a number of distinct different timetables");
  mvprintw (9, 15, "s : print individual timetables for each student");
  mvprintw (10, 15, "a : print individual timetables for each agent");
  mvprintw (11, 15, "h : help about printing options");
  mvprintw (12, 15, "r : return to main program screen");
  mvprintw (13, 15, "q : quit program");
  mvprintw (15, 15, "Type chosen option here: ");
  move (15, 41);
  clrtoeol ();
}


void
makehelpscreen()
{
  clear();
  mvprintw (1, 1,
	    "------------------------------------------------------------------------");

  mvprintw (2, 1,
  " GATT: Genetic Algorithm TimeTabler - Help on Options");
  mvprintw (3, 1,
	    "------------------------------------------------------------------------");

  mvprintw (5, 3, "Which option would you like help on ? ");
  mvprintw (8, 15, "c : on Continue");
  mvprintw (9, 15, "r : on Restart");
  mvprintw (10, 15, "p : on Printing");
  mvprintw (14, 15, "b : return to main screen");
  mvprintw (15, 15, "q : quit program");
  mvprintw (18, 15, "Type chosen option here: ");
  move (18, 41);
  clrtoeol ();

}


void 
makeprinthelpscreen ()
{

  clear();
  mvprintw (1, 1,
	    "------------------------------------------------------------------------");

  mvprintw (2, 1,
  " GATT: Genetic Algorithm TimeTabler - Help on Printing");
  mvprintw (3, 1,
	    "------------------------------------------------------------------------");

  mvprintw (5, 3, "Which option would you like help on ? ");
  mvprintw (8, 15, "p : on printing the best timetable");
  mvprintw (9, 15, "n : on printing a number of distinct different timetables");
  mvprintw (10, 15, "s : on printing individual timetables for each student");
  mvprintw (11, 15, "a : on printing individual timetables for each agent");
  mvprintw (14, 15, "r : return to print screen");
  mvprintw (15, 15, "q : quit program");
  mvprintw (18, 15, "Type chosen option here: ");
  move (18, 41);
  clrtoeol ();
}

int 
sort_out_print_action (pops, evs, gen)
     CHROMOSOME *pops;
     int evs, gen;
{

  int mm = 0;		/* What's returned to the main menu */
  int num;			/* Number of distinct timetables to print */
  int bests;

  bests = find_num_distinct_bests(pops); 
 
  while(mm == 0) switch (getch ())
    {
    case 'p':
    case 'P':
      move(21,5);clrtoeol();
      move(22,5);clrtoeol();
      print_best_tt (pops);
      advise_printed_best_tt ();
      move (15, 41);
      clrtoeol ();
      refresh ();
      mm = sort_out_print_action (pops, evs, gen);
      break;

      case 'n': case 'N':
	move(21,5);clrtoeol();
	move(22,5);clrtoeol();
                       num = ask_how_many_to_print(bests);
                       print_different_tts(pops,num);
                       advise_printed_n_tts(num);
		       move (15, 41);
		       clrtoeol ();
		       refresh ();
		       mm = sort_out_print_action (pops, evs, gen);
		       break;
        
      case 's': case 'S':
	move(21,5);clrtoeol();
	move(22,5);clrtoeol();
                    print_student_tts(pops);
                    advise_printed_student_tts();
		    move (15, 41);
		    clrtoeol ();
		    refresh ();
		    mm = sort_out_print_action (pops, evs, gen);
		    break;
   
     case 'a': case 'A':
       move(21,5);clrtoeol();
      move(22,5);clrtoeol();
       if (numagents == 0){
	 cant_print_agent();
       }
       else{
	 print_agent_tts(pops);
	 advise_printed_agent_tts();
       }	 
       move (15, 41);
       clrtoeol ();
       refresh ();
       mm = sort_out_print_action (pops, evs, gen);
       break;
	 

     case 'h': case 'H':
                    makeprinthelpscreen();
		    refresh();
                    mm=sort_help_action();
                    makeprintscreen(); refresh();
                    mm=sort_out_print_action(pops,evs,gen);
                    break;
	

    case 'r':
    case 'R':
      clear ();
      makerunscreen (pops, evs, gen - 1);
      refresh ();
      mm = ASK_AGAIN;
      break;

    case 'q':
    case 'Q':
      mm = QUIT;
      break;
    }

  return (mm);
}

void  help_printed_best_tt ()
{
  clear();

  mvprintw(5,5, "Help on printing the best timetable");
  mvprintw(7,5, 
	        "This option prints out the best timetable found so far");
 mvprintw(8,5, "to a file called best_tt. If there are clashes in this");
 mvprintw(9,5, "timetable, then the events and people involved are listed.");
 mvprintw(10,5,"The whole timetable is then printed out.");


  mvprintw(13,5, "Press any key to return to help screen     ");
  while (!getch());
}


void  help_student_tt ()
{
  clear();
   mvprintw(5,5, "Help on printing individual STUDENT timetables");
   mvprintw(7,5, "With this option, individual timetables for each student");
   mvprintw(8,5,"are printed, showing the exams taken by the student, with");
   mvprintw(9,5,"the date and time of each.");
   mvprintw(12,5, "Press any key to return to help screen     ");
  while (!getch());
}

void  help_agent_tt ()
{
  clear();
   mvprintw(5,5, "Help on printing individual AGENT timetables");
   mvprintw(7,5, "With this option, individual timetables for each agent");
   mvprintw(8,5,"are printed, showing the exams invigilated by the agent");
   mvprintw(9,5,"and the date and time of each exam.");
   mvprintw(12,5, "Press any key to return to help screen     ");
  while (!getch());
}


void  help_number_tt ()
{
  clear();
   mvprintw(5,5, "Help on printing a number of timetables");
   mvprintw(7,5, "GATT maybe able to find a number of different timetables");
   mvprintw(8,5,"of equivalent quality - this option prints out ALL such");
   mvprintw(9,5,"timetables");
   mvprintw(12,5, "Press any key to return to help screen     ");
  while (!getch());
}

void help_restart()
{
  clear();
 mvprintw(5,5, "Help on (R)estart");
 mvprintw(7,5, "Press R to restart GATT from scratch - this re-initialises");
 mvprintw(8,5,"the population randomly and starts from 0 evaluations again. ");
  mvprintw(12,5, "Press any key to return to help screen     ");
  while (!getch());
}





void help_print(){
 clear();
 mvprintw(5,5, "Help on (P)rinting");
 mvprintw(7,5, "Press P to print out a timetable. This invokes the printing");
 mvprintw(8,5,"screen from which the following options are available:");
 mvprintw(10,5,"Print the best timetable");
 mvprintw(11,5,"Print a number of timetables");
 mvprintw(12,5,"Print individual student timetables");
 mvprintw(13,5,"Print individual agent timetables");
 mvprintw(14,5, "Help about printing");
 mvprintw(17,5,"Press any key to return to help screen     ");
  while (!getch());
}




void help_continue()
{
 clear();
 mvprintw(5,5, "Help on (C)ontinue");
 mvprintw(7,5, "Press C to run the genetic algorithm for a set number");
 mvprintw(8,5,"of evaluations. The actual number can be set using the ");
 mvprintw(9,5,"parameter ga_gens_stage. ");
 mvprintw(10,5,"For example,ga_gens_stage : 10000 ");
 mvprintw(12,5,"The screen is updated in intervals set by the parameter");
 mvprintw(13,5,"ga_reporting_interval.");
 mvprintw(14,5,"For example, ga_reporting_interval     : 100");
 mvprintw( 17,5, "Press any key to return to help screen     ");
 while (!getch());
}
 
int 
sort_help_action (pops, evs, gen)
     CHROMOSOME *pops;
     int evs, gen;
{
  int mm = 0;			/* What's returned to the main menu */
  
  while(mm == 0) switch (getch ())
    {
    case 'p':
    case 'P':
      help_printed_best_tt ();
      makeprinthelpscreen();
      move (18, 41);
      clrtoeol ();
      refresh ();
      mm = sort_help_action (pops, evs, gen);
      break;

      case 'n': case 'N':
	help_number_tt();
        makeprinthelpscreen();
	move (18, 41);
	clrtoeol ();
	refresh ();
	mm = sort_help_action (pops, evs, gen);
	break;
        
      case 's': case 'S':
	help_student_tt();
	 makeprinthelpscreen();
	 move (18, 41);
	clrtoeol ();
	refresh ();
	mm = sort_help_action (pops, evs, gen);
	break;
	
     case 'a': case 'A':
       help_agent_tt(pops);
       makeprinthelpscreen();
       move (18, 41);
       clrtoeol ();
       refresh ();
       mm = sort_help_action (pops, evs, gen);
       
       break;
		      
    case 'r':
    case 'R':
      clear ();
      makeprinthelpscreen ();
      refresh ();
      mm = ASK_AGAIN;
      break;

    case 'q':
    case 'Q':
      mm = QUIT;
      break;
    }

  return (mm);
}

int 
sort_main_help_action (pops, evs, gen)
     CHROMOSOME *pops;
     int evs, gen;
{
  int mm = 0;			/* What's returned to the main menu */
  
  while(mm == 0) switch (getch ())
    {
   
    case 'c': case 'C' :
      help_continue ();
      makehelpscreen();
      move (18, 41);
      clrtoeol ();
      refresh ();
      mm = sort_main_help_action (pops, evs, gen);
      break;

      case 'r': case 'R':
	help_restart();
        makehelpscreen();
	move (18, 41);
	clrtoeol ();
	refresh ();
	mm = sort_main_help_action (pops, evs, gen);
	break;
        
      case 'p': case 'P':
	help_print();
	 makehelpscreen();
	 move (18, 41);
	clrtoeol ();
	refresh ();
	mm = sort_main_help_action (pops, evs, gen);
	break;
	
		      
    case 'b':
    case 'B':
      poss_actions_id = 0;
      clear ();
      makerunscreen (pops, evs, gen - 1);
      refresh ();
      mm = ASK_AGAIN;
      break;

    case 'q':
    case 'Q':
      mm = QUIT;
      break;
    }

  return (mm);
}




void
advise_printed_agent_tts()
{
  mvprintw (17, 5,
      "Details of agents appended to file: agent_tt");
  clrtoeol ();
}

void 
advise_printed_best_tt ( )
{
  mvprintw (17, 5,
      "Details of a best timetable found so far appended to file: best_tt");
  clrtoeol ();
}


void 
advise_printed_n_tts (int bests)
{
  mvprintw (17, 5,
      "Details of  %d timetables appended to file: best_tt",bests);
  clrtoeol ();
}

void advise_printed_student_tts()
{
 mvprintw (17, 5,
      "Details of  %d student timetables appended to file: student_tt",
	   numstudents);
  clrtoeol ();
}
  
int ask_how_many_to_print(int bests)
{
  int ok;
  int newpen;
  char buffer[50],  *p;
  buffer[0] = 5;
  ok = 0;
  p = (char *)malloc(50);

  while (ok == 0)
    {
      move(17,5);
      clrtoeol();
      mvprintw (17,5, "There are %d distinct timetables -\n", bests);
      mvprintw (18,5, "How many would you like to print ?   ");
      clrtoeol();
      refresh();
      getstr(p);
      newpen = atoi (p);

      if (newpen > (bests)){
	refresh();
        mvprintw (21, 5, "There are only %d distinct timetables\n", bests);
	mvprintw (22, 5, "      Please enter another number\n");
	refresh();
      }
      else{
	ok = 1;
	move(17,5);
	clrtoeol();
	move(18,5);
	clrtoeol();
	move(24,5);
	clrtoeol();
	move(25,5);
	clrtoeol();
	refresh();
      }
    }
  return(newpen);
}    


void cant_print_agent(){
  mvprintw (21, 5, "There are no agents specified !");
  mvprintw (22, 5, "Please enter another option\n");
  refresh();
}
