/************************************************************************
 *                                                                      *
 *               GATT - Genetic Algorithm Time-Tabler                   *
 *                                                                      *
 *                By Dave Corne                                         *
 *                   Peter Ross (peter@aisb.ed.ac.uk),                  *
 *                   Emma Collingwood (emmac@aisb.ed.ac.uk)             *
 *                                                                      *
 ************************************************************************/

/* 
   Copyright (C) 1996, Dave Corne, Peter Ross and Emma Collingwood

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. 
*/




/**********************************************************************
 *  Filename: tt0.c 
 * 
 *  Functions for mutating the timetable in various
 *  ways, checking for constraint violations, and evaluating the
 *  quality of a timetable. 
 *
 **********************************************************************/ 

#include "gatt.h"
#include <stdio.h>
#include "tt0.h"
extern FILE *fopen ();
int intcompare();
void eval_consecutives();
extern int maxconsecutive;   /* max consecutive exams */  
extern int *tnscs;
extern int *tnrcs;
extern int *tnacs;
extern int studentoomanypen;

extern int nsimgroups;
extern int geno_size;
extern int chromoformat;
extern int nummods;
extern int numdays;
extern int numedges;
extern int numjuxts;
extern int numchromo;
extern int maxns;
extern int slotseatconstraints;
extern int sloteventconstraints;
extern int slotseatmax;
extern int sloteventmax;
extern int studmodsperday;
extern int abs ();
extern int chromo_id;
extern int *evals;
extern int *checks;
extern FILE *output;
extern char *strcpy ();
extern char *strcat ();
extern int *modprefs;
extern int getname ();
extern int twins;
extern int pmergep;
extern int days;
extern int efm;
extern long random ();
extern double dsm_rate;
extern double mute_rate;
extern double diversity_measure;
extern double efm_pressure;

extern char *agentslotexcs;
extern char *modagentexcs;
extern char *modslotexcs;
extern char *modroomexcs;
extern int maxpref;
extern int adaptive_efm;
extern int lazy_eval;
extern int numtocheck;
/* extern int tnchoose_size;*/
int tnchange_size ();
int tnrchange_size ();
int tnachange_size ();

int tnchange();
int tnrchange();
int tnachange();
int sim_tnchange();

extern int (*dmchoose) ();
extern int (*dmchange) ();

int readline ();
void eval_students();
void init_eval_agent ();
void eval_agent ();
void init_eval_slot ();
void eval_slot ();
void init_eval_room ();
void eval_room ();
double init_eval_tt ();
double eval_tt ();
double eval_h ();

void check_juxt_constraint ();
void check_edge_violation ();
extern double sqrt ();
extern int roomsmatter;
extern int agentsmatter;
extern int alminsbetweenmods;
extern int aldaysbetweenmods;
extern int eval_nmodsperday;
extern int npdpen;
extern int consecpen;
extern int clashpen;
extern int nclashpen;
extern int juxtpen;
extern int tooclosepen;
extern int ntooclosepen;
extern int modexcslotpen;
extern int modavslotpen;
extern int ssloadpen;
extern int seloadpen;
extern int eventcappen;
extern int studcappen;
extern int modexcroompen;
extern int modavroompen;
extern int agentexcslotpen;
extern int agentavslotpen;
extern int agentoverhpen;
extern int agentunderhpen;
extern int agentoverepen;
extern int agentunderepen;
extern int agentexcmodpen;
extern int agentavmodpen;
extern int agentprefpen;
extern int agentclashpen;
extern int agenttoomanypen;
extern int agentmodsperday;
extern double drandom ();
extern int clashperexam_pen;

extern void create_chromo ();
extern CHROMOSOME copychromo ();
extern void copy_genotype ();
extern void copy_loads ();
extern void fix_for_sim_groups ();

CHROMOSOME slot_mutate ();
CHROMOSOME sa_slot_mutate ();
CHROMOSOME room_mutate ();
CHROMOSOME agent_mutate ();
CHROMOSOME repair_clash ();
CHROMOSOME repair_equal ();
/* CHROMOSOME switch_juxt();*/

CHROMOSOME sa_mutate (genotype, parent, pop)
     char *genotype;
     CHROMOSOME *parent;
     int pop;
{
  CHROMOSOME newchromo;
  
  if (!(roomsmatter) && !(agentsmatter))
    newchromo = sa_slot_mutate (genotype, parent, pop);
  return (newchromo);
}

CHROMOSOME dsm_mutate (genotype, parent, pop)
     char *genotype;
     CHROMOSOME *parent;
     int pop;
{
  CHROMOSOME newchromo;
  int flip;

  if (drandom ()< mute_rate)
    {
      newchromo = repair_clash (genotype, parent, pop);
      return (newchromo);
    }

  if ( ((nrunspecs==0)||(!(roomsmatter))) && !(agentsmatter))
    newchromo = slot_mutate (genotype, parent, pop);
  else if ((roomsmatter) && (agentsmatter))
    {
      flip = random ()% 3;
      if (flip == 0)
	newchromo = slot_mutate (genotype, parent, pop);
      else if (flip == 1)
	newchromo = room_mutate (genotype, parent, pop);
      else
	newchromo = agent_mutate (genotype, parent, pop);
    }
  else if (roomsmatter)
    {
      flip = random ()% 2;
      if (flip == 0)
	{
	  newchromo = slot_mutate (genotype, parent, pop);
	  /*  findbestchromandcheck(0,newchromo); */
	}
      else if (flip == 1)
	{
	  newchromo = room_mutate (genotype, parent, pop);
	  /*    findbestchromandcheck(1,newchromo); */
	}
    }
  else
    {
      flip = random ()% 2;
      if (flip == 0)
	newchromo = slot_mutate (genotype, parent, pop);
      else if (flip == 1)
	newchromo = agent_mutate (genotype, parent, pop);
    }

  return (newchromo);
}


CHROMOSOME mutate_tt0 (original, rate)
     CHROMOSOME original;
     double rate;
{
  register int place;
  CHROMOSOME newchromo;
  nschanges = nrchanges = nachanges = 0;


  create_chromo (&newchromo);	/* First, create this child  */
  copy_genotype (&original, &newchromo);	/* copy the genotype across */
  copy_loads (&original, &newchromo);

  for (place = 0; place < nummods; place++)
    {
      if (a_modules[place]->slotspec < 0)
	if (drandom ()<= rate)
	  {
	    newchromo.gdp->genotype[place] =
	      (original.gdp->genotype[place] + 1 + (random ()% (numslots - 1)))
	      % numslots;
	    schanges[nschanges++] = place;
	    schanged[place] = 1;
	  }
	else
	  schanged[place] = 0;
    }
  if (roomsmatter)
    for (place = 0; place < nummods; place++)
      {
	if (a_modules[place]->roomspec < 0)
	  if (drandom ()<= rate)
	    {
	      newchromo.gdp->genotype[nummods + place] =
		(original.gdp->genotype[nummods + place] + 1 + (random ()% (numrooms - 1)))
		% numrooms;
	      rchanges[nrchanges++] = place;
	      rchanged[place] = 1;
	    }
	  else
	    rchanged[place] = 0;
      }
  if (agentsmatter)
    for (place = 0; place < nummods; place++)
      {
	if (a_modules[place]->agentspec < 0)
	  if (drandom ()<= rate)
	    {
	      newchromo.gdp->genotype[nummods + (nummods * roomsmatter) + place] =
		(original.gdp->genotype[nummods + (nummods * roomsmatter) + place] + 1 +
		 (random ()% (numagents - 1))) % numagents;
	      achanges[nachanges++] = place;
	      achanged[place] = 1;
	    }
	  else
	    achanged[place] = 0;
      }
  return (newchromo);
}


CHROMOSOME agent_mutate (genotype, parent, pop)
     char *genotype;
     CHROMOSOME *parent;
     int pop;
{
  int i,changer, newagent;
  CHROMOSOME newchromo;

  nschanges = nrchanges = nachanges = 0;
  for (i = 0; i < nummods; i++)
    achanged[i] = 0;

  if (naunspecs == 0)
    {
      newchromo = copychromo (parent);	/* If all agents specified, return */
      return (newchromo);
    }
  create_chromo (&newchromo);	/* First, create this child  */
  copy_genotype (parent, &newchromo);	/* copy the genotype across */
  copy_loads (parent, &newchromo);

  /* Now randomly pick an unspecified-agent event whose agent to mutate */
  changer = aunspecs[random ()% naunspecs];

  /* Now pick where it goes */
  newagent = tnachange (genotype, changer, parent, pop);

  /* Now put it there */
  newchromo.gdp->genotype[nummods +(roomsmatter*nummods)+ changer] = newagent;
  achanges[nachanges++] = changer;
  achanged[changer] = 1;

  /* Finally, evaluate this new chromosome. We can send it to eval_tt */
  /* which will just consider the changes between it and its parent */
  newchromo.fitness =
    eval_tt (newchromo.gdp->genotype, parent->gdp->genotype,
	     &newchromo, parent, pop);
  return (newchromo);
}


CHROMOSOME room_mutate (genotype, parent, pop)
     char *genotype;
     CHROMOSOME *parent;
     int pop;
{
  int i,changer, newroom;
  CHROMOSOME a_chromo;

  nschanges = nrchanges = nachanges = 0;
  for (i = 0; i < nummods; i++)
    rchanged[i] = schanged[i] = 0;

  if (nrunspecs == 0)
    {				/* If no mutations allowed, go */
      a_chromo = copychromo (parent);
      return (a_chromo);
    }

  create_chromo (&a_chromo);	/* First, create this child  */
  copy_genotype (parent, &a_chromo);	/* copy the genotype across */
  copy_loads (parent, &a_chromo);	/* copy other things across */

  /* Now randomly pick an unspecified-room event whose room to mutate */
  changer = runspecs[random ()% nrunspecs];

  /* Now pick where it goes */
  newroom = tnrchange (genotype, changer, parent, pop);

  /* Now put it there */
  a_chromo.gdp->genotype[nummods + changer] = newroom;
 
  rchanges[nrchanges++] = changer;
  rchanged[changer] = 1;

  /* Finally, evaluate this new chromosome. We can send it to eval_tt */
  /* which will just consider the changes between it and its parent */

  a_chromo.fitness =
    eval_tt (a_chromo.gdp->genotype, parent->gdp->genotype,
	     &a_chromo, parent, pop);
  return (a_chromo);
}

/* CHROMOSOME   repair_juxt(genotype, parent, pop)
   char         *genotype;
   CHROMOSOME   *parent;
   int          pop;
   {
   int        i, s, cl, tc, viol, choice, changer,newslot,e1,e2,t1,t2,ns;
   int        nves;

   CHROMOSOME newchromo;
   JUXT       *j, *js[MAXMODS];
   nschanges=nrchanges=nachanges=0;
   for(i=0;i<nummods;i++) schanged[i]=0;
   if(roomsmatter) for(i=0;i<nummods;i++) rchanged[i]=0;

   if(nsunspecs==0)
   {newchromo = copychromo(parent);
   return(newchromo);
   }

   create_chromo(&newchromo);
   copy_genotype(parent,&newchromo);
   copy_loads(parent,&newchromo);


CHROMOSOME   switch_juxt(genotype, parent, pop)
char         *genotype;
CHROMOSOME   *parent;
int          pop;
{
  int        i, s, cl, tc, viol, choice, changer,newslot,e1,e2,t1,t2,ns;
  int        nves;

  CHROMOSOME newchromo;
  JUXT       *j, *js[MAXMODS];
  nschanges=nrchanges=nachanges=0;
  for(i=0;i<nummods;i++) schanged[i]=0;
  if(roomsmatter) for(i=0;i<nummods;i++) rchanged[i]=0;

  if(nsunspecs==0)
      {newchromo = copychromo(parent);
      return(newchromo);
      }

  create_chromo(&newchromo);
  copy_genotype(parent,&newchromo);
  copy_loads(parent,&newchromo);


  nvjs=0;
  for(j=juxts;j;j=j->next) {
            e1 = j->a->id;
            e2 = j->b->id;
            check_juxt_constraint(j,genotype[e1],genotype[e2],&viol);
           if(viol) js[nvjs++]=j;
       }

   if(nves>0)
    { j = js[random()%nves];

       e1 = j->a->id;
       e2 = j->b->id;

     choice = genotype[e1];
     newchromo.gdp->genotype[e1]=genotype[e2];
     newchromo.gdp->genotype[e2]=choice;
     schanges[nschanges++]=e1;
     schanges[nschanges++]=e2;
     schanged[e1]=schanged[e2]=1;
     if(roomsmatter)
      {rchanges[nrchanges++]=e1;
       rchanges[nrchanges++]=e2;
      rchanged[e1]=rchanged[e2]=1;}
  }
     newchromo.fitness =
     eval_tt(newchromo.gdp->genotype,parent->gdp->genotype,
             &newchromo,parent,pop);
     return (newchromo);

}
*/

CHROMOSOME repair_equal (genotype, parent, pop)
     char *genotype;
     CHROMOSOME *parent;
     int pop;
{
  int i, viol,changer, newslot, e1, e2;
  int nves;

  CHROMOSOME newchromo;
  JUXT *j, *js[MAXMODS];

  nschanges = nrchanges = nachanges = 0;
  for (i = 0; i < nummods; i++)
    schanged[i] = 0;
  if (roomsmatter)
    for (i = 0; i < nummods; i++)
      rchanged[i] = 0;

  if (nsunspecs == 0)		/* If no slots are mutatable anyway,don't*/
    {
      newchromo = copychromo (parent);
      return (newchromo);
    }

  create_chromo (&newchromo);	/* First, create this child */
  copy_genotype (parent, &newchromo);	/* copy the genotype across */
  copy_loads (parent, &newchromo);

  /* First, see if any equal-time constraints are violated */

  nves = 0;
  for (j = juxts; j; j = j->next)
    {
      if (j->rel == 3)		/* Same time */
	{
	  e1 = j->a->id;
	  e2 = j->b->id;
	  check_juxt_constraint (j, genotype[e1], genotype[e2], &viol);
	  if (viol)
	    js[nves++] = j;
	}
    }

  /* Now, nve same-time problems are identified, and the violated */
  /* juxts are in js. */

  if (nves > 0)
    j = js[random ()% nves];
  else
    for (i = 0, j = juxts; i < random ()% numjuxts; i++)
      j = j->next;

  e1 = j->a->id;
  e2 = j->b->id;

  /* either give e1's time to e2 or vice versa */

  if (drandom ()< 0.5)
    {
      changer = e2;
      newslot = genotype[e1];
    }
  else
    {
      changer = e1;
      newslot = genotype[e2];
    }

  if (nves > 0)
    newchromo.gdp->genotype[changer] = newslot;
  schanges[nschanges++] = changer;
  schanged[changer] = 1;
  if (roomsmatter)
    {
      rchanges[nrchanges++] = changer;
      rchanged[changer] = 1;
    }

  /* Finally, evaluate this new chromosome. We can send it to eval_tt */
  /* which will just consider the changes between it and its parent   */

  newchromo.fitness =
    eval_tt (newchromo.gdp->genotype, parent->gdp->genotype,
	     &newchromo, parent, pop);
  return (newchromo);

}

CHROMOSOME repair_clash (genotype, parent, pop)
     char *genotype;
     CHROMOSOME *parent;
     int pop;
{
  int i, cl, tc, changer, newslot, e1, e2, ns;
  int p1[MAXSLOTS], p2[MAXSLOTS], np1, np2, bad, nevs;

  CHROMOSOME newchromo;
  EDGE *e, *evs[MAXMODS * MAXMODS / 2];
  LL_EDGES *lle;
  nschanges = nrchanges = nachanges = 0;
  for (i = 0; i < nummods; i++)
    schanged[i] = 0;
  if (roomsmatter)
    for (i = 0; i < nummods; i++)
      rchanged[i] = 0;

  if (nsunspecs == 0)
    {				/* If no slots are mutatable anyway,don't*/
      newchromo = copychromo (parent);
      return (newchromo);
    }

  create_chromo (&newchromo);	/* First, create this child */
  copy_genotype (parent, &newchromo);	/* copy the genotype across */
  copy_loads (parent, &newchromo);

  nevs = 0;
  for (e = edges; e; e = e->next)
    {
      e1 = e->mod1->id;
      e2 = e->mod2->id;
      ns = e->ns;
      check_edge_violation (e1, e2, genotype[e1], genotype[e2], &cl, &tc);
      if (cl)			/* If these modules clash */
	evs[nevs++] = e;
    }

  /* choose a bad edge */

  if (nevs > 0)
    e = evs[random ()% nevs];
  else
    {
      for (i = 0, e = edges; i < random ()% numedges; i++)
	e = e->next;
    }

  e1 = e->mod1->id;
  e2 = e->mod2->id;

  /* Try to find a new place for e1 */
  np1 = np2 = 0;
  for (i = 0; i < numslots; i++)
    {
      bad = 0;
      for (lle = a_modules[e1]->edges; lle; lle = lle->next)
	{
	  if (lle->e->mod1->id == e1)
	    {
	      if (genotype[lle->e->mod2->id] == i)
		bad = 1;
	      break;
	    }
	  else
	    {
	      if (genotype[lle->e->mod1->id] == i)
		bad = 1;
	      break;
	    }
	}
      if (bad == 0)
	p1[np1++] = i;
      bad = 0;
    }

  /* Now see what possibilities there are for e2 */

  for (i = 0; i < numslots; i++)
    {
      bad = 0;
      for (lle = a_modules[e2]->edges; lle; lle = lle->next)
	{
	  if (lle->e->mod1->id == e1)
	    {
	      if (genotype[lle->e->mod2->id] == i)
		bad = 1;
	      break;
	    }
	  else
	    {
	      if (genotype[lle->e->mod1->id] == i)
		bad = 1;
	      break;
	    }
	}
      if (bad == 0)
	p2[np2++] = i;
      bad = 0;
    }

  if ((np1 > 0) && (np2 > 0))
    {
      if (drandom ()< 0.5)
	changer = e1;
      else
	changer = e2;
      if (changer == e1)
	newslot = p1[random ()% np1];
      else
	newslot = p2[random ()% np2];
    }
  else if (np1 > 0)
    {
      changer = e1;
      newslot = p1[random ()% np1];
    }
  else if (np2 > 0)
    {
      changer = e2;
      newslot = p2[random ()% np2];
    }
  else if (nevs == 0)
    {
      changer = random ()% nummods;
      newslot = random ()% numslots;
    }
  else
    /* there are clashes, but not simply solved. So tn_move */
    /* one of the clashers */
    {
      if (drandom ()< 0.5)
	changer = e1;
      else
	changer = e2;
      newslot = tnchange (genotype, changer, parent, pop);
    }

  newchromo.gdp->genotype[changer] = newslot;
  schanges[nschanges++] = changer;
  schanged[changer] = 1;

  if (roomsmatter)
    {
      rchanges[nrchanges++] = changer;
      rchanged[changer] = 1;
    }

  /* Finally, evaluate this new chromosome. We can send it to eval_tt */
  /* which will just consider the changes between it and its parent   */

  newchromo.fitness =
    eval_tt (newchromo.gdp->genotype, parent->gdp->genotype,
	     &newchromo, parent, pop);
  return (newchromo);

}

CHROMOSOME sa_slot_mutate (genotype, parent, pop)
     char *genotype;
     CHROMOSOME *parent;
     int pop;
{
  int i, changer, newslot;
  int mutating_a_sim_group = 0, group;
  CHROMOSOME newchromo;

  nschanges = nrchanges = nachanges = 0;
  for (i = 0; i < nummods; i++)
    schanged[i] = 0;
  if (roomsmatter)
    for (i = 0; i < nummods; i++)
      rchanged[i] = 0;

  if (nsunspecs == 0)		/* If no slots are mutatable anyway,don't*/
    {
      newchromo = copychromo (parent);
      return (newchromo);
    }

  create_chromo (&newchromo);	/* First, create this child */
  copy_genotype (parent, &newchromo);	/* copy the genotype across */
  copy_loads (parent, &newchromo);

  /* Now randomly pick an unspecified-slot event whose slot to mutate */
  changer = sunspecs[random ()% nsunspecs];

  /* Now pick where it goes */
  if (a_modules[changer]->sim > 0)	/* it's in a sim group */
    {				/* Make main changer the base of the group */
      group = a_modules[changer]->simgroup;
      changer = simgroups[group]->sims[0];
      mutating_a_sim_group = 1;
    }

  newslot = random ()% numslots;
  while (modslotexcs[nummods * newslot + changer] == 1)
    newslot = random ()% numslots;

  if (mutating_a_sim_group == 0)
    {

      newchromo.gdp->genotype[changer] = newslot;
      schanges[nschanges++] = changer;
      schanged[changer] = 1;
      if (roomsmatter)
	{
	  rchanges[nrchanges++] = changer;
	  rchanged[changer] = 1;
	}

      /* Finally, evaluate this new chromosome. We can send it to eval_tt */
      /* which will just consider the changes between it and its parent   */

      newchromo.fitness =
	eval_tt (newchromo.gdp->genotype, parent->gdp->genotype,
		 &newchromo, parent, pop);
      return (newchromo);
    }
  /* We only get to here if we are mutating a sim group whose */
  /* base is changer */

  newchromo.gdp->genotype[changer] = newslot;

  for (i = 0; i < simgroups[group]->num; i++)
    {
      newchromo.gdp->genotype[simgroups[group]->sims[i]] = newslot;
      schanges[nschanges++] = simgroups[group]->sims[i];
      schanged[simgroups[group]->sims[i]] = 1;
      if (roomsmatter)
	{
	  rchanges[nrchanges++] = simgroups[group]->sims[i];;
	  rchanged[simgroups[group]->sims[i]] = 1;
	}
    }
  /* Finally, evaluate this new chromosome. We can send it to eval_tt */
  /* which will just consider the changes between it and its parent   */

  newchromo.fitness =
    eval_tt (newchromo.gdp->genotype, parent->gdp->genotype,
	     &newchromo, parent, pop);
  return (newchromo);
}


CHROMOSOME slot_mutate (genotype, parent, pop)
     char *genotype;
     CHROMOSOME *parent;
     int pop;
{
  int i, changer, newslot;
  int mutating_a_sim_group = 0, group;
  CHROMOSOME newchromo;


  nschanges = nrchanges = nachanges = 0;
  for (i = 0; i < nummods; i++)
    schanged[i] = 0;
  if (roomsmatter)
    for (i = 0; i < nummods; i++)
      rchanged[i] = 0;

  if (nsunspecs == 0)		/* If no slots are mutatable anyway,don't*/
    {
      newchromo = copychromo (parent);
      return (newchromo);
    }

  create_chromo (&newchromo);	/* First, create this child */
  copy_genotype (parent, &newchromo);	/* copy the genotype across */
  copy_loads (parent, &newchromo);

  /* Now randomly pick an unspecified-slot event whose slot to mutate */
  changer = sunspecs[random ()% nsunspecs];

  /* Now pick where it goes */
  if (a_modules[changer]->sim > 0)	/* it's in a sim group */
    {				/* Make main changer the base of the group */
      group = a_modules[changer]->simgroup;
      changer = simgroups[group]->sims[0];
      mutating_a_sim_group = 1;
    }

  if (mutating_a_sim_group == 0)
    {
      newslot = tnchange (genotype, changer, parent, pop);

      newchromo.gdp->genotype[changer] = newslot;
      schanges[nschanges++] = changer;
      schanged[changer] = 1;
      if (roomsmatter)
	{
	  rchanges[nrchanges++] = changer;
	  rchanged[changer] = 1;
	}

      /* Finally, evaluate this new chromosome. We can send it to eval_tt */
      /* which will just consider the changes between it and its parent   */

      newchromo.fitness =
	eval_tt (newchromo.gdp->genotype, parent->gdp->genotype,
		 &newchromo, parent, pop);
      return (newchromo);
    }
  /* We only get to here if we are mutating a sim group whose */
  /* base is changer */

  newslot = sim_tnchange (genotype, changer, group, parent, pop);

  newchromo.gdp->genotype[changer] = newslot;

  for (i = 0; i < simgroups[group]->num; i++)
    {
      newchromo.gdp->genotype[simgroups[group]->sims[i]] = newslot;
      schanges[nschanges++] = simgroups[group]->sims[i];
      schanged[simgroups[group]->sims[i]] = 1;
      if (roomsmatter)
	{
	  rchanges[nrchanges++] = simgroups[group]->sims[i];;
	  rchanged[simgroups[group]->sims[i]] = 1;
	}
    }
  /* Finally, evaluate this new chromosome. We can send it to eval_tt */
  /* which will just consider the changes between it and its parent   */

  newchromo.fitness =
    eval_tt (newchromo.gdp->genotype, parent->gdp->genotype,
	     &newchromo, parent, pop);
  return (newchromo);
}




/*********************************************************************/
/**       Evaluate the initial random timetables                    **/
/*********************************************************************/

double init_eval_tt (genotype, oldgeno, chromo, oldchromo, pop)
     char *genotype, *oldgeno;
     CHROMOSOME *chromo, *oldchromo;
     int pop;
{
  int totalpen = 0;

  evals[pop]++;
  chromo->student_over = 0;
  
  init_eval_slot (genotype, chromo);
  if (maxconsecutive > 0)
    eval_consecutives(genotype, chromo);
  
  /* work out number of exams each student has per day
   * if this is a constraint */

  if (studmodsperday > 0)
    eval_students(genotype, chromo);
  
 /* Add up violations: */
  totalpen += (chromo->clashes * clashpen +	/* clashes,           */
	       chromo->nclash * nclashpen +	/* N clashes          */
	       chromo->toocloses * tooclosepen +/* too-closes         */
	       chromo->ntooclose * ntooclosepen + /* N too-closes       */
	       chromo->modexcslots * modexcslotpen + /* excluded slots     */
	       chromo->modavslots * modavslotpen + /* avoided slots      */
	       chromo->juxtscore * juxtpen) +	/* juxtapositions     */
    chromo->student_over * studentoomanypen +   /* too many exams per day */
    chromo->nclash_exams * clashperexam_pen +
    chromo->consecutives * consecpen;
  
  
  if (slotseatconstraints)
    totalpen += chromo->ssload * ssloadpen;
  if (sloteventconstraints)
    totalpen += chromo->seload * seloadpen;

  if (eval_nmodsperday)
    totalpen += chromo->nperday * npdpen;

  if (roomsmatter)
    {
      init_eval_room (genotype, chromo);

      totalpen += (chromo->eventcapviols * eventcappen +	/* event capacities */
		   chromo->studcapviols * studcappen +	/* student capac'es */
		   chromo->modexcrooms * modexcroompen +	/* excluded rooms   */
		   chromo->modavrooms * modavroompen);	/* avoided rooms    */


    }
  if (agentsmatter)
    {

      init_eval_agent (genotype, chromo);

      totalpen += (chromo->agentexcslots * agentexcslotpen +  /* excluded slots */
		   chromo->agentavslots * agentavslotpen +	/* avoided slots  */
		   chromo->agentoverh * agentoverhpen +	/* agent overtime */
		   chromo->agentunderh * agentunderhpen +	/* agent undertime */
		   chromo->agentovere * agentoverepen +	/* agent toomanye */
		   chromo->agentundere * agentunderepen +	/* agent undere   */
		   chromo->agentexcmod * agentexcmodpen +	/* don't dos */
		   chromo->agentavmod * agentavmodpen +	/* can do's  */
		   chromo->agentprefs * agentprefpen +	/* preferences */
		   chromo->agentclash *agentclashpen + /* clashes */
		   chromo->agenttoomany * agenttoomanypen);
  }
  return (1.0 / (1.0 + (double) totalpen));	/* return total fitness */
}

void eval_consecutives(genotype, chromo)
  char *genotype;
  CHROMOSOME *chromo;
{

  _STUDENT *thisStudent;
  LL_MODS *thisMod;
  int i,j,n,day,student_id,index;
  int pos;
  int **usesSlots;
  int *studTotal;
  
  studTotal = (int *)calloc(numstudents, sizeof(int));
  usesSlots = calloc(numstudents,sizeof(int *));

  for (i=0;i<numstudents;i++)
    usesSlots[i] = (int *)calloc(numslots, sizeof(int));
  
  
  /* for each students, go through exam lists and work out how many exams
   * per day - if number is greater than modsperday than increment counter
   */
  
  chromo->consecutives = 0;
 
  thisStudent = students;
  for(thisStudent;thisStudent!=NULL;thisStudent=thisStudent->next){
    thisMod = thisStudent->mods;
    student_id = thisStudent->id;
    studTotal[student_id] = 0;
    for (thisMod;thisMod != NULL;thisMod = thisMod->next){
      usesSlots[student_id][studTotal[student_id]++] =
        genotype[thisMod->mod->id];
    }
    
  }

  /* now sort the arrays into ascending order of slots */
  for (i=0;i<numstudents;i++)
    qsort((char *)usesSlots[i],studTotal[i], sizeof(int),intcompare);

  /* now calculate biggest consecutive group for each student */
  for (i=0;i<numstudents;i++){
    index = 0;
    pos = 0;
    while (pos < studTotal[i]){
      while (usesSlots[i][pos+1] == usesSlots[i][pos]+1){
        index++;
        pos++;
      }
      if (index >= maxconsecutive)
        chromo->consecutives++;
      index = 0;
      pos++;
    }
    
  }

  
  free(studTotal);
  for (i=0;i<numstudents;i++)
    free(usesSlots[i]);
  free(usesSlots);
  
  
}


void eval_students(genotype, chromo)
  char *genotype;
  CHROMOSOME *chromo;
{

  _STUDENT *thisStudent;
  LL_MODS *thisMod;
  int i,j,n,day,student_id;
  int exams_per_day[MAXSTUDENTS][MAXDAYS];
  
  
  /* initialise */
  n = 0;
  for (i=0;i<numstudents;i++)
    for (j=0;j<numdays;j++)
      exams_per_day[i][j] = 0;
  
  chromo->student_over = 0;
  
  
  /* for each students, go through exam lists and work out how many exams
   * per day - if number is greater than modsperday than increment counter
   */

  
  thisStudent = students;
  for(thisStudent;thisStudent!=NULL;thisStudent=thisStudent->next){
    thisMod = thisStudent->mods;
    student_id = thisStudent->id;
    for (thisMod;thisMod != NULL;thisMod = thisMod->next){
      i = genotype[thisMod->mod->id];
      day = slots[i]->day;
      exams_per_day[student_id][day]++;
    }
  }

  /* now check how many students have too many per day */
  for (i=0;i<numstudents;i++)
    for (j=0;j<numdays;j++)
      if (exams_per_day[i][j] > studmodsperday)
        n++;
  chromo->student_over = n;

  
}


     
double eval_tt (genotype, oldgeno, chromo, oldchromo, pop)
     char *genotype, *oldgeno;
     CHROMOSOME *chromo, *oldchromo;
     int pop;
{
  int totalpen = 0;

  evals[pop]++;

  eval_slot (genotype, oldgeno, chromo, oldchromo, pop);
  if (maxconsecutive > 0)
    eval_consecutives(genotype,chromo);
  
  /* work out number of exams each student has per day
   * if this is a constraint */

  if (studmodsperday > 0)
    eval_students(genotype, chromo);
  
  /* Add up violations: */
  totalpen += (chromo->clashes * clashpen +	/* clashes,       */
	       chromo->toocloses * tooclosepen +	/* too-close's    */
	       chromo->modexcslots * modexcslotpen +	/* excluded slots */
	       chromo->modavslots * modavslotpen +	/* avoided slots  */
	       chromo->juxtscore * juxtpen) +	/* juxtapositions */
    chromo->student_over * studentoomanypen  +   /* too many exams per day */
    chromo->nclash_exams * clashperexam_pen +
    chromo->consecutives * consecpen;
  
  if (slotseatconstraints)
    totalpen += chromo->ssload * ssloadpen;
  if (sloteventconstraints)
    totalpen += chromo->seload * seloadpen;

  if (eval_nmodsperday)
    totalpen += chromo->nperday * npdpen;

  if (roomsmatter)
    {
      eval_room (genotype, oldgeno, chromo, oldchromo);

      totalpen += (chromo->eventcapviols * eventcappen +	/* event capacities */
		   chromo->studcapviols * studcappen +	/* student capac'es */
		   chromo->modexcrooms * modexcroompen +	/* excluded rooms   */
		   chromo->modavrooms * modavroompen);	/* avoided rooms    */


    }
  if (agentsmatter)
    {

      eval_agent (genotype, oldgeno, chromo, oldchromo);

      totalpen += (chromo->agentexcslots * agentexcslotpen +	/* excluded slots */
		   chromo->agentavslots * agentavslotpen +	/* avoided slots  */
		   chromo->agentoverh * agentoverhpen +	/* agent overtime */
		   chromo->agentunderh * agentunderhpen +	/* agent undertime */
		   chromo->agentovere * agentoverepen +	/* agent toomanye */
		   chromo->agentundere * agentunderepen +	/* agent undere   */
		   chromo->agentexcmod * agentexcmodpen +	/* don't dos */
		   chromo->agentavmod * agentavmodpen +	/* can do's  */
		   chromo->agentprefs * agentprefpen +	/* preferences */
	           chromo->agentclash * agentclashpen + /* clashes */
		   chromo->agenttoomany * agenttoomanypen); /* too many */
 }
  return (1.0 / (1.0 + (double) totalpen));	/* return total fitness */
}


void 
init_eval_agent (genotype, chromo)
     char *genotype;
     CHROMOSOME *chromo;
{
  register int i, j, s, a, l, v;
  int day, agentclash = 0;
  int slot_agent[MAXSLOTS][MAXAGENTS];
  int agent_per_day[MAXAGENTS][MAXDAYS];

  chromo->agentexcslots = chromo->agentavslots = chromo->agentoverh =
    chromo->agentunderh = chromo->agentovere = chromo->agentundere =
    chromo->agentexcmod = chromo->agentavmod = chromo->agentprefs
    = chromo->agentclash = chromo->agenttoomany = 0;

  /* Initialise agent loads */

  for (i = 0; i < numagents; i++)
    {
      chromo->aeloads[i] = 0;
      chromo->ahloads[i] = 0;
      for (j = 0;j< numdays;j++)
	agent_per_day[i][j] = 0;
    }
  /* Now, look through this chromosome and suss out the agents' teaching
   loads. This is aeloads and ahloads (events and hours). */

  for (i = 0; i < nummods; i++)
    {
      a = genotype[nummods + (roomsmatter * nummods) + i];	/* The agent teaching */
      /* (or whatever) event i */
      l = a_modules[i]->len;	/* i's length */
      s = genotype[i];		/* it's slot */
      chromo->aeloads[a]++;
      chromo->ahloads[a] += l;

      /* now work out what day this slot is on and increment day counter
       * for that module */
      day = slots[s]->day;
      agent_per_day[a][day]++;


      if ((v = (modagentexcs[nummods * a + i]) > 0))
	if (v == 1)
	  chromo->agentexcmod++;
	else
	  chromo->agentavmod++;

      /* agent slot exclusions ? */
      
      if ((v = (agentslotexcs[numagents * s + a]) > 0))
	if (v == 1)
	  chromo->agentexcslots++;
	else
	  chromo->agentavslots++;

      /* Now for agent preferences. This is a funny old thing. Penalise */
      /* when an agent is doing something they'd rather not, to an amount */
      /* in tune with their distaste at doing it */

      /**********************************************************
       V1.6 - CHANGED THIS AS ORIGINAL CODE PENALISES A CHROMOSOME
       MORE IF AN AGENT TAKES  A MODULES WITH HIGH PREFERENCE !
       *********************************************************/
       
      for (j = 0; j < agents[a]->numcando; j++)
	if (agents[a]->cando[j]->mod == i)
	  chromo->agentprefs +=
            (agents[a]->maxpref - agents[a]->cando[j]->des);
    }

  for (i = 0; i < numagents; i++)
    {
      if (chromo->aeloads[i] > agents[i]->maxmods && agents[i]->maxmods > 0)
	chromo->agentovere += chromo->aeloads[i] - agents[i]->maxmods;
      if (chromo->aeloads[i] < agents[i]->minmods && agents[i]->minmods > 0)
	chromo->agentundere += agents[i]->minmods - chromo->aeloads[i];
      if ((chromo->ahloads[i] > agents[i]->maxhours) && (agents[i]->maxhours > 0))
	chromo->agentoverh += chromo->ahloads[i] - agents[i]->maxhours;
      if ((chromo->ahloads[i] < agents[i]->minhours) && (agents[i]->minhours > 0))
	chromo->agentunderh += agents[i]->minhours - chromo->ahloads[i];

      /* check how many slots per day an agent is associated with */
      for (j=0;j < numdays;j++)
	if (agentmodsperday > 0 && agent_per_day[i][j] > agentmodsperday)
	  chromo->agenttoomany++;

    }

  /***********************************************************
   * ADDED 23/5/97
   * Code added to calculate agent-
   clashes, ie agents that a
   * have been assigned to concurrent exams
   ***********************************************************/


  /* now go through the agents and work out if there are any
   * agent clashes, ie an agent who supervises more than 1 module
   * that occurs in the same slot
   */

  /* initialise array first */

  for (i=0;i<numslots;i++)
    for (j=0;j<numagents;j++)
      slot_agent[i][j] = 0;

  /* first go through the chromosome and find out which agents
   * teach in which slot
   */
  
  for (i=0;i<numslots;i++)
    for (j=0;j<nummods;j++)
      if (genotype[j] == i) 
	slot_agent[i][genotype[nummods + (roomsmatter * nummods) + j]]++;
	
  
  /* now any slot-agent array that has a value >1 means that agent
   * is doing something more than once in the slot */
 
  for (i=0;i<numslots;i++)
    for(j=0;j<numagents;j++)
      if (slot_agent[i][j] > 0)
	agentclash = agentclash + slot_agent[i][j]-1;

  /* finally, tell the chromsome how many clashes */
  chromo->agentclash = agentclash;
  
}


void eval_agent (genotype, oldgeno, chromo, oldchromo, pop)
     char *genotype, *oldgeno;
     CHROMOSOME *chromo, *oldchromo;
     int pop;
{
  register int i, j;
  int  e, a, oa,day,s,v;
  int  agentclash = 0;
  int slot_agent[MAXSLOTS][MAXAGENTS];
  int agent_per_day[MAXAGENTS][MAXDAYS];

  /* Remember to set everything to zero */
 
  /***************************************************************
   * 22/5/96
   * added code to initialise all agent related variables to 0
   ***************************************************************/
  
  chromo->agentexcslots = chromo->agentavslots = chromo->agentoverh =
    chromo->agentunderh = chromo->agentovere = chromo->agentundere =
    chromo->agentexcmod = chromo->agentavmod = chromo->agentprefs 
    = chromo->agentclash = chromo->agenttoomany = 0;
 
  /* initialise array */
  for (i=0;i<numagents;i++)
    for (j=0;j<numdays;j++)
      agent_per_day[i][j] = 0;

  for (i = 0; i < nachanges; i++)
    {
      e = achanges[i];
      a = genotype[nummods + (roomsmatter * nummods) + e];
      oa = oldgeno[nummods + (roomsmatter * nummods) + e];

      /* First see if this affects the agent's loads and hours */

      chromo->aeloads[a]++;	/* Augment loads for */
      chromo->ahloads[a] += a_modules[e]->len;	/* the new agent */

      chromo->aeloads[oa]--;	/* Decrement for the */
      chromo->ahloads[oa] -= a_modules[e]->len;	/* old agent */

    }


  /* and now work out agent loads per day and exclusions*/
  for (i=0;i<nummods;i++){
    day = slots[genotype[i]]->day;
    agent_per_day[genotype[nummods + (roomsmatter * nummods) + i]][day]++;
    
    a =  genotype[nummods + (roomsmatter * nummods) + i];
    s = genotype[i];

    /* preferences */
    for (j = 0; j < agents[a]->numcando; j++)
      if (agents[a]->cando[j]->mod == i)
        chromo->agentprefs +=
          (agents[a]->maxpref - agents[a]->cando[j]->des);
    

    /* agent slot exclusions ? */
    if ((v = (agentslotexcs[numagents * s + a]) > 0))
      if (v == 1)
	chromo->agentexcslots++;
      else
	chromo->agentavslots++;

    /********************************************************/
    /* VERSION 1.6- ADDED THIS CODE To CHECK MODS  EXCLUDED
       FOR EACH AGENT */
    if ((v = (modagentexcs[nummods * a + i]) > 0))
      if (v == 1)
        chromo->agentexcmod++;
      else
        chromo->agentavmod++;
    /* END OF ADDITIONS */
    /********************************************************/

  }
    
  /* Now gauge the overall effect */
  for (i = 0; i < numagents; i++)
    {
      if (chromo->aeloads[i] > agents[i]->maxmods && agents[i]->maxmods > 0 )
	chromo->agentovere += chromo->aeloads[i] - agents[i]->maxmods ;
      if (chromo->aeloads[i] < agents[i]->minmods && agents[i]->minmods > 0)
	chromo->agentundere += agents[i]->minmods- chromo->aeloads[i];
      if (chromo->ahloads[i] > agents[i]->maxhours && agents[i]->maxhours > 0)
	chromo->agentoverh += chromo->ahloads[i] - agents[i]->maxhours;
      if (chromo->ahloads[i] < agents[i]->minhours && agents[i]->minhours > 0 )
	chromo->agentunderh += agents[i]->minhours -chromo->ahloads[i];

      for (j=0;j < numdays;j++)
	if (agentmodsperday > 0 && agent_per_day[i][j] > agentmodsperday)
	  chromo->agenttoomany++;
      
    }

  /***********************************************************
   * ADDED 23/5/97
   * Code added to calculate agent-clashes, ie agents that a
   * have been assigned to concurrent exams
   ***********************************************************/


  /* now go through the agents and work out if there are any
   * agent clashes, ie an agent who supervises more than 1 module
   * that occurs in the same slot
   */

  /* initialise first */
  for (i=0;i<numslots;i++)
    for (j=0;j<numagents;j++)
      slot_agent[i][j] = 0;

  for (i=0;i<numslots;i++)
    for (j=0;j<nummods;j++)
      if (genotype[j] == i) 
	slot_agent[i][genotype[nummods + (roomsmatter * nummods) + j]]++;


  /* now check each slot-agent and see how many are greater than 1 */
  for (i=0;i<numslots;i++)
    for(j=0;j<numagents;j++)
      if (slot_agent[i][j] > 0)
	agentclash = agentclash+slot_agent[i][j]-1;


  chromo->agentclash +=agentclash;

  

}




/*********************************************************************/
/**       Evaluate the event/slot constraints                       **/
/*********************************************************************/

void init_eval_slot (genotype, chromo)
     char *genotype;
     CHROMOSOME *chromo;
{

  register int i, s,e1, e2,ns;
  int cl, tc, viol;
  JUXT *j;
  EDGE *e;
  _MODULE *m;
  int clash_exam[MAXMODS];
  LL_SPEC *ms;

  chromo->clashes = chromo->toocloses = chromo->modexcslots =
    chromo->modavslots = chromo->juxtscore = chromo->juxtviols =
    chromo->nperday = chromo->seload = chromo->ssload =
    chromo->nclash = chromo->ntooclose = chromo->nclash_exams = 0;


  for (i=0;i<nummods;i++)
    clash_exam[i] = 0;
  
  /* Look at every edge constraint */
  for (e = edges; e; e = e->next)
    if (e->ns > 0)
      {
	e1 = e->mod1->id;
	e2 = e->mod2->id;
	ns = e->ns;
	check_edge_violation (e1, e2, genotype[e1], genotype[e2], &cl, &tc);
	if (cl)			/* If these modules clash */
	  {
	    chromo->clashes += ns;	/* Augment clash score in chromosome */
	    chromo->nclash++;
            /* record which module was involved in clash */
            clash_exam[e1]++;
            clash_exam[e2]++; 
	  }
	if (tc)			/* If these modules are too close */
	  {
	    chromo->toocloses += ns;	/* Augment tooclose score */
	    chromo->ntooclose++;
	  }
      }

  /* Look at exclusions and avoidances */
  /****************************************
   * V1.6 - if a slot needs to be avoided and a preference is specified,
   * then multiply the penalty by the preference to avoid this slot
   ****************************************/
  for (m = modules; m; m = m->next)
    if (modslotexcs[nummods * genotype[m->id] + m->id])
      if (modslotexcs[nummods * genotype[m->id] + m->id] == 1)
	chromo->modexcslots++;
      else{
        /* V1.7 updated priority checking so that it
             checks priority of correct slot */
        /* find the priority for this slot */
        for (ms = m->exslots;ms;ms=ms->next)
          if (ms->id == genotype[m->id]){
            chromo->modavslots += (ms->priority);
            break;
          }
      }
  
  /* Now look at juxtaposition constraints */
  for (j = juxts; j; j = j->next)
    {
      e1 = j->a->id;
      e2 = j->b->id;
      check_juxt_constraint (j, genotype[e1], genotype[e2], &viol);
      if (viol)
	chromo->juxtscore += j->priority;
    }

  if (slotseatconstraints)
    {
      for (i = 0; i < nummods; i++)
	{
	  chromo->ssloads[genotype[i]] += a_modules[i]->size;
	  for (s = genotype[i] + 1; s < numslots; s++)
	    { 
	      if ((s < numslots) &&
		  (slots[genotype[i]]->day == slots[s]->day) &&
		  ((slots[genotype[i]]->time + a_modules[i]->len) >
		   slots[s]->time))
		chromo->ssloads[s] += a_modules[i]->size;
	      else
		break;
	    }
	}
      /* Now adjust chromo->ssloads to take into account any */
      /* bother caused by sims having to be in one place but */
      /* sharing students */
      for (i = 0; i < nsimgroups; i++)
	chromo->ssloads[genotype[simgroups[i]->sims[0]]] -=
	  (simgroups[i]->total - simgroups[i]->size);
      for (s = 0; s < numslots; s++)
	if (chromo->ssloads[s] > slots[s]->sscap)
	  {
	    chromo->ssload += (chromo->ssloads[s] - slots[s]->sscap);
	  }
    }
  if (sloteventconstraints)
    {
      for (i = 0; i < nummods; i++)
	{
	  chromo->seloads[genotype[i]]++;
	  for (s = genotype[i] + 1; s < numslots; s++)
	    {
	      if ((slots[genotype[i]]->time + a_modules[i]->len) >
		  slots[s]->time)
		chromo->seloads[s]++;
	      else
		break;
	    }
	}
      for (s = 0; s < numslots; s++)
	if (chromo->seloads[s] > slots[s]->secap)
	  {
	    chromo->seload++;
	  }

    }
  for (i=0;i<nummods;i++)
    if (clash_exam[i] > 0)
      chromo->nclash_exams++;
 
  
}


void eval_slot (genotype, oldgeno, chromo, oldchromo, pop)
     char *genotype, *oldgeno;
     CHROMOSOME *chromo, *oldchromo;
     int pop;
{
  register int i, k, e1, e2,ns;
  int cl, tc, viol;
  LL_EDGES *e;
  LL_JUXTS *ju;
  JUXT *jc;
  EDGE *edge;
  _MODULE *m;
  int clash_exam[MAXMODS];
  LL_SPEC *ms;
  FILE *f;
  

  chromo->clashes = chromo->toocloses = chromo->modexcslots =
    chromo->modavslots = chromo->juxtscore = chromo->juxtviols =
    chromo->nperday = chromo->ssload = chromo->seload =
    chromo->nclash = chromo->ntooclose = chromo->nclash_exams =0;

  for (i=0;i<nummods;i++)
    clash_exam[i] = 0;
  
 /* Look at every edge constraint HACK */
  for (edge = edges; edge; edge = edge->next)
    if (edge->ns > 0)
      {
	e1 = edge->mod1->id;
	e2 = edge->mod2->id;
	ns = edge->ns;
	check_edge_violation (e1, e2, genotype[e1], genotype[e2], &cl, &tc);
	if (cl)			/* If these modules clash */
	  {
            clash_exam[e1]++;
            clash_exam[e2]++; 
	  }
      }
  
  
  /*  if(slotseatconstraints) for(i=0;i<numslots;i++)chromo->ssloads[i]=0;*/
  /*   if(sloteventconstraints) for(i=0;i<numslots;i++)chromo->seloads[i]=0;*/

  /* changes[i] carries id of ith changed module. We can now just look */
  /* at the edges and juxts relating to these changed modules. Some    */
  /* bookkeeping necessary though, to make sure we don't check any of   */
  /* these edges twice. Hence the check each time. An edge of mod i has */
  /* been checked already if both of its modules are changed, and the id */
  /* of the one which isn't i is smaller than i */

  for (i = 0; i < nschanges; i++)
    {
      m = a_modules[schanges[i]];	/* m is the module pointer */
      for (e = m->edges; e; e = e->next)
	if (e->e->ns > 0)
	  {
	    edge = e->e;
	    e1 = edge->mod1->id;
	    e2 = edge->mod2->id;
	    if (((e1 < m->id) && (schanged[e1])) ||
		((e2 < m->id) && (schanged[e2])))

	      /* Then we have already looked at this edge */
	      continue;

	    ns = edge->ns;
	    /* Now deal with this constraint */
	    checks[pop]++;	/* One more constraint checked */

	    check_edge_violation (e1, e2, genotype[e1], genotype[e2], &cl, &tc);
	    if (cl)		/* If these modules clash */
	      {
		chromo->clashes += ns;	/* Augment clash score in chromosome */
		chromo->nclash++;
              }
	    if (tc)		/* If these modules are too close */
	      {
		chromo->toocloses += ns;
		chromo->ntooclose++;
	      }

	    /* We have seen what has happened to this constraint in the */
	    /* new genotype, but now need to subtract the effect of this */
	    /* constraint in the old genotype */

	    check_edge_violation (e1, e2, oldgeno[e1], oldgeno[e2], &cl, &tc);
	    if (cl)		/* If these modules clashed */
	      {
		chromo->clashes -= ns;	/* Augment clash score in chromosome */
		chromo->nclash--;
              }
	    if (tc)		/* If these modules were too close */
	      {
		chromo->toocloses -= ns;
		chromo->ntooclose--;
	      }
	  }
    }


  for (i=0;i<nummods;i++)
    if (clash_exam[i] > 0)
      chromo->nclash_exams++;
  

  /* That's the edge constraints dealt with. We just need now to */
  /* update hots, clashes and toocloses properly, later */
  /* Now add exclusion violations */

  for (i = 0; i < nschanges; i++)
    {
      m = a_modules[schanges[i]];
      if (modslotexcs[nummods * genotype[m->id] + m->id])
	if (modslotexcs[nummods * genotype[m->id] + m->id] == 1)
	  chromo->modexcslots++;
	else{
          /* V1.7 updated priority checking so that it
             checks priroty of correct slot */
          /* find the priority for this slot */
          for (ms = m->exslots;ms;ms=ms->next)
            if (ms->id == genotype[m->id]){
              chromo->modavslots += (ms->priority);
              break;
            }
        }
      
      if (modslotexcs[nummods * oldgeno[m->id] + m->id])
	if (modslotexcs[nummods * oldgeno[m->id] + m->id] == 1)
	  chromo->modexcslots--;
	else{
           /* find the priority for this slot */
          for (ms = m->exslots;ms;ms=ms->next)
            if (ms->id == oldgeno[m->id]){
              chromo->modavslots -= (ms->priority);
              break;
            }
        }
          

      /* V1.6 changed modavslot penalty so that it is in proportion to
         the weight with which we want to avoid it.*/
    }

  /* Now add up juxtaposition violations */

  for (i = 0; i < nschanges; i++)
    {
      m = a_modules[schanges[i]];
      for (ju = m->juxts; ju; ju = ju->next)
	{
	  viol = 0;
	  jc = ju->j;
	  e1 = jc->a->id;
	  e2 = jc->b->id;
	  if (((e1 < m->id) && (schanged[e1])) || ((e2 < m->id) && (schanged[e2])))
	    continue;
	  checks[pop]++;
	  check_juxt_constraint (jc, genotype[e1], genotype[e2], &viol);
	  if (viol)
	    chromo->juxtscore += jc->priority;
	  check_juxt_constraint (jc, oldgeno[e1], oldgeno[e2], &viol);
	  if (viol)
	    chromo->juxtscore -= jc->priority;
	}
    }

  if (slotseatconstraints)
    {
      for (i = 0; i < nschanges; i++)
	{
	  m = a_modules[schanges[i]];
	  e1 = m->id;
	  chromo->ssloads[genotype[e1]] += m->size;
	  chromo->ssloads[oldgeno[e1]] -= m->size;
	}

      /* Right -- if this was a sim-group, it's now been moved */
      /* we need to recover chromo->ssloads[oldgeno[e1]]  and  */
      /* chromo->ssloads[genotype[e1]] to their correct values */

      /* only if there were indeed any slot changes though....*/
      if (nschanges > 0)
	if (a_modules[e1]->sim > 0)
	  {
	    chromo->ssloads[genotype[e1]] -=
	      (simgroups[a_modules[e1]->simgroup]->total -
	       simgroups[a_modules[e1]->simgroup]->size);
	    chromo->ssloads[oldgeno[e1]] +=
	      (simgroups[a_modules[e1]->simgroup]->total -
	       simgroups[a_modules[e1]->simgroup]->size);
	  }

      for (k = 0; k < numslots; k++)
	if (chromo->ssloads[k] > slots[k]->sscap)
	  chromo->ssload += (chromo->ssloads[k] - slots[k]->sscap);
    }

  if (sloteventconstraints)
    {
      for (i = 0; i < nschanges; i++)
	{
	  m = a_modules[schanges[i]];
	  e1 = m->id;
	  chromo->seloads[genotype[e1]]++;
	  chromo->seloads[oldgeno[e1]]--;
	}
      for (k = 0; k < numslots; k++)
	if (chromo->seloads[k] > slots[k]->secap)
	  chromo->seload++;
    }
  

  chromo->clashes += oldchromo->clashes;
  chromo->toocloses += oldchromo->toocloses;
  chromo->juxtscore += oldchromo->juxtscore;
  chromo->modexcslots += oldchromo->modexcslots;
  chromo->modavslots += oldchromo->modavslots;
  chromo->nclash += oldchromo->nclash;
  chromo->ntooclose += oldchromo->ntooclose;
}


/*************************************************************************/
/*  Find problems with the timetable associated with violations of room  */
/* constraints. There are four kinds of constraints on rooms: event      */
/* loads, student loads, avoided rooms, and excluded rooms. The critical */
/* thing to do here first with regard to event loads and student loads   */
/* is to store, in the chromosome, the event loads and student loads for */
/* each room in each slot. This makes it easy to keep a check on things  */
/* incrementally via delta evaluation.  The seat loads are kept in       */
/* chromo.sloads, and the event loads in chromo.eloads. ........... */

void init_eval_room (genotype, chromo)
     char *genotype;
     CHROMOSOME *chromo;
{
  int a, d, i, j, k, l, r, s, t, v;

  chromo->eventcapviols = chromo->studcapviols =
    chromo->modexcrooms = chromo->modavrooms = 0;

  /* initialise the room loads */
  for (i=0;i<numslots*numrooms;i++){
    chromo->eloads[i] = 0;
    chromo->sloads[i] = 0;
  }


  /* Add up the capacity violations. First find the load for */
  /* each room in each slot, then compare against roomcaps.  */

  for (i = 0; i < nummods; i++)
    {
      r = genotype[nummods + i];/*  The room it's in       */
      s = genotype[i];		/*  Slot it starts in      */
      t = slots[s]->time;	/*  Actual time it starts  */
      d = slots[s]->aday;	/*  Actual day it starts   */
      l = a_modules[i]->len;	/*  Its length             */
      a = a_modules[i]->size;	/*  Its size               */
      k = 0;			/*  Extra slots moved into */

      while (((s + k) < numslots)
	     && (slots[s + k]->aday == d)	/* for slots on this day */
	     && ((t + l) > slots[s + k]->time))	/* which it moves into ... . */
	{
	  chromo->eloads[numrooms * (s + k) + r]++;	/* increment event loads */
	  chromo->sloads[numrooms * (s + k) + r] += a;	/* increment student loads */
	  k++;			/* Try the next slot */
	}
    }

  for (i = 0; i < numrooms; i++)
    for (j = 0; j < numslots; j++)
      {
	/* if roomcaps < 0, then nothing has been specified so can ignore */
	if (roomcaps[i][j][0] > 0)
	  if ((v = chromo->eloads[numrooms * j + i]) > roomcaps[i][j][0])
	    chromo->eventcapviols += (v - roomcaps[i][j][0]);
        
	if (roomcaps[i][j][1] > 0)
	  if ((v = chromo->sloads[numrooms * j + i]) > roomcaps[i][j][1])
	    chromo->studcapviols += (v - roomcaps[i][j][1]);
      }


  /* Now look at excluded rooms for each module */
  for (i = 0; i < nummods; i++)
    {
      r = genotype[nummods + i];
      if (modroomexcs[nummods * r + i])
	{
	  if (modroomexcs[nummods * r + i] == 1)
	    chromo->modexcrooms++;
	  else
	    chromo->modavrooms++;
	}
    }
}




void eval_room (genotype, oldgeno, chromo, oldchromo)
     char *genotype, *oldgeno;
     CHROMOSOME *chromo, *oldchromo;
{
  int a, d, i, j, k, l, r, s, t, or, os, ot, od, v;
  _MODULE *m;
  chromo->eventcapviols = chromo->studcapviols =
    chromo->modexcrooms = chromo->modavrooms = 0;

  for (i = 0; i < nrchanges; i++)
    {
      m = a_modules[rchanges[i]];	/* m is the module pointer */
      r = genotype[nummods + rchanges[i]];	/* its current room */
      or = oldgeno[nummods + rchanges[i]];	/* its old room */
      s = genotype[rchanges[i]];/* It's current slot */
      os = oldgeno[rchanges[i]];/* It's old slot */
      t = slots[s]->time;	/* It's start time */
      ot = slots[os]->time;	/* It's old start time */
      d = slots[s]->aday;	/* It's day */
      od = slots[os]->aday;	/* It's old day */
      l = m->len;		/*  Its length             */
      a = m->size;		/*  Its size               */
      k = 0;			/*  Extra slots moved into */


      /* Now, work out how this new room placement alters things */

      while (((s + k) < numslots)
	     && (slots[s + k]->aday == d)	/* for slots on this day */
	     && ((t + l) > slots[s + k]->time))	/* which it moves into ... */
	{
	  chromo->eloads[numrooms * (s + k) + r]++;	/* increment event loads */
	  chromo->sloads[numrooms * (s + k) + r] += a;	/* increment student loads */
	  k++;			/* Try the next slot */
	}

      /*  But subtract the effect of its previous position */
      k = 0;
      while (((os + k) < numslots)
	     && (slots[os + k]->aday == od)	/* for slots on this day */
	     && ((ot + l) > slots[os + k]->time))	/* which it moves into ... */
	{
	  chromo->eloads[numrooms * (os + k) + or]--;	/* decrement event loads */
	  chromo->sloads[numrooms * (os + k) + or] -= a;	/* decrement student loads */
	  k++;			/* Try the next slot */
	}


      /* Now look at exclusions and avoidances. */
      if (modroomexcs[nummods * r + rchanges[i]])
	{			/* If excluded from new */
	  if (modroomexcs[nummods * r + rchanges[i]] == 1)	/* place, penalise */
	    chromo->modexcrooms++;
	  else
	    chromo->modavrooms++;
	}
      if (modroomexcs[nummods * or + rchanges[i]])
	{			/* If excluded from old */
	  if (modroomexcs[nummods * or + rchanges[i]] == 1)	/* place, depenalise */
	    chromo->modexcrooms--;
	  else
	    chromo->modavrooms--;
	}
    }
  /* That has taken care of adjusting eloads and sloads. Now see if there */
  /* are violations */

  for (i = 0; i < numrooms; i++)
    for (j = 0; j < numslots; j++)
      {
	if ((v = chromo->eloads[numrooms * j + i]) > roomcaps[i][j][0])
	  if (roomcaps[i][j][0] > 0)
	    chromo->eventcapviols += (v - roomcaps[i][j][0]);
	if ((v = chromo->sloads[numrooms * j + i]) > roomcaps[i][j][1])
	  if (roomcaps[i][j][1] > 0)
	    chromo->studcapviols += (v - roomcaps[i][j][1]);
	 
      }

  chromo->modexcrooms += oldchromo->modexcrooms;
  chromo->modavrooms += oldchromo->modavrooms;
  /*      chromo->eventcapviols+=oldchromo->eventcapviols; */
  /*    chromo->studcapviols+=oldchromo->studcapviols;  */

}


void 
check_edge_violation (m1, m2, s1, s2, cl, tc)
     int m1, m2, s1, s2, *cl, *tc;
{
  register int d1, d2, t1, t2, ad1, ad2;
  *tc = *cl = 0;
  d1 = slots[s1]->day;
  d2 = slots[s2]->day;
  ad1 = slots[s1]->aday;
  ad2 = slots[s2]->aday;
  t1 = slots[s1]->time + ad1 * 1440;	/* Get actual times */
  t2 = slots[s2]->time + ad2 * 1440;
  if ((t1 == t2)		/* If they start at the same time */
      || ((t1 < t2) && (t1 + (a_modules[m1]->len) > t2))	/* or they overlap */
      || ((t2 < t1) && (t2 + (a_modules[m2]->len) > t1)))
    {
      *cl = 1;
      return;
    }
  else
  /* see if they are too close together */ if ((d1 == d2) && (alminsbetweenmods > -1))
    {
      if
	(((t1 < t2) && ((t1 + (a_modules[m1]->len) + alminsbetweenmods) > t2))
	 || ((t2 < t1) && ((t2 + (a_modules[m2]->len) + alminsbetweenmods) > t1)))
	*tc = 1;
    }
  else
    {
      if (aldaysbetweenmods > -1)
	if (abs (ad1 - ad2) < aldaysbetweenmods)
	  *tc = 1;

    }

}

void 
check_juxt_constraint (j, s1, s2, viol)
     JUXT *j;
     int s1, s2, *viol;
{
  register int rel, elm, dsm, gap, t1, t2, d1, d2, diff;

  rel = j->rel;
  elm = j->elm;
  dsm = j->dsm;
  gap = j->gap;

  t1 = slots[s1]->time + ((slots[s1]->aday) * 1440);
  t2 = slots[s2]->time + ((slots[s2]->aday) * 1440);
  d1 = slots[s1]->aday;
  d2 = slots[s2]->aday;
  if (dsm == 0)
    diff = d2 - d1;
  else
    diff = t2 - t1;


  *viol = 0;
  switch (rel)
    {
    case 0:			/****************** mod1 before mod2 *****************/
      if (elm == 0)
	{			/*    exactly  */
	  if (diff != gap)
	    *viol = 1;
	}
      else if (elm == 1)
	{			/*   at least  */
	  if ((diff > 0 && diff < gap) || (diff <= 0))
	    *viol = 1;
	}
      else if (elm == 2)
	{			/*  at most */
	  if ((diff > 0 && diff > gap) || (diff <= 0))
	    *viol = 1;
	}
      break;
    case 1:			/****************** mod1 after mod2 *****************/
      if (elm == 0)
	{			/*     exactly  */
	  if (-1 * diff != gap)
	    *viol = 1;
	}
      else if (elm == 1)
	{			/*  at least  */
	  if ((diff < 0 && diff > (-1 * gap)) || (diff >= 0))
	    *viol = 1;
	}
      else if (elm == 2)
	{			/*  at most   */
	  if ((diff < 0 && diff < (-1 * gap)) || (diff >= 0))
	    *viol = 1;
	}
      break;
    case 2:			/*********** mod1 mod2 -- between -- ************/
      if (elm == 0)
	{			/*    exactly  */
	  if (abs (diff) != gap)
	    *viol = 1;
	}
      else if (elm == 1)
	{			/*   at least  */
	  if (abs (diff) < gap)
	    *viol = 1;
	}
      else if (elm == 2)
	{			/*   at most   */
	  if (abs (diff) > gap)
	    *viol = 1;
	}
      break;
    case 3:			/* mod1 and mod2 start at same time */
      if (t1 != t2)
	*viol = 1;
      break;
    case 4:			/* mod1 and mod2 are on the same day */
      if (d1 != d2)
	*viol = 1;
      break;
    }
}

/* read in a line of text from the given file */
int 
readline (file_ptr, text_line)
     FILE *file_ptr;
     char *text_line;
{
  int charnum;
  char c;

  charnum = 0;
  while ((c = getc (file_ptr)) != EOF && c != '\n')
    if (charnum < MAXCHARS)
      {
	text_line[charnum] = c;
	charnum++;
      }
  if (c == EOF)
    return (-1);
  text_line[charnum++] = 0;
  return charnum;
}

int 
gt (a, b)
     int *a, *b;
{
  if (*a > *b)
    return 1;
  else if (*a == *b)
    return 0;
  else
    return -1;
}


int 
randchoose (violations, choices, nchoices)
     int *violations, *choices, nchoices;
{
  return (random ()% nummods);
}

/* int tnchoose(violations,choices,nchoices)
int *violations, *choices, nchoices;
{
   int choice, chosen, i;

    choice = choices[random()%nchoices];
    chosen = choice;

  for(i=1;i<tnchoose_size;i++) {
   choice = choices[random()%nchoices];
   if (violations[choice]>violations[chosen]) chosen = choice;
   else if (violations[choice]==violations[chosen])
     {if(drandom()<0.5) chosen = choice;}
 }

   return chosen;
 }
*/

/* Choose a best (highest-scoring) gene to mutate */
int 
abestchoose (violations, choices, nchoices)
     int *violations, *choices, nchoices;
{

  int worst, i;

  worst = 0;

  for (i = 1; i < nchoices; i++)
    if (violations[choices[i]] > violations[choices[worst]])
      worst = i;
    else if (violations[choices[i]] == violations[choices[worst]])
      {
	if (drandom ()< 0.5)
	  worst = i;
      }
  return worst;

}

int 
randchange (genotype, event, parent, pop)
     char *genotype;
     int event, pop;
     CHROMOSOME *parent;
{
  return ((genotype[event] + 1 + random ()% (numslots - 1)) % numslots);
}

int 
abestchange (genotype, event, parent, pop)
     char *genotype;
     int event, pop;
     CHROMOSOME *parent;
{
  int s, choice, i, viols[MAXSLOTS], e1, e2, t1,
    t2, ns, cl, tc, viol;
  LL_EDGES *e;
  LL_JUXTS *ju;
  JUXT *jc;

  for (i = 0; i < numslots; i++)
    viols[i] = 0;		/* Initialise violations */
  viols[genotype[event]] = 1000000;


  for (s = 0; s < numslots; s++)/* Now work out violations for each */
    {
      if (s == genotype[event])
	continue;

      /* Check relevant edge constraints */
      for (e = a_modules[event]->edges; e; e = e->next)
	{
	  checks[pop]++;
	  e1 = e->e->mod1->id;
	  e2 = e->e->mod2->id;
	  if (e1 == event)
	    {
	      t1 = s;
	      t2 = genotype[e2];
	    }
	  else
	    {
	      t2 = s;
	      t1 = genotype[e1];
	    }
	  ns = e->e->ns;
	  check_edge_violation (e1, e2, t1, t2, &cl, &tc);
	  if (cl)
	    viols[i] += (clashpen * ns);
	  if (tc)
	    viols[i] += (tooclosepen * ns);
	}
      for (ju = a_modules[event]->juxts; ju; ju = ju->next)
	{
	  checks[pop]++;
	  e1 = ju->j->a->id;
	  e2 = ju->j->b->id;
	  jc = ju->j;
	  if (e1 == event)
	    {
	      t1 = s;
	      t2 = genotype[e2];
	    }
	  else
	    {
	      t2 = s;
	      t1 = genotype[e1];
	    }
	  check_juxt_constraint (jc, t1, t2, &viol);
	  if (viol)
	    viols[i] += juxtpen;
	}

      /* Check exclusion and avoidance constraints */
      if (modslotexcs[nummods * s + event])
	{
	  if (modslotexcs[nummods * s + event] == 1)
	    viols[i] += modexcslotpen;
	  else
	    viols[i] += modavslotpen;
	}
      /* Check slotseat constraints, if any */

      if (slotseatconstraints)
	{			/* Need to see if putting the event in this slot overcrowds it */
	  if ((parent->ssloads[s] + a_modules[event]->numtakers) >
	      slots[s]->sscap)
	    viols[i] += ssloadpen;
	}
      if (sloteventconstraints)
	{			/* Need to see if putting the event in this slot overcrowds it */
	  if ((parent->seloads[s] + 1) > slots[s]->secap)
	    viols[i] += seloadpen;
	}

    }
  /* viols now has all the information. Choose one with minimal trouble */

  choice = 0;
  for (s = 1; s < numslots; s++)
    if (viols[s] < viols[choice])
      choice = s;
    else if (viols[s] == viols[choice])
      if (drandom ()< 0.5)
	choice = s;

  return choice;

}

int 
tnrchange (genotype, event, parent, pop)
     char *genotype;
     int event, pop;
     CHROMOSOME *parent;
{
  int r, or, choice, i, erooms[MAXROOMS], viols[MAXROOMS], s, t,
    a, d, l, v, k;

  or = parent->gdp->genotype[nummods + event];
  s = genotype[event];
  t = slots[s]->time;
  a = a_modules[event]->size;
  d = slots[s]->aday;
  l = a_modules[event]->len;

  for (i = 0; i < tnrchange_size (event); i++)
    {
      /* Pick tnchange_size new rooms at random */
      r = random ()% numrooms;
      while (modroomexcs[nummods * r + event] == 1)
	r = random ()% numrooms;
      erooms[i] = r;
    }

  for (i = 0; i < tnrchange_size (event); i++)	/* Now work out violations for each */
    {
      r = erooms[i];

      /* Given this new room or slot for the event, see how we affect */
      /*  sloads[numrooms*t+r] and eloads */

      viols[i] = 0;
      k = 0;
      while (((s + k) < numslots)
	     && (slots[s + k]->aday == d)	/* for slots on this day */
	     && ((t + l) > slots[s + k]->time))	/* which it moves into ... */
	{

	  if
	    ((v = parent->eloads[numrooms * (s + k) + r] + 1 -
	      roomcaps[r][s + k][0]) > 0)
	    viols[i] += eventcappen;

	  if
	    ((v = parent->sloads[numrooms * (s + k) + r] + a -
	      roomcaps[r][s + k][1]) > 0)
	    if (v > a)
	      viols[i] += studcappen * (v - a);
	    else
	      viols[i] += studcappen * v;

	  /* Now subtract the effect of having moved from the old room */

	  if
	    ((v = parent->eloads[numrooms * (s + k) + or] -
	      roomcaps[r][s + k][0]) > 0)
	    viols[i] -= eventcappen;

	  if
	    ((v = parent->sloads[numrooms * (s + k) + or] -
	      roomcaps[r][s + k][1]) > 0)
	    if (v > a)
	      viols[i] -= studcappen * (v - a);
	    else
	      viols[i] -= studcappen * v;
	  k++;
	}

      /* See also how we affect room exclusions and avoidances */

      if (modroomexcs[nummods * r + event])
	{			/* If excluded from */
	  if (modroomexcs[nummods * r + event] == 1)	/* here, penalise */
	    viols[i] += modexcroompen;
	  else
	    viols[i] += modavroompen;
	}

      if (modroomexcs[nummods * or + event])
	{			/* If excluded from */
	  if (modroomexcs[nummods * or + event] == 1)	/* here, depenalise */
	    viols[i] -= modexcroompen;
	  else
	    viols[i] -= modavroompen;

	}
    }

  /* viols[i] now has relative penalty information for each of */
  /* the candidates erooms[i]. So just choose a minimally poor */
  /* one as the new slot */

  choice = 0;
  for (i = 1; i < tnrchange_size (event); i++)
    if (viols[i] < viols[choice])
      choice = i;
    else if (viols[i] == viols[choice])
      if (drandom ()< 0.5)
	choice = i;

  return erooms[choice];
}


/* Currently just chooses a random new agent. MUST DO PROPERLY */

int 
tnachange (genotype, event, parent, pop)
     char *genotype;
     int event, pop;
     CHROMOSOME *parent;
{

  return (random ()% numagents);
}

int 
sim_tnchange (genotype, simbase, group, parent, pop)
     char *genotype;
     int simbase, group, pop;
     CHROMOSOME *parent;
{
  int s, choice, i, eslots[MAXSLOTS], viols[MAXSLOTS], e1, e2, t1,
    t2, ns, cl, tc, viol, ev, event, totsize = 0;
  LL_EDGES *e;
  LL_JUXTS *ju;
  JUXT *jc;

  for (i = 0; i < tnchange_size (simbase); i++)
    {
      s = random ()% numslots;
      while (modslotexcs[nummods * s + simbase] == 1)
	s = random ()% numslots;
      eslots[i] = s;		/* Pick tnchange_size new slots at random */
    }

  for (i = 0; i < tnchange_size (simbase); i++)
    viols[i] = 0;		/* Initialise violations */
  for (i = 0; i < tnchange_size (simbase); i++)	/* Now work out violations for each */
    {
      s = eslots[i];

      /* Check relevant edge constraints */
      /* Need to look at each sim-group member in turn */
      for (ev = 0; ev < simgroups[group]->num; ev++)
	{
	  event = simgroups[group]->sims[ev];
	  for (e = a_modules[event]->edges; e; e = e->next)
	    if (e->e->ns > 0)
	      {
		checks[pop]++;
		e1 = e->e->mod1->id;
		e2 = e->e->mod2->id;
		if (e1 == event)
		  {
		    t1 = s;
		    t2 = genotype[e2];
		  }
		else
		  {
		    t2 = s;
		    t1 = genotype[e1];
		  }
		ns = e->e->ns;
		check_edge_violation (e1, e2, t1, t2, &cl, &tc);
		if (cl)
		  viols[i] += ((clashpen * ns) + nclashpen);
		if (tc)
		  viols[i] += ((tooclosepen * ns) + ntooclosepen);
	      }
	}
      for (ev = 0; ev < simgroups[group]->num; ev++)
	{
	  event = simgroups[group]->sims[ev];
	  for (ju = a_modules[event]->juxts; ju; ju = ju->next)
	    {
	      checks[pop]++;
	      e1 = ju->j->a->id;
	      e2 = ju->j->b->id;
	      jc = ju->j;
	      if (e1 == event)
		{
		  t1 = s;
		  t2 = genotype[e2];
		}
	      else
		{
		  t2 = s;
		  t1 = genotype[e1];
		}
	      check_juxt_constraint (jc, t1, t2, &viol);
	      if (viol)
		viols[i] += jc->priority;
	    }
	}
      /* Check exclusion and avoidance constraints */

      for (ev = 0; ev < simgroups[group]->num; ev++)
	{
	  event = simgroups[group]->sims[ev];
	  if (modslotexcs[nummods * s + event])
	    {
	      if (modslotexcs[nummods * s + event] == 1)
		viols[i] += modexcslotpen;
	      else
		viols[i] += modavslotpen;
	    }
	}
      /* Check slotseat constraints, if any */

      if (slotseatconstraints)
	{			/* Need to see if putting the event in this slot overcrowds it */
	  /* get total size of events */
	  for (ev = 0; ev < simgroups[group]->num; ev++)
	    totsize += a_modules[simgroups[group]->sims[ev]]->size;
	  if (parent->gdp->genotype[simbase] == s)
	    totsize = 0;
	  if ((parent->ssloads[s] + totsize) >
	      slots[s]->sscap)
	    viols[i] += ssloadpen * (parent->ssloads[s] + totsize - slots[s]->sscap);
	}
      if (sloteventconstraints)
	{			/* Need to see if putting the event in this slot overcrowds it */
	  totsize = simgroups[group]->num;
	  if (parent->gdp->genotype[simbase] == s)
	    totsize = 0;
	  if ((parent->seloads[s] + totsize) > slots[s]->secap)
	    viols[i] += seloadpen * (parent->seloads[s] + totsize - slots[s]->secap);
	}
    }

  /* viols[i] now has relative penalty information for each of */
  /* the candidates eslots[i]. So just choose a minimally poor */
  /* one as the new slot */

  choice = 0;
  for (i = 1; i < tnchange_size (simbase); i++)
    if (viols[i] < viols[choice])
      choice = i;
    else if (viols[i] == viols[choice])
      if (drandom ()< 0.5)
	choice = i;

  return eslots[choice];

}


int 
tnchange (genotype, event, parent, pop)
     char *genotype;
     int event, pop;
     CHROMOSOME *parent;
{
  int s, choice, i, eslots[MAXSLOTS], viols[MAXSLOTS], e1, e2, t1,
    t2, ns, cl, tc, viol, totsize;
  LL_EDGES *e;
  LL_JUXTS *ju;
  JUXT *jc;

  for (i = 0; i < tnchange_size (event); i++)
    {
      s = random ()% numslots;
      while (modslotexcs[nummods * s + event] == 1)
	s = random ()% numslots;
      eslots[i] = (int) s;	/* Pick tnchange_size new slots at random */
    }

  for (i = 0; i < tnchange_size (event); i++)
    viols[i] = 0;		/* Initialise violations */

  for (i = 0; i < tnchange_size (event); i++)	/* Now work out violations for each */
    {
      s = eslots[i];

      /* Check relevant edge constraints */
      for (e = a_modules[event]->edges; e; e = e->next)
	if (e->e->ns > 0)
	  {
	    checks[pop]++;
	    e1 = e->e->mod1->id;
	    e2 = e->e->mod2->id;
	    if (e1 == event)
	      {
		t1 = s;
		t2 = genotype[e2];
	      }
	    else
	      {
		t2 = s;
		t1 = genotype[e1];
	      }
	    ns = e->e->ns;
	    check_edge_violation (e1, e2, t1, t2, &cl, &tc);
	    if (cl)
	      viols[i] += ((clashpen * ns) + nclashpen);
	    if (tc)
	      viols[i] += ((tooclosepen * ns) + ntooclosepen);
	  }
      for (ju = a_modules[event]->juxts; ju; ju = ju->next)
	{
	  checks[pop]++;
	  e1 = ju->j->a->id;
	  e2 = ju->j->b->id;
	  jc = ju->j;
	  if (e1 == event)
	    {
	      t1 = s;
	      t2 = genotype[e2];
	    }
	  else
	    {
	      t2 = s;
	      t1 = genotype[e1];
	    }
	  check_juxt_constraint (jc, t1, t2, &viol);
	  if (viol)
	    viols[i] += jc->priority;
	}

      
      /* Check exclusion and avoidance constraints */

      if (modslotexcs[nummods * s + event])
	{
	  if (modslotexcs[nummods * s + event] == 1)
	    viols[i] += modexcslotpen;
	  else
	    viols[i] += modavslotpen;
	}
      /* Check slotseat constraints, if any */

      if (slotseatconstraints)
	{			/* Need to see if putting the event in this slot overcrowds it */
	  if (parent->gdp->genotype[event] == s)
	    totsize = 0;
	  else
	    totsize = a_modules[event]->size;
	  if ((parent->ssloads[s] + totsize) >
	      slots[s]->sscap)
	    viols[i] += ssloadpen * (parent->ssloads[s] + totsize - slots[s]->sscap);
	}
      if (sloteventconstraints)
	{			/* Need to see if putting the event in this slot overcrowds it */
	  if (parent->gdp->genotype[event] == s)
	    totsize = 0;
	  else
	    totsize = 1;
	  if ((parent->seloads[s] + totsize) > slots[s]->secap)
	    viols[i] += seloadpen * (parent->seloads[s] + totsize - slots[s]->secap);
	}

    }

  /* viols[i] now has relative penalty information for each of */
  /* the candidates eslots[i]. So just choose a minimally poor */
  /* one as the new slot */

  choice = 0;
  for (i = 1; i < tnchange_size (event); i++)
    if (viols[i] < viols[choice])
      choice = i;
    else if (viols[i] == viols[choice])
      if (drandom ()< 0.5)
	choice = i;

  return eslots[choice];

}


int 
findbestchromandcheck (int flag, CHROMOSOME chromo)
{
  int i, j, k, scv, thisroomstuds, cscv;


  scv = chromo.studcapviols;
  cscv = 0;
  for (i = 0; i < numslots; i++)

    {
      for (j = 0; j < numrooms; j++)
	{
	  thisroomstuds = 0;
	  {
	    for (k = 0; k < nummods; k++)
	      if ((chromo.gdp->genotype[nummods + k] == j) &&
		  (chromo.gdp->genotype[k] == i))
		thisroomstuds += a_modules[k]->size;
	    if (roomcaps[j][i][1] < thisroomstuds)
	      cscv += (thisroomstuds - roomcaps[j][i][1]);
	  }
	}
    }

  if (scv == cscv)
    return (1);
  else
    {
      fprintf (output, "\n");
      if (flag == 0)
	fprintf (output, "buggered after slot mutate\n");
      else
	fprintf (output, "buggered after room mutate\n");
      return (1);
    }
}


int 
tnchange_size (int event)
{
  return (tnscs[event]);

}

int 
tnrchange_size (int event)
{
  return (tnrcs[event]);
}

int 
tnachange_size (int event)
{
  return (tnacs[event]);
}

int 
maxof (a, b)
     int a, b;
{
  if (a < b)
    return b;
  else
    return a;
}



/* Take a randomly chosen pair of distinct slots, move the things in */
/* one slot to the other slot */

CHROMOSOME 
period_swap_mutate (genotype, parent, pop, n)
     char *genotype;
     CHROMOSOME *parent;
     int pop, n;
{
  CHROMOSOME newchromo;
  int s1, s2, i;

  nschanges = nrchanges = nachanges = 0;
  for (i = 0; i < nummods; i++)
    schanged[i] = 0;
  if (roomsmatter)
    for (i = 0; i < nummods; i++)
      rchanged[i] = 0;


  create_chromo (&newchromo);	/* First, create this child */
  copy_genotype (parent, &newchromo);	/* copy the genotype across */
  copy_loads (parent, &newchromo);

  /* Pick two random slots */

  s1 = random ()% numslots;
  s2 = random ()% numslots;

  while (s2 == s1)
    s2 = random ()% numslots;

  for (i = 0; i < geno_size; i++)
    if (newchromo.gdp->genotype[i] == s1)
      {
	if (modslotexcs[nummods * s2 + i] != 1)
	  newchromo.gdp->genotype[i] = s2;
      }
    else if (newchromo.gdp->genotype[i] == s2)
      {
	if (modslotexcs[nummods * s1 + i] != 1)
	  newchromo.gdp->genotype[i] = s1;
      }

  return (newchromo);
}


int intcompare(int *i, int *j)
{
  if (*i > *j)
    return (1);
  if (*i < *j)
    return (-1);
  return (0);
}

